/*
 Licensed Materials - Property of IBM - some unique copyrights (i.e., the ability to copy,  modify,
 distribute) may be licensed in certain circumstances (e.g., sample/template code  intended to be used by
 customers to build upon and distribute as their own products).

 (c) Copyright IBM Corporation 2011, 2015. All Rights Reserved.

 U.S. Government Users Restricted Rights: Use, duplication or disclosure restricted by GSA ADP
 Schedule Contract with IBM Corp.
*/
(function () {
dojo.provide("com.ibm.rdm.web.artifactgrid.ArtifactCommands");

dojo.require("dojox.xml.parser");
dojo.require("jazz.core.loader");

dojo.require("com.ibm.rdm.web.artifactgrid.ArtifactConstants");
dojo.require("com.ibm.rdm.web.common.MimeTypes");
dojo.require("com.ibm.rdm.web.common.MessageHandler");
dojo.require("com.ibm.rdm.web.common.PermalinkDialog");
dojo.require("com.ibm.rdm.web.common.ProjectCache");
dojo.require("com.ibm.rdm.web.common.ResourceLinksHandler");
dojo.require("com.ibm.rdm.web.common.StatusDisplay");
dojo.require("com.ibm.rdm.web.external.CsvExport");
dojo.require("com.ibm.rdm.web.external.ImportWizard");
dojo.require("com.ibm.rdm.web.grid.Operation");
dojo.require("com.ibm.rdm.web.links.linkutils");
dojo.require("com.ibm.rdm.web.locks.LockManager");
dojo.require("com.ibm.rdm.web.model.resources.ResourceManager");
dojo.require("com.ibm.rdm.web.model.types.RRMTypes");
dojo.require("com.ibm.rdm.web.model.types.TypesManager");
dojo.require("com.ibm.rdm.web.notifications.ChangeNotifier");
dojo.require("com.ibm.rdm.web.pages.ActionViewCache");
dojo.require("com.ibm.rdm.web.picker.CopyFolderPicker");
dojo.require("com.ibm.rdm.web.picker.MoveFolderPicker");
dojo.require("com.ibm.rdm.web.project.DashboardConstants");
dojo.require("com.ibm.rdm.web.project.EditManager");
dojo.require("com.ibm.rdm.web.project.ManageSecurityForm");
dojo.require("com.ibm.rdm.web.project.NewArtifactForm");
dojo.require("com.ibm.rdm.web.query.ViewServiceClient");
dojo.require("com.ibm.rdm.web.reporting.ReportWizard");
dojo.require("com.ibm.rdm.web.review.reviewutils");
dojo.require("com.ibm.rdm.web.server.cache.ProcessOperationsCache");
dojo.require("com.ibm.rdm.web.server.client.utils.ResourceUtil");
dojo.require("com.ibm.rdm.web.suspicion.ClearSuspicionForm");
dojo.require("com.ibm.rdm.web.suspicion.SuspectLinksUtil");
dojo.require("com.ibm.rdm.web.tags.TagUtil");
dojo.require("com.ibm.rdm.web.util.clipboard");
dojo.require("com.ibm.rdm.web.viewer.CollectionsMenu");
dojo.require("com.ibm.rdm.web.links.SimpleLinkDialog");

// Loaded via load_async
//dojo.require("com.ibm.rdm.web.linksexplorer.LinksExplorerDialog");

dojo.requireLocalization("com.ibm.rdm.web.artifactgrid", "ArtifactCommands");
/*
 * Supports the Resource Grid with a basic set of (non-edit) commands.
 */
 
var ChangeNotifier = com.ibm.rdm.web.notifications.ChangeNotifier;
var CopyFolderPicker = com.ibm.rdm.web.picker.CopyFolderPicker;
var CsvExport = com.ibm.rdm.web.external.CsvExport;
var ManageSecurityForm = com.ibm.rdm.web.project.ManageSecurityForm;
var MimeTypes = com.ibm.rdm.web.common.MimeTypes;
var MoveFolderPicker = com.ibm.rdm.web.picker.MoveFolderPicker;
var NewArtifactForm = com.ibm.rdm.web.project.NewArtifactForm;
var ProcessOperationsCache = com.ibm.rdm.web.server.cache.ProcessOperationsCache;
var PermalinkDialog = com.ibm.rdm.web.common.PermalinkDialog;
var ProjectCache = com.ibm.rdm.web.common.ProjectCache;
var RRMTypes = com.ibm.rdm.web.model.types.RRMTypes;
var ResourceLinksHandler = com.ibm.rdm.web.common.ResourceLinksHandler;
var ResourceManager = com.ibm.rdm.web.model.resources.ResourceManager;
var ResourceUtil = com.ibm.rdm.web.server.client.utils.ResourceUtil;
var StatusDisplay = com.ibm.rdm.web.common.StatusDisplay;
var TagUtil = com.ibm.rdm.web.tags.TagUtil;
var ViewServiceClient = com.ibm.rdm.web.query.ViewServiceClient;
var reviewutils = com.ibm.rdm.web.review.reviewutils;
var TypesManager = com.ibm.rdm.web.model.types.TypesManager;
var SuspectLinksUtil = com.ibm.rdm.web.suspicion.SuspectLinksUtil;
var LockManager = com.ibm.rdm.web.locks.LockManager;
var	linkutils = com.ibm.rdm.web.links.linkutils;
var clipboard = com.ibm.rdm.web.util.clipboard;
var Operation = com.ibm.rdm.web.grid.Operation;

var COPY_ACTION = clipboard.Actions.COPY;
var MAX_NUM_ROWS_TO_ENABLE_OPEN_IN_NEW_TAB_OR_WINDOW = 5;


dojo.declare(
	"com.ibm.rdm.web.artifactgrid.ArtifactCommands",
	null,
{
	//The type of the view
	_viewType: null,
	
	//The DataCache to use
	_dataCache: null,
	
	//The TypeCache to use
	_typeCache: null,
	
	//The project URI in context
	_projectURI: null,
	
	//Internationalisation
	_messages: null,
	
	//The operations (sort of like permissions) allowed on this project
	_projectOperations: null,
	
	// _selectedFolderURI: URI
	//		The URI of the last selected folder in a copy dialog.
	_selectedFolderURI: null,
	
	// _isInHistoryMode: URI
	//		bool - artifact history view
	_isInHistoryMode: false,
	
	// _isInReviewMode: URI
	//		bool - Artifact is in review
	_isInReviewMode: false,

/////////////////////////////////////////////////////////////////////////////////
//// Creation and Initialiation

	constructor: function(params) {
		// TODO: should use dojo.mixin
		this._dataCache = params.dataCache;
		this._typeCache = params.typeCache;
		this._projectURI = params.projectURI;
		this._baselineURI = params.baselineURI;
		this._viewType = params.viewType;
		this._isInHistoryMode = params.isInHistoryMode;
		this._isInReviewMode = params.isInReviewMode;
		this.grid = params.grid;
		this._messages = dojo.i18n.getLocalization("com.ibm.rdm.web.artifactgrid", "ArtifactCommands");
		
		this._onExecConn = dojo.connect(Operation.prototype, "onExecute", this, function () {
			this.clearGridSelection();
		});
		
	},
	
/////////////////////////////////////////////////////////////////////////////////
//// Main API

	setEditInterface: function(editInterface) {
		this._editInterface = editInterface;
	},
	
	setLinksInterface: function(linksInterface) {
		this._linksInterface = linksInterface;
	},
	
	setDashboardInterface: function(dashboardInterface) {
		this._dashboardInterface = dashboardInterface;
	},

	setProjectOperations: function(projectOperations) {
		this._projectOperations = projectOperations;
	},

	getOperations: function(caller) {
		var self = this;
		var menuOperations = {
			".delete": //Params: _id, _iconClass, _displayText, _isEnabledMethod, _executeMethod
				new Operation(".delete", "deleteCommand",
					function(numSelected) {return self.deleteText(numSelected); },
					function(rowIds, count) {return self.isDeleteEnabled(rowIds, count);},
					function(rowIds) {return self.executeDelete(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getDeleteEnablementDeferred(rowIds, count);}),
			".open": //Params: _id, _iconClass, _displayText, _isEnabledMethod, _executeMethod
				new Operation(".open", "openCommand",
					function(numSelected) {return self._messages.openArtifact; },
					function(rowIds, count) {return self.isOpenEnabled(rowIds, count);},
					function(rowIds) {self.executeOpen(rowIds);},
					function(rowIds, count) {
						return count === 1 && self._defaultIsVisible(rowIds, count);
					}),
			".openInNewTab":
				new Operation(".openInNewTab", "openInNewTabCommand",
					function(numSelected) {return self.openArtifactInNewTabText(numSelected);},
					function(rowIds, count) {return self.isOpenInNewTabOrWindowEnabled(count);},
					function(rowIds) {self.executeOpenInNewTab(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					null,
					dojo.hitch(this, function(count){
						return this._buildPreexistingBrowserBasedKeyCombo(".openInNewTab", count);
					})),
			".openInNewWindow":
				new Operation(".openInNewWindow", "openInNewWindowCommand",
					function(numSelected) {return self.openArtifactInNewWindowText(numSelected);},
					function(rowIds, count) {return self.isOpenInNewTabOrWindowEnabled(count);},
					function(rowIds) {self.executeOpenInNewWindow(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					null,
					dojo.hitch(this, function(count){
						return this._buildPreexistingBrowserBasedKeyCombo(".openInNewWindow", count);
					})),
			".openInNewSideBySideWindows":
				new Operation(".openInNewSideBySideWindows", "openInNewSideBySideWindowsCommand",
					function(numSelected) {return self._messages.openArtifactsSideBySide;},
					function(rowIds, count) {return count === 2;},
					function(rowIds) {self.executeOpenInNewSideBySideWindows(rowIds);},
					function(rowIds, count) {
						return count === 2 && self._defaultIsVisible(rowIds, count);
					}),
			".selectTags":
				new Operation(".selectTags", "selectTagsCommand",
					function(numSelected) {return self.selectTagsText(numSelected); },
					function(rowIds, count) {return self.isSelectTagsEnabled(rowIds, count);},
					function(rowIds) {self.executeSelectTags(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getSelectTagsDeferred(rowIds, count);}),
			".move":
				new Operation(".move", "moveCommand",
					function(numSelected) {return self.moveText(numSelected); },
					function(rowIds, count) {return self.isMoveEnabled(rowIds, count);},
					function(rowIds) {self.executeMove(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					// Note: by design, we don't provide a "getEnablementDeferred" argument
					// here.  It used to be that we'd check for whether the current user
					// is allowed to delete an artifact of the selected artifact's type
					// before enabling Move but that led to the customer issue reported
					// in Defect 67442 and therefore we're removing this check ...
					//
					// Update by T. Dalbo:  We still need the permissions server call to have
					// returned before a move can be enabled.  There simply won't be a check to
					// see whether a user has DELETE permissions.
					function(rowIds, count) {return self.getMoveEnabledDeferred(rowIds, count);}
				),
			".showUrl":
				new Operation(".showUrl", "showUrlCommand",
					function(numSelected) {return self.showArtifactUrlText(numSelected);},
					function(rowIds, count) {return true;},
					function(rowIds) {self.executeShowUrl(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);}
				),
			".lock":
				new Operation(".lock", "lockCommand",
					function(numSelected) {return self.lockText(numSelected); },
					// disable lock command until we know that the user has permission
					function(rowIds, count) {return false;},
					function(rowIds) {self.executeLock(rowIds);},
					function(rowIds, count) {return self.isLockVisible(rowIds, count);},
					function(rowIds, count) {return self.getLockRowEnablementDeferred(rowIds, count);}),				
			".unlock":
				new Operation(".unlock", "unlockCommand",
					function(numSelected) {return self.unlockText(numSelected);},
					function(rowIds, count) {return false;},
					function(rowIds) {self.executeUnlock(rowIds);},
					function(rowIds, count) {return self.isUnlockVisible(rowIds, count);},
					function(rowIds, count) {return self.getReleaseLockEnablementDeferred(rowIds, count);}),					
			".copy":
				new Operation(".copy", "copyCommand",
					function(numSelected) {return self.copyText(numSelected); },
					function(rowIds, count) {return self.isCopyEnabled(rowIds, count);},
					function(rowIds) {self.executeCopy(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getCopyEnabledDeferred(rowIds, count);},
					dojo.hitch(this, function(){return this._getKeyComboForOperation(".copy");})),
			".duplicate":
				new Operation(".duplicate", "duplicateCommand",
					function(numSelected) {return self.duplicateText(numSelected); },
					function(rowIds, count) {return self.isCopyEnabled(rowIds, count);},
					function(rowIds) {self.executeDuplicate(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getCopyEnabledDeferred(rowIds, count);},
					dojo.hitch(this, function(){return this._getKeyComboForOperation(".duplicate");})),
			".pasteAsLink":
				new Operation(".pasteAsLink", "pasteAsLinkCommand",
					function(numSelected) {return self.pasteText(numSelected); },
					function(rowIds, count) {return self.isPasteAsLinkEnabled(rowIds, count);},
					function(rowIds) {self.executePasteAsLink(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					null,
					dojo.hitch(this, function(){return this._getKeyComboForOperation(".pasteAsLink");})),
			".createSimilar":
				new Operation(".createSimilar", "newArtifactCommand",
					function(numSelected, rowIds, key) {return self.createSimilarText(numSelected, rowIds, key); },
					function(rowIds, count) {return self.isCreateSimilarEnabled(rowIds, count);},
					function(rowIds) {return self.executeCreateSimilar(rowIds, false);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getCreateSimilarEnabled(rowIds, count);},
					dojo.hitch(this, function(){return this._getKeyComboForOperation(".createSimilar");})),
			".createModuleChild":
				new Operation(".createModuleChild", "createModuleChildCommand",
					function(numSelected, rowIds, key) {return self.moduleChildText(numSelected, rowIds, key); },
					function(rowIds, count) {return self.isCreateSimilarEnabled(rowIds, count);},
					function(rowIds) {return self.executeCreateSimilar(rowIds, true);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getCreateSimilarEnabled(rowIds, count);}),
			".export":
				new Operation(".export", "exportCommand",
					function(numSelected) {return self.exportText(numSelected); },
					function(rowIds, count) {return self.isExportEnabled(rowIds, count);},
					function(rowIds) {self.executeExport(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getExportEnabledDeferred(rowIds, count);}),
			".manageSecurity":
				new Operation(".manageSecurity", "manageSecurityCommand",
					function(numSelected, rowIds) {return self.selectManageSecurityText(numSelected, rowIds);},
					function(rowIds, count) {return self.isManageSecurityEnabled(rowIds, count);},
					function(rowIds) {self.executeManageSecurity(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getManageSecurityEnablementDeferred(rowIds, count);}),
			".createReview":
				new Operation(".createReview", "createReviewCommand",
					function(numSelected, rowIds) {return self.selectCreateReviewText(numSelected, rowIds);},
					function(rowIds, count) {return self.isCreateReviewEnabled(rowIds, count);},
					function(rowIds) {self.executeCreateReview(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getCreateReviewDeferred(rowIds, count);}),
			".addToCollection":
				new Operation(".addToCollection", "addToCollectionCommand",
					function(numSelected) {return self.selectAddToCollectionText(numSelected);},
					function(rowIds, count) {return self.isAddToCollectionEnabled(rowIds, count);},
					function(rowIds) {self.executeAddToCollection(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getAddToCollectionDeferred(rowIds, count);}),
			".loading":
				new Operation(".loading", "loadingDisplay",
					function() {return self._messages.loadingLinkTypes;},
					function() {return false;},
					function() {}),				
			".remove": //Params: _id, _iconClass, _displayText, _isEnabledMethod, _executeMethod
				new Operation(".remove", "removeCommand",
					function(numSelected) {return self.removeText(numSelected); },
					function(rowIds, count) {return self.isRemoveEnabled(rowIds, count);},
					function(rowIds) {self.executeRemove(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getRemoveEnablementDeferred(rowIds, count);}),
			".clearSuspicion": //Params: _id, _iconClass, _displayText, _isEnabledMethod, _executeMethod
				new Operation(".clearSuspicion", "clearSuspicionCommand",
					function(numSelected) {return self._messages.clearSuspicion; },
					function(rowIds, count) {return self.isClearSuspicionEnabled(rowIds, count);},
					function(rowIds) {self.executeClearSuspicion(rowIds);},
					function(rowIds, count) {return self._isClearSuspicionVisible(rowIds, count);},
					function(rowIds, count) {return self.getClearSuspicionEnablementDeferred(rowIds, count);}),
			".createReport": //Params: _id, _iconClass, _displayText, _isEnabledMethod, _executeMethod
				new Operation(".createReport", "createReportCommand",
					function(numSelected) {return self.reportGenerateText(numSelected); },
					function(rowIds, count) {return self.isReportGenerate(rowIds, count);},
					function(rowIds) {self.doGenerateReport(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);}),
			".parseRequirements": //Params: _id, _iconClass, _displayText, _isEnabledMethod, _executeMethod
				new Operation(".parseRequirements", "parseRequirementsCommand",
					function(numSelected) {return self.parseRequirementText(numSelected); },
					function(rowIds, count) {return self.isParseEnabled(rowIds, count);},
					function(rowIds) {self.doParseRequirements(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getParseEnabledDeferred(rowIds, count);})					
		};
		
		if (this._editInterface) {
			menuOperations[".editAttribs"] = 
				new Operation(".editAttribs", "editAttribsCommand",
					function(numSelected) {return self._editInterface.editAttribsText(numSelected);},
					function(rowIds, count) {
						// Note: We need to hear back from deferreds (in
						// getEditAttributesEnablementDeferred) so for now return
						// false (knowing that we'll conditionally enable this
						// menu item when we hear back from our deferreds).
						return false;
					},
					function(rowIds) {
						var rowIdsArray = self._getUniqueKeysFromRowIds(rowIds);
						var resourceInfos = self._getRecordInfos(rowIdsArray);
						self._editInterface.executeEditAttribs(resourceInfos);
					},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getEditAttributesEnablementDeferred(rowIds, count);});
			menuOperations[".editRow"] = 
				new Operation(".editRow", "editRowCommand",
					function(numSelected) {return self._messages.editArtifact; },
					function(rowIds, count) {
						// Note: We need to hear back from deferreds (in
						// getEditRowEnablementDeferred) so for now return
						// false (knowing that we'll conditionally enable this
						// menu item when we hear back from our deferreds).
						return false;
					},
					function(rowIds) {self._editInterface.editRow(rowIds);},
					function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
					function(rowIds, count) {return self.getEditRowEnablementDeferred(rowIds, count);});
		}
		
		menuOperations[".openDiagramView"] = //Params: _id, _iconClass, _displayText, _isEnabledMethod, _executeMethod
			new Operation(".openDiagramView", "openDiagramViewCommand",
				function(numSelected) {return self.openLinksExplorerText(numSelected); },
				function(rowIds, count) {return true;},
				function(rowIds) {self.openLinksExplorer(rowIds);},
				function(rowIds, count) {return self._defaultIsVisible(rowIds, count);});


		var isCollection = (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.COLLECTION);
		var isModule = (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE);
		var supportsEdit = !!this._editInterface;
		var supportsLinks = !!this._linksInterface;

		//This defines how the menu is laid out - ie the grouping etc.
		var menuLayout = [];

		var editGroupOne = {
			groupName: "EditGroupOne",
			isShown: false
		},
		editGroupTwo = {
			groupName: "EditGroupTwo",
			isShown: false
		};

		var operations = [];
		operations.push({id: ".createSimilar"});
		if (isModule) {
		    operations.push(
		    	{id: ".createModuleChild"},
			    {id: ".toggleHeading"}
		    );
		}
		editGroupOne.operations = operations;
		menuLayout.push(editGroupOne);

		var openGroup = {
			groupId: "OpenGroup",
			iconClass: "openCommand",
			separatorAbove: true,
			groupName: dojo.hitch(this, function(count) {
				if (count > 1) {
					return this._messages.openArtifacts;
				}
				return this._messages.openArtifact;
			}),
			isShown: true,
			isEnabled: dojo.hitch(this, function(rowIds, count) {return true;}),
			isVisible: dojo.hitch(this, function(rowIds, count) {return this._defaultIsVisible(rowIds, count);}),
			operations: [
				{id: ".open"},
				{id: ".openInNewSideBySideWindows"},
				{id: ".openInNewTab"},
				{id: ".openInNewWindow"}
			]
		};
		menuLayout.push(openGroup);

		operations = [];
		if (supportsEdit) {
			operations.push(
				{id: ".editAttribs"},
				{id: ".editRow"}
			);
		}
		operations.push(
			{id: ".selectTags", separatorAbove: true},
			{id: ".copy", separatorAbove: true},
			{id: ".duplicate"},
			{id: ".move"},
			{id: ".pasteAsLink"}
		);
		if (!isCollection) { operations.push({id: ".delete"}); }
		if (supportsEdit) {		
			operations.push(
				{id: ".lock", separatorAbove: true},
				{id: ".unlock", separatorAbove: true}
			);
		}
		operations.push(
			{id: ".manageSecurity"},
			{id: ".createReview", separatorAbove: true},
			{id: ".addToCollection"},
			{id: ".export"}
		);
		if (!isModule && !isCollection) { operations.push({id: ".parseRequirements"}); }
		if (!isModule) { operations.push({id: ".createReport"}); }
		operations.push({id: ".rename"});
		if (isModule) { operations.push({id: ".comment", separatorAbove: true}); }
		operations.push({id: ".clearSuspicion"});
		if (isCollection) { operations.push({id: ".remove"}); }
		if (!isModule) { operations.push({id: ".openDiagramView", separatorAbove: true}); }
		editGroupTwo.operations = operations;
		menuLayout.push(editGroupTwo);
		
		if (supportsLinks) {
			var group2 = {
				//Intially the linking group is loading.
				groupId: "CreateLinks", 
				groupName: function(count) {return self._linksInterface.getLinkGroupText(count);},
				isShown: true,
				isEnabled: function(rowIds, count) {return self._linksInterface.isLinkGroupEnabled(rowIds, count);},
				isVisible: function(rowIds, count) {return self._defaultIsVisible(rowIds, count);},
				iconClass: "addLinksCommand",
				operations: [{id: ".loading"} ]
			};
			menuLayout.push(group2);
		}
		
		var group3 = {
			groupId: "ShowURLGroup",
			isShown: false,
			isEnabled: true,
			operations: [{id: ".showUrl"}]
		};
		menuLayout.push(group3);
		
		return {operations: menuOperations, layout: menuLayout};
	},

	openArtifactInNewTabText: function(count){
		if (count === 1 || count > MAX_NUM_ROWS_TO_ENABLE_OPEN_IN_NEW_TAB_OR_WINDOW) {
			return this._messages.openArtifactInNewTabOne;
		}
		return dojo.string.substitute(this._messages.openArtifactInNewTabN, [count]);
	},

	openArtifactInNewWindowText: function(count){
		if (count === 1 || count > MAX_NUM_ROWS_TO_ENABLE_OPEN_IN_NEW_TAB_OR_WINDOW) {
			return this._messages.openArtifactInNewWindowOne;
		}
		return dojo.string.substitute(this._messages.openArtifactInNewWindowN, [count]);
	},

	isOpenInNewTabOrWindowEnabled: function(count) {
		return count >= 1 && count <= MAX_NUM_ROWS_TO_ENABLE_OPEN_IN_NEW_TAB_OR_WINDOW;
	},

	executeOpenInNewTab: function(rowIds) {
		var resourceURIs = this._getRowIdsSortedByRowIndex(rowIds),
			i;
		for (i = 0; i < resourceURIs.length; i++) {
			rowIds = {};
			rowIds[resourceURIs[i]] = true;
			this.executeOpen(rowIds, true);
		}
	},

	executeOpenInNewWindow: function(rowIds) {
		// summary:
		//		Open the passed rowIds in new browser windows.
		// description:
		//		Use a tiling approach to open new browser windows starting at position
		//		(100, 100).  If possible, try to make these new windows use the same
		//		internal dimensions as those found in the current browser window.

		var getDimensionWhichStillEnablesTiling = dojo.hitch(
			this,
			function(originWidthOrHeight, screenWidthOrHeight, browserWidthOrHeight,
					 tileIncrement, numWindowsToOpen, isHeight) {
				// summary:
				//		Get the width or height to use for a launched window.
				// description:
				//		If possible, we want to use the same internal content dimensions
				//		as the current browser window's content area (i.e. window.innerWidth
				//		and window.innerHeight) but it _may_ be the case that, due to
				//		starting at originWidthOrHeight and then using a tiling placement
				//		approach, we may not have enough real estate to use honor mimicking
				//		the current browser window's internal content dimesions.  This means
				//		we have to shrink it enough to still allow us to introduce the new
				//		windows at a particular origin and tile them appropriately.
				var valueToUse,
					tilingDelta = tileIncrement * (numWindowsToOpen-1),
					screenGapAffordance = 100, // To account for the screen menu area on Mac computers
											   // as well as taskbar (potentially on the right hand side
											   // with large icons) on Windows computers.
					browserHeaderAffordance = 50; // To account for the fact that if you pass height
												  // to window.open it's the height of the content
												  // area only.
				if (browserWidthOrHeight
					> (screenWidthOrHeight - originWidthOrHeight - tilingDelta
					   - screenGapAffordance - browserHeaderAffordance)
					) {
					valueToUse = Math.min(
						(screenWidthOrHeight - originWidthOrHeight - tilingDelta
						 - screenGapAffordance - browserHeaderAffordance),
						(
							isHeight ?
							this._getHeightValueToUseForWindowOpen() :
							this._getWidthValueToUseForWindowOpen()
						)
					);
				} else {
					if (isHeight) {
						valueToUse = this._getHeightValueToUseForWindowOpen();
					} else {
						valueToUse = this._getWidthValueToUseForWindowOpen();
					}
				}
				return valueToUse;
			}
		);
		var sortedResourceURIs = this._getRowIdsSortedByRowIndex(rowIds),
			numWindowsToOpen = sortedResourceURIs.length,
			screenWidth = window.screen.width,
			screenHeight = window.screen.height,
			browserWidth = window.outerWidth,
			browserHeight = window.outerHeight,
			top = 100,
			left = 100,
			xTileIncrement = 25,
			yTileIncrement = 25,
			widthToUse = getDimensionWhichStillEnablesTiling(
				left, screenWidth, browserWidth, xTileIncrement, numWindowsToOpen, false
			),
			heightToUse = getDimensionWhichStillEnablesTiling(
				top, screenHeight, browserHeight, yTileIncrement, numWindowsToOpen, true
			);

		dojo.forEach(
			sortedResourceURIs,
			dojo.hitch(this, function(resourceURI) {
				this._openInNewWindow(
					resourceURI,
					top,
					left,
					heightToUse,
					widthToUse
				);
				top += yTileIncrement;
				left += xTileIncrement;
			})
		);
	},

	executeOpenInNewSideBySideWindows: function(rowIds) {
		// summary:
		//		Open the two selected artifacts (found in the passed 'rowIds')
		//		in new browser windows in a side-by-side fashion.
		// description:
		//		Open two browser windows for the passed artifacts in a
		//		side-by-side fashion in such a way that we take all the
		//		available display real-estate that we can get our hands on.
		//		Some notes about this:
		//		- On Windows, Chrome doesn't properly account for the taskbar
		//		  [in the window.screen.availHeight value] if it's positioned
		//		  on the top or bottom (and there's no way for us to sense
		//		  this), so we default to using 90% of the available height for
		//		  Chrome on Windows (otherwise such things as the browser's
		//		  status bar would not be visible).
		//		- On Windows, in order to have the left window appear with its
		//		  right border "touching" the right window's left border, a
		//		  certain amount of "tweaking" is needed (via setting a
		//		  'browserSpecificWidthSubtractor' variable to an appropriate
		//		  value).

		var sortedResourceURIs = this._getRowIdsSortedByRowIndex(rowIds),
			availableWidth = window.screen.availWidth,
			availableHeight = window.screen.availHeight,
			widthToUse = Math.floor(availableWidth / 2.0),
			browserSpecificWidthSubtractor = 0,
			heightToUse = (
				dojo.isChrome && !dojo.isMac ?
				Math.floor(availableHeight * 0.9) :
				availableHeight
			),
			originX = 0,
			originY = 0;

		// Set 'browserSpecificWidthSubtractor' to an appropriate value.
		if (dojo.isSafari
			|| dojo.isChrome
			|| dojo.isFF
			) {
			// Handle Safari/Chrome/Firefox.
			if (!dojo.isMac) {
				if (dojo.isFF) {
					browserSpecificWidthSubtractor = 16;
				}
				else {
					browserSpecificWidthSubtractor = 10;
				}
			}
		}
		else if (dojo.isIE9) {
			// Handle IE9.
			browserSpecificWidthSubtractor = 17;
		}
		else if (dojo.isIE) {
			// Handle IE10 or greater.
			browserSpecificWidthSubtractor = 20;
		}

		// Open the left window.
		this._openInNewWindow(
			sortedResourceURIs[0],
			originY,
			originX,
			heightToUse,
			widthToUse - browserSpecificWidthSubtractor
		);
		// Open the right window.
		this._openInNewWindow(
			sortedResourceURIs[1],
			originY,
			originX + widthToUse,
			heightToUse,
			widthToUse - browserSpecificWidthSubtractor
		);
	},

	///////////////////////////////////////////////////////////////////////////
	// Helper Functions for Command Implementations

	_openInNewWindow: function(resourceURI, top, left, height, width) {
		var windowOpenFeatures =
			"toolbar=no, scrollbars=yes, resizable=yes,"
			+ " top=" + top + ","
			+ " left=" + left + ","
			+ " width=" + width + ","
			+ " height=" + height;
		var rowIds = {};
		rowIds[resourceURI] = true;
		this.executeOpen(rowIds, true, windowOpenFeatures);
	},

	_getHeightValueToUseForWindowOpen: function() {
		// summary:
		//		Get the ideal height value to use when launching a new browser
		//		window such that the resulting window's internal content area
		//		height will closely match the current browser window's internal
		//		content area height.

		var heightToUse = window.innerHeight;
		if (dojo.isIE) {
			heightToUse -= 3;
		}
		return heightToUse;
	},

	_getWidthValueToUseForWindowOpen: function() {
		// summary:
		//		Get the ideal width value to use when launching a new browser
		//		window such that the resulting window's internal content area
		//		width will closely match the current browser window's internal
		//		content area width.

		var widthToUse = window.innerWidth;
		if (dojo.isIE) {
			widthToUse -= 5;
		}
		return widthToUse;
	},

	_getRowIdsSortedByRowIndex: function(rowIds) {
		var sortedResourceURIs = [];
		if (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE) {
			var rowIdsArray = this._getUniqueKeysFromRowIds(rowIds);
			sortedResourceURIs = this.moduleStructure.arraySortOnBookorder(rowIdsArray);
		} else {
			var rowsOrder = this.grid.getRowsOrder(),
				numRowIds = this._getUniqueKeysFromRowIds(rowIds).length,
				resourceURI,
				i;
			for (i = 0; i < rowsOrder.length; i++) {
				resourceURI = rowsOrder[i].key;
				if (rowIds[resourceURI]) {
					sortedResourceURIs.push(resourceURI);
					if (sortedResourceURIs.length === numRowIds) {
						break;
					}
				}
			}
		}
		return sortedResourceURIs;
	},

	_getUniqueKeysFromRowIds: function (rowIds) {
		// summary:
		//		gets the unique keys from rowIds
		
		//Now get an array
		var keysArray = [];
		for(var item in rowIds) {
			keysArray.push(item);
		}
		
		return keysArray;
	},
	
	_getRecordInfosFromRowIds: function(rowIds) {
		var keys = this._getUniqueKeysFromRowIds(rowIds);
		return this._getRecordInfos(keys);
	},

	_getRecordInfos: function(keys) {
		var recordInfos = [];
		dojo.forEach(keys, dojo.hitch(this, function(key) {
			var ri = this._dataCache.getRecordInfo(key);
			if (ri) {
				recordInfos.push(ri);
			}
		}));

		return recordInfos;
	},
	
	containsCollectionRow: function(rowIds) {
		var collectionRowFound = false;
		// TODO Ask Art why RRMTypes doesn't have a constant that can be
		// referenced instead of hard-coding "Collection" in the following ...
		var collectionFormatURI = RRMTypes.getArtifactFormats(this._projectURI).Collection;
		for (var resourceURI in rowIds) {
			var resourceInfo = this._dataCache.getRecordInfo(resourceURI);
			var artifactFormat = resourceInfo.getEntityType();
			if (artifactFormat === collectionFormatURI) {
				collectionRowFound = true;
				break;
			}
		}
		return collectionRowFound;
	},
	
	containsModuleRow: function(rowIds) {
		var moduleRowFound = false;
		var moduleFormatURI = RRMTypes.getArtifactFormats(this._projectURI).Module;
		for (var resourceURI in rowIds) {
			var resourceInfo = this._dataCache.getRecordInfo(resourceURI);
			var artifactFormat = resourceInfo.getEntityType();
			if (artifactFormat === moduleFormatURI) {
				moduleRowFound = true;
				break;
			}
		}
		return moduleRowFound;
	},
	
	containsCalmRow: function(rowIds) {
		//iterates over rows to see if any rows are C/ALM rows
		for (var key in rowIds) {
			if (rowIds[key] === "calmRow") {
				return true;
			}
		}
		return false;
	},
	
	containsRelatedRow: function(rowIds) {
		// summary:
		//		iterates over rows to see if any rows are related rows, ie drilled into

		for (var key in rowIds) {
			if (rowIds[key] === "relatedRow") {
				return true;
			}
		}
		return false;
	},
	
	_defaultIsVisible: function(rowIds, count) {
		return true;
	},

	clearGridSelection: function() {
		if (this._nonClearingCommand) {
			return;
		}
		this.grid.clearSelection();
	},

	setNonClearingCommand: function() {
		this._nonClearingCommand = true;
		var self = this;
		setTimeout(function () {self._nonClearingCommand = false;}, 10);
	},
	
	
	///////////////////////////////////////////////////////////////////////////
	/// Operation Implementations
	
	
	openLinksExplorerText: function (num) {
		// This calculates the display text for the command.
		if (num === 1) {
			return this._messages["openDiagramView"];
		}
		return dojo.string.substitute(this._messages["openDiagramViewN"], [num]);
	},

	openLinksExplorer: function(rowIds) {
		var resourceURIs = [];

		for (var key in rowIds) {
			resourceURIs.push(key);
		}
		var resourceInfo = resourceInfo = this._dataCache.getRecordInfo(resourceURIs[0]),
			processAreaURI = resourceInfo.getProcessAreaURI() || this._projectURI;	
		try {
			jazz.core.loader
				.load_async("com.ibm.rdm.web.linksexplorer.LinksExplorerDialog", dojo.hitch(this, function() {
		        var constructor = dojo.getObject("com.ibm.rdm.web.linksexplorer.LinksExplorerDialog.createNewInstance");
		        if(!constructor) { throw "No constructor found"; }
		        var linksExplorerDialog = constructor({
					_resourceURIs: resourceURIs,
					_projectURI: this._projectURI,
					_baselineURI: this._baselineURI,
					viewer: {
						getProcessAreaURI: dojo.hitch(this, function() {
							return processAreaURI;
						}),
						getProjectURI: dojo.hitch(this, function() {
							return this._projectURI;
						}),
						projectSnapshotURI: this._baselineURI
					}
				});
		        linksExplorerDialog.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
		    }));
		} catch (e) {
		    console.error("Failed to load LinksExplorer : " + e.message);
		}
	},
	
	reportGenerateText: function(num) {
		// This calculates the display text for the command.
		var isModule = (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULES);
		if (num === 1) {
			return (isModule) ? this._messages.createReportModuleOne : this._messages.createReportOne;
		}
		var sub = (isModule) ? this._messages.createReportModuleN : this._messages.createReportN;
		return dojo.string.substitute(sub, [num]);
	},
	
	isReportGenerate: function(rowIds, count) {
		if (count > 0) {
			//This function decides if report creation is allowed.

			//since C/ALM rows do no contain the same methods, disable action.
			if (this.containsCalmRow(rowIds)) {
				return false;
			}
			//TODO: Add extra logic to determine if report should be generated
			var isReportingAllowed = true;
			/*
			for (var key in rowIds) {
				dojo.forEach(this.getArrayListOfResources(), function (item) {
					if (item.getResourceURI() === key) {
						//set isReportingAllowed
					}
				});
			}*/
			return isReportingAllowed;
		}
		return false;
	},
	
	doGenerateReport: function(rowIds) {
		//This command brings up the reporting wizard and populates it with the 
		//current selections			
		var callbackScope= this;			
		var getReportArtifacts = function(){
			var toAddToReport = callbackScope._getRecordInfosFromRowIds(rowIds);
			return toAddToReport;
		};
		
		//Determine the report context to generate a list of appropriate report types.
		var context = null;
		switch(this._viewType){
		case com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE:
		case com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULES:
			context = com.ibm.rdm.web.reporting.ReportWizard.viewType.MODULE;
			break;
		case com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.COLLECTION:
		case com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.COLLECTIONS:
			context = com.ibm.rdm.web.reporting.ReportWizard.viewType.COLLECTION;
			break;
		default:
			context = com.ibm.rdm.web.reporting.ReportWizard.viewType.VIEW;
			break;
		}
		
		var wizard = new com.ibm.rdm.web.reporting.ReportWizard({
			"projectURI": this._projectURI,
			"baselineURI": this._baselineURI,
			"getReportArtifactsMethod": getReportArtifacts,
			"context": context
		});
		wizard.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},
	
	selectManageSecurityText: function(num, rowIds) {
		if (num === 1) {
			return this._messages.manageSecurityOne;
		}
		return dojo.string.substitute(this._messages.manageSecurityN, [num]);
	},

	selectCreateReviewText: function(num, rowIds) {
		if (num === 1) {
			if (this.containsCollectionRow(rowIds)) {
				// Special case in which we are creating a review
				// from the selected artifact (a collection)
				return this._messages.createReviewFromCollection;
			}
			return this._messages.createReviewOne;
		}
		return dojo.string.substitute(this._messages.createReviewN, [num]);
	},

	selectAddToCollectionText: function(num) {
		if (num === 1) {
			return this._messages.addToCollectionOne;
		}
		return dojo.string.substitute(this._messages.addToCollectionN,[num]);
	},

	isManageSecurityEnabled: function(rowIds, count) {
		// summary:
		//		Reurn false, with the expectation that the "appropriate"
		//		setting will shortly be provided via a subsequent call to the
		//		getManageSecurityEnablementDeferred function (as we need to
		//		work with deferreds in order to really know whether the "Assign
		//		Team Ownership for Artifact" menu item should be enabled).
		return false;
	},
	
	_anyLockedByOther: function(rowIds) {
		// summary:
		//		Returns true if any of the URI's in rowIds are
		//		thought to be locked by another user.
		// returns:
		//		true if any locked
		for (var key in rowIds) {
			if (LockManager.isLockedByOther(key)) {
				 return true;
			}
		}
		
		return false;
	},

	getManageSecurityEnablementDeferred: function(rowIds, count) {
		// summary:
		//		Returns a deferred which will be called back with a single param
		//		representing whether the "Assign Team Ownership for Artifact"
		//		menu item should be enabled.
		// returns:
		//		A Deferred.
		var deferred = new dojo.Deferred();
		
		if (this._anyLockedByOther(rowIds)) {
			// WI: 73709, if any artifacts are locked by another user, this will operation will fail
			deferred.callback(false);
		}
		else if (count > 1 && !this._baselineURI) {
			deferred.callback(true);
		}
		else if (this._baselineURI){
			deferred.callback(false);
		}
		else {
			var artifactURIs = [];
			for (var key in rowIds) {
				artifactURIs.push(key);
			}
			var resourceInfo = this._dataCache.getRecordInfo(artifactURIs[0]);
			var processAreaURI =
				resourceInfo.getProcessAreaURI() || this._projectURI;
			var projectOperationsDeferred =
				ProcessOperationsCache.isSaveTeamSecuritySpecificationAllowed(
					processAreaURI, this._projectURI
				);
			projectOperationsDeferred.addCallback(
				this,
				function(isSaveTeamSecuritySpecificationAllowed){
					deferred.callback(isSaveTeamSecuritySpecificationAllowed);
				}
			);
		}
		return deferred;
	},

	isCreateReviewEnabled: function(rowIds, count) {
		// We're calling the getCreateReviewDeferred right after, so default to false
		return false;
	},

	getCreateReviewDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
		
			if (this.isUsedInModule(rowIds)) {
				deferred.callback(false);
			}
		
			var rowCount = 0,
				resourceURI = null;
		
			for(resourceURI in rowIds) {
				if(rowCount > 1){
					break;
				}
				rowCount++;
			}
		
			if (rowCount === 1 && this.containsCollectionRow(rowIds)) {
				// In this case, we will be creating a review from
				// the selected artifact, which is a collection.
				deferred.callback(true);
			} else {
		
				if (count <= 0) {
					deferred.callback(false);
				}
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}
				if (this.containsCollectionRow(rowIds)) {
					deferred.callback(false);
				}
				
				var isCreateReviewEnabled = (count > 0 &&
						reviewutils.isReviewSavingPermissionFoundInUserRole(this._projectOperations)  && 
																			!this.baselineURI && 
																			!this._isInHistoryMode && 
																			!this._isInReviewMode);
				deferred.callback(isCreateReviewEnabled);
			}
		}));
		
		return deferred;
	},

	isAddToCollectionEnabled: function(rowIds, count) {
		// We're using a deferred getAddToCollectionDeferred(), so just
		// return false until it sorts out the details
		return false;
	},

	getAddToCollectionDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
			if (this.isUsedInModule(rowIds) || this.containsCollectionRow(rowIds)) {
				deferred.callback(false);
			}
			else {
				var isAddToCollectionEnabled = (
					!this.containsCalmRow(rowIds) && // No CLM rows
					count > 0 &&// Something is actually selected
					// User has permissions to create artifacts
					this._projectOperations.item(ProjectCache.MODIFY_RESOURCE) &&
					!this._baselineURI &&
					!this._isInHistoryMode &&
					!this._isInReviewMode
				);

				deferred.callback(isAddToCollectionEnabled);
			}
		}));
		
		return deferred;
	},

	executeManageSecurity: function(rowIds) {
		// summary:
		//		Display a dialog in which the user is able to assign the process
		//		area ownship that should be used for the selected artifacts.
		var artifactURIs = [];
		for (var key in rowIds) {
			artifactURIs.push(key);
		}
		var artifactName;
		if (artifactURIs.length === 1) {
			var resourceInfo = this._dataCache.getRecordInfo(artifactURIs[0]);
			artifactName = resourceInfo.getTitle();
		}
		var projectDeferred = ProjectCache.getProject(this._projectURI);
		projectDeferred.addCallback(
			this,
			function(project) {
				// Note: Even though this artifact's processAreaURI is available
				// via "resourceInfo.getProcessAreaURI();", we're explicitly NOT
				// passing a processAreaURI field in the following constructor
				// call.  This forces the ManageSecurityForm to do a GET to
				// obtain the most up to date setting for this artifact's
				// process area assignment.  We do this to avoid potentially
				// using a stale assignment value (e.g. perhaps the user just
				// did a recursive folder team area assignment and one of the
				// resulting tweaked artifacts is the row that was just selected
				// for this artifact team area assignment operation, or perhaps
				// some other user has changed this artifact's team area
				// assignment _after_ this row was last refreshed in the results
				// area).  Please note that the GET that's performed by the
				// ManageSecurityForm will just make use of the GET results that
				// were previously obtained when the grid row's context menu was
				// opened.
				var form = new ManageSecurityForm({
					projectURI: this._projectURI,
					projectName: project.name,
					destroyContents: true,
					artifactURIs: artifactURIs,
					artifactName:  artifactName
				});
				form.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
			}
		);
	},

	executeCreateReview: function(rowIds) {
		var resourceURI = null;
		if (this.containsCollectionRow(rowIds)) {
			// We're creating a review from a collection.
			for(resourceURI in rowIds) {
				break;
			}
			
			ResourceManager.get({
				resourceURI: resourceURI
			}, dojo.hitch(this, function(collection){
				reviewutils.createNewReviewFromCollection(
					this._projectURI,
					collection,
					dojo.hitch(this, this._onLaunchedDialogHasClosed)
				);
				return collection;
			}));
		}
		else {
			// We're creating a review and adding rowIds to it.
			var resourceURIs = [];
			for (resourceURI in rowIds) {
				resourceURIs.push(resourceURI);
			}
			reviewutils.createNewReview(
				this._projectURI,
				resourceURIs,
				dojo.hitch(this, this._onLaunchedDialogHasClosed)
			);
		}
	},

	executeAddToCollection: function(rowIds) {
		// summary:
		//		Invoked when the "Add N Artifacts to Collection" menu
		//		item is selected from the context menu on the grid
		var resourceURIs = [],
			newArtifactForm = null;
		for (var resourceURI in rowIds) {
			resourceURIs.push(resourceURI);
		}

		var addToCollections = function(selectedCollections) {
			
			newArtifactForm.displayInfoMessage(this._messages.addingToCollection);
			var collectionTitles = null;

			dojo.forEach(selectedCollections, function(collectionInfo) {

				var collectionDeferred = ResourceManager.get({
					resourceURI: collectionInfo.resourceURI,
					skipCache: true
				});
				com.ibm.rdm.web.viewer.CollectionsMenu.removeResource(collectionInfo.resourceURI);
				
				collectionTitles = collectionTitles ? collectionTitles + ", " +collectionInfo.getTitle() : collectionInfo.getTitle();

				collectionDeferred.addCallback(this, function(collection){

					var currentURIs = collection.getArtifactURIs();
					var culen = currentURIs.length;
					// Gotta make sure we're not duping members, since
					// the server does not check for this...
					for (var x in resourceURIs) {
						if (dojo.indexOf(currentURIs, resourceURIs[x]) === -1) {
							currentURIs.push(resourceURIs[x]);
							collection.setArtifactURIs(currentURIs);
						}
					}

					// Only save if the URI list has grown
					if (collection.getArtifactURIs().length > culen) {
						
						// Save the new collection. We must refresh here
						// to make sure that subsequent operations on the
						// new collection from the grid context menu work
						// correctly.
						var saveDeferred = collection.save({
							refreshAfterSave: true
						});
						saveDeferred.addCallback(this, function(saveResponse){
							// Remove the collection from the cache if one exists.
							com.ibm.rdm.web.pages.ActionViewCache.removeCachedActionView("showArtifact", collection.resourceURI);

							newArtifactForm.displayOkMessage(this._messages.addedToCollection,
								[collectionTitles]);
						});
						saveDeferred.addErrback(this, function(errback){
							newArtifactForm.displayMessage(7338);
						});
					}
					else {
						// The artifacts were already members of the collection
						newArtifactForm.displayOkMessage(
							this._messages.addedToCollection,
							[collectionTitles]);
					}

					return collection; // in case we're chained
				});
				
				collectionDeferred.addErrback(this,function(errback){
					newArtifactForm.displayMessage(7338);
				});
			}, this); // dojo.forEach - "this" provides scope
		};

		var params = {
			projectURI: this._projectURI,
			selectionOptions: {
				selectArtifactFormat: ["Collection"],
				displayArtifactCreation: true,
				newArtifactFormat: RRMTypes.getArtifactFormats(this._projectURI).Collection
			},
			style: com.ibm.rdm.web.picker.ArtifactPickerForm.formStyle.ADD_CLOSE,
			filterArtifacts: dojo.clone(resourceURIs),
			description: this._messages.addToCollectionDialogDesc,
			title: this._messages.addToCollectionDialogTitle,
			onAddArtifacts: addToCollections,
			onAddCallbackScope: this,
			surrenderClose: false,
			defaultFolderForArtifactCreation: this._dashboardInterface.getSelectedFolder(),
			pauseOnAddAndClose: true,
			disablePickerModuleReuseMode: true
		};

		newArtifactForm = new com.ibm.rdm.web.picker.ArtifactPickerForm(params);

		newArtifactForm.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},
	
	executeDelete: function(rowIds) {
		//This command actually deletes the artifact.
		//Find the artifacts
		var toDelete = this._getRecordInfosFromRowIds(rowIds);
		var isInModule = (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE);

		var onLaunchedDialogHasClosed = dojo.hitch(this, this._onLaunchedDialogHasClosed);

		//Now we know which artifacts to delete - do it
		var deleteDeferred;
		if (toDelete.length > 1) {
			deleteDeferred = ResourceUtil.deleteMultipleResources(toDelete, isInModule, this.moduleStructure);
		} else {
			var isWrapper = (toDelete[0].getEntityType() === MimeTypes.getArtifactFormat(MimeTypes.WRAPPER_RESOURCE_NAME, this._projectURI));
			if (isWrapper && !isInModule) {
				deleteDeferred = ResourceUtil.deleteResource(
					toDelete[0].getResourceURI(), toDelete[0].getProjectURI(),
					toDelete[0].getDisplayTitle(), toDelete[0].getWrappedResourceURI(), isInModule,
					false, this.moduleStructure, onLaunchedDialogHasClosed
				);
			} else {
				deleteDeferred = ResourceUtil.deleteResource(
					toDelete[0].getResourceURI(), toDelete[0].getProjectURI(),
					toDelete[0].getDisplayTitle(), null, isInModule,
					toDelete[0].getEntityType() == com.ibm.rdm.web.model.types.RRMTypes.getArtifactFormats(this._projectURI)["Module"],
					this.moduleStructure, onLaunchedDialogHasClosed
				);
			}
		}
		deleteDeferred.promise.always(onLaunchedDialogHasClosed);
		return deleteDeferred;
	},

	executeRemove: function(rowIds, skipWarning) {
		//This command removes the artifact form the current contents (but does not delete it).
		//Find the artifacts
		var doRemove = dojo.hitch(this, function(rowIds){
			for (var key in rowIds) {
				this._dataCache.removeResource(key);
			}

			this._dashboardInterface.colRemovedCallback(rowIds);

		}, rowIds);
		
		if (skipWarning){
			doRemove();
		}
		else{
			var msgForm = new com.ibm.rdm.web.common.MessageForm({
				confirmMessage: this._messages.confirmRemove,
				dialogTitle: this._messages.removeOne,
				dialogStyle: "yesNo",
				onOk: doRemove
			});
			msgForm.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
		}
	},
	

	executeOpen: function(rowIds, /*Boolean?*/useWindowOpen, /*String?*/windowOpenFeatures) {
		// Opens an artifact in its own viewer.
		if (rowIds.length < 1) {
			return;
		}
		for ( var rowId in rowIds) {
			var resourceURI = rowId;
			var isOpenInHistory = !!this._baselineURI;
			
			var args = {
				"isRemoveFromCache": false,
				"isOpenInHistory": isOpenInHistory,
				"projectSnapshotURI": this._baselineURI,
				"useWindowOpen": useWindowOpen,
				"windowOpenFeatures": windowOpenFeatures
			};
			
			//open core artifact instead of modulefact if navigating from module view
			/*if (this.isUsedInModule(rowIds) && this._viewType !== com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE) {
				args.artifactInModule = this._dataCache.getRecordInfo(rowId).getModuleHostURI();
			}
			else if (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE && 
				this._dataCache.getRecordInfo(rowId) && 
				this._dataCache.getRecordInfo(rowId).getModuleBoundArtifact && 
				this._dataCache.getRecordInfo(rowId).getModuleBoundArtifact()) {
					resourceURI = this._dataCache.getRecordInfo(rowId).getModuleBoundArtifact();
			}*/
			
			ResourceLinksHandler.showInternalResource(resourceURI, args);
			break;
		}
	},
	

	moveText: function(num){
		// summary:
		//		Text that should be used for move menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			return this._messages.moveOne;
		}
		return dojo.string.substitute(this._messages.moveN, [num]);
	},

	showArtifactUrlText: function(num){
		// summary:
		//		Text that should be used for show artifact url option
		// num: int
		//		The number of rows that are selected in the grid
		if (num === 1){
			return this._messages.showUrlOne;
		}
		return dojo.string.substitute(this._messages.showUrlN, [num]);
	},
	
	isMoveEnabled: function(rowIds, count){
		// summary:
		//		Determine if the Move menu option should be enabled. It should
		//		only be enabled if the current user has the correct permission
		//		to move the selected resources.
		// rowIds: Object
		//		A map of identifier to rowtype for each of the selected rows.
		// count: int
		//		The count of the rows that have been selected.
		return false;
	},

	getMoveEnabledDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
			if (count > 0 &&
				!this._baselineURI
				) {
				//This function decides if move is allowed.
				//since C/ALM rows do not contain the same methods, disable action.
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}
				var projectOperations = this._projectOperations;
			
				var isMoveAllowedWithoutConsideringTeamOperations = projectOperations.item(ProjectCache.MODIFY_RESOURCE) && !this._isInHistoryMode && !this._isInReviewMode;
				deferred.callback(isMoveAllowedWithoutConsideringTeamOperations);

			} 
			else {
				deferred.callback(false);
			}
		}));
		
		return deferred;
	},

	executeMove: function(rowIds){
		// summary:
		//		Execute the Move on selected rows.
		// rowIds:
		//		A map of identifier to rowtype for each of the rows.

		var toMove = this._getRecordInfosFromRowIds(rowIds);
		
		var getFolder = function(folderURI) {
			var moveDeferred = null;
			if (toMove.length > 1) {
				moveDeferred = ResourceUtil.moveMultipleResources(toMove, folderURI);
			}
			else{
				ResourceUtil.moveToFolder(
					toMove[0].getProjectURI(),
					toMove[0].getResourceURI(),
					folderURI
				);
			}

			//Update the context to show a change has happend to make the cache refresh...
			com.ibm.rdm.web.common.UpdateContext.cachedArtifactHasBeenSaved();
			return moveDeferred;
		};

		//Let's just use the first one because all resources should
		//have the same value at this point.
		var descriptionDeferred = com.ibm.rdm.web.model.resources.ResourceManager.getResourceDescribe({
			resourceURI: toMove[0].getResourceURI(),
			skipCache: true
		});
		descriptionDeferred.addBoth(this, function(descriptionResponse){
			var picker =
				new MoveFolderPicker({
					projectURI: toMove[0].getProjectURI(),
					currentName: toMove[0].getTitle(),
					onGetMoveResult: dojo.hitch(this, getFolder),
					isMultiMove: (toMove.length > 1),
					artifactTypeURI: this._getArtifactTypeURI(toMove[0])
				});
			picker.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
		});

	},
	
	executeShowUrl: function(rowIds){
		var dialog;
		var links = [];
		var names = [];
		var info = [];
		var width = "450px";
		
		var recordInfos = this._getRecordInfosFromRowIds(rowIds);
		
		// Extract the artifact URIs and names from each artifact in our rows
		// Note: _getRecordInfosFromRowIds returns an array, but we're only feeding
		//       it one object at a time so we must always access the first element.
		for (var i = 0; i < recordInfos.length; ++i){
			var recordInfo = recordInfos[i];
			if (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE && recordInfo.inSnapshot){
				links.push(recordInfo.getResourceURI() + "?moduleBaseline=" + encodeURIComponent(recordInfo.getSnapshotURI()));
				width = "525px";
			}
			else if (recordInfo.inSnapshot){
			// We are in a baseline, but not a module.  Show the revision.
				links.push(recordInfo.revision);
			}
			else {
			// Otherwise a normal case, no baselines or modules. Just show the artifact.
				links.push(recordInfo.resourceURI);
			}
			names.push(dojo.string.substitute(this._messages.artifactDisplayName, [recordInfo.getID(), recordInfo.getTitle()]));
		}
		
		info.push(links);
		info.push(names);
		
		dialog = new PermalinkDialog({
			initDialogWidth: width,
			initInfoArray: info
		});
		dialog.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},
	
///// Lock and Unlock Commands

	lockText: function(num){
		// summary:
		//		Text that should be used for lock menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			return this._messages.lockOne;
		}
		return dojo.string.substitute(this._messages.lockN, [num]);
	},

	
	
	isLockVisible: function(rowIds, count){
		return this._isLockVisibleImpl(rowIds, count);
	},	

	_isLockVisibleImpl: function(rowIds, count){
		// summary:
		//		Determine if the Lock menu option is visible. It should
		//		only be visible we are not in edit mode, and all the resources can
		//		be locked.
		// rowIds: Object
		//		A map of identifier to rowtype for each of the selected rows.
		// count: int
		//		The count of the rows that have been selected.
		var visible = this._defaultIsVisible(rowIds, count);
		if (!visible) {
			return visible;
		}
		//Check we can lock these items
		var allLockable = true;
		
		for (var key in rowIds) {
			if (LockManager.isLockedByMe(key) || LockManager.isLockedByOther(key)) {
				allLockable = false;
				break;
			}
		}
		
		return allLockable;		
	},
	
	executeLock: function(rowIds){
		// summary:
		//		Execute the Lock on selected rows.
		// rowIds:
		//		A map of identifier to rowtype for each of the rows.

		var toLock = this._getUniqueKeysFromRowIds(rowIds);

		this._executeLockImpl(toLock, toLock);
	},
	
	_executeLockImpl: function(rowKeys, rowURIs){
		// summary:
		//		Execute the Lock on selected rows.
		// rowKeys:
		//		An array of row keys for the grid (this may or may not be the underlying URIs).
		// rowURIs:
		//		An array of corresponding URIs (for the actual artifacts, ie base artifacts in modules)

		var deferred;
		
		if (rowURIs.length > 1) {
			//Note these are explicit locks, so are preserved while 'implicit' edits occur.
			deferred = LockManager.lockResources(rowURIs, true);
			
			var refresh = [];
			dojo.forEach(rowKeys, function(resourceURI) {
				refresh.push({
					key: resourceURI,
					rowtype: com.ibm.rdm.web.artifactgrid.ArtifactConstants.RESOURCE_ROW
				});
			});
			
			deferred.addCallback(this, function (results) {
				var message = null;
				if (!results.allLocked) {
					if (results.allFailed) {
						//We got none - tell the user
						message = this._messages.lockAllFailed;
					} else {
						//Some failed, some didn't
						message = this._messages.lockSomeFailed;
					}
				}
				this.grid.refreshRows(refresh);
				
				if (message) {
					var msgForm = new com.ibm.rdm.web.common.MessageForm({
						confirmMessage: message,
						dialogTitle: this._messages.notifyLocks,
						dialogWidth: "350px",
						dialogStyle: "ok",
						okButtonText: this._messages.notifyLocksOK
						//onDialogOpen: dojo.hitch(this, function() {
						//	dojo.fadeIn({node: clearSuspicionForm.domNode, duration: 400}).play();
						//}),
					});
					msgForm.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
				}
			});
		} else {
			//Only one to lock
			var resourceURI = rowURIs[0];
			
			//Note this is an explicit lock, so is preserved while 'implicit' edits occur.
			deferred = LockManager.lockResource(resourceURI, true);
			
			var refresh = [];
			refresh.push({
				key: rowKeys[0],
				rowtype: com.ibm.rdm.web.artifactgrid.ArtifactConstants.RESOURCE_ROW
			});
			
			deferred.addCallback(this, function (lockStatus) {
			
				LockManager.showLockedMessage(LockManager.MSG_LOCK, lockStatus);

				this.grid.refreshRows(refresh);
			});
		}
	},
	
	////////// Unlock Command
	
	unlockText: function(num){
		// summary:
		//		Text that should be used for unlock menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			return this._messages.unlockOne;
		}
		return dojo.string.substitute(this._messages.unlockN, [num]);
	},
	
	isUnlockEnabled: function(rowIds, count){
		return this._isUnlockEnabledImpl(rowIds, count);
	},	

	_isUnlockEnabledImpl: function(rowIds, count) {
		// summary:
		//		Determine if the Unlock menu option should be enabled. It should
		//		only be enabled if the current user has the edit permission
		//		on the selected resources, and owns the lock for the selected
		//		resource(s).
		// rowIds: Object
		//		A map of identifier to rowtype for each of the selected rows.
		// count: int
		//		The count of the rows that have been selected.
		return false;
	},

	getUnlockEnabledDeferred: function(rowIds, count) {
		//	getReleaseLockEnablementDeferred() will call this function.
		//	When there are multiple artifacts selected and they are all locked, we will return true to enable the Unlock menu.
		//	If there is only one artifact selected and it is locked, if the lock is mine, it will enable the Unlock menu.
		//	If there is only one artifact selected and it is locked, if the lock is not mine, it will check the modify lock
		//	permission. If override permission is allowed, the Unlock menu will be enabled; otherwise, it will be disabled.
		//  If a positive response is returned (Unlock menu is enabled), we will check to see whether the current user has
		//	break lock permissions during the execute unlock, so we can decide whether to allow the override or display
		//	a warning.
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
			if (count > 0 &&
				!this._baselineURI
				) {
				//since C/ALM rows do not contain the same methods, disable action.
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}

				var projectOperations = this._projectOperations;
	 
				var isUnlockAllowedWithoutConsideringTeamOperations = projectOperations.item(ProjectCache.MODIFY_RESOURCE)
																	&& !this._isInHistoryMode 
																	&& !this._isInReviewMode;

				var modifyLockDeferred = new dojo.Deferred();
				var enableLock = true;
				if (count == 1) {
					var uris = this._getUniqueKeysFromRowIds(rowIds);
					dojo.forEach(uris, function(uri) {
						var mine = LockManager.isLockedByMe(uri);
						if (!mine) {
							// not my lock, check override permission
							var resourceInfo = this._dataCache.getRecordInfo(uri);
							
							var furtherProcess = dojo.hitch(this, function(resourceInfo, projectURI) {
								var resourceURI = resourceInfo.getResourceURI();
								var resourceProcessAreaURI = resourceInfo.getProcessAreaURI() || projectURI;

								var isReleaseLocksAllowedDeferred =
									ProcessOperationsCache.isReleaseLocksAllowed(
										resourceProcessAreaURI,
										projectURI
									);
								isReleaseLocksAllowedDeferred.addCallback(
									this,
									function(isReleaseLocksAllowed){
										modifyLockDeferred.callback(isReleaseLocksAllowed);
									}
								);
							});
							
							if (resourceInfo) {
								furtherProcess(resourceInfo, this._projectURI);
							} else {
								//Base artifact for module - so go get it
								ResourceManager.get(
									{resourceURI: uri}, 
									dojo.hitch(this, function(resourceInfo){
										furtherProcess(resourceInfo, this._projectURI);
										return resourceInfo;
									})
								);				
							}
						} else {
							modifyLockDeferred.callback(enableLock);
						}
					}, this);
				} else {
					// When displaying the Unlock menu, we chose to allow the unlock option if multiple locked artifacts are selected.
					// This is done for speed purposes.
					modifyLockDeferred.callback(enableLock);
				}
				
				modifyLockDeferred.then(dojo.hitch(this, function(isLockEnabled) {
					deferred.callback(isLockEnabled && isUnlockAllowedWithoutConsideringTeamOperations);	
				}));
			} else {
				deferred.callback(false);
			}
		}));
		
		return deferred;
	},	

	isUnlockVisible: function(rowIds, count){
		return this._isUnlockVisibleImpl(rowIds, count);
	},	
		
	_isUnlockVisibleImpl: function(rowIds, count){
		// summary:
		//		Determine if the Unlock menu option is visible. It should
		//		only be visible we are not in edit mode, and all the resources are
		//		currently locked.
		// rowIds: Object
		//		A map of identifier to rowtype for each of the selected rows.
		// count: int
		//		The count of the rows that have been selected.
		var visible = this._defaultIsVisible(rowIds, count);
		if (!visible) {
			return visible;
		}
		//Check the items are all locked
		var allLocked = true;
		for (var key in rowIds) {
			if (!(LockManager.isLockedByMe(key) || LockManager.isLockedByOther(key))) {
				allLocked = false;
				break;
			}
		}
		
		return allLocked;			
	},	

	getResourceInfo: function(artifactURI){
		var resourceInfo = this._dataCache.getRecordInfo(artifactURI);
		if (resourceInfo) {
			var deferred = new dojo.Deferred();
			deferred.callback(resourceInfo);
			return deferred;
		}
		//Base artifact for module - so go get it
		return ResourceManager.get({ resourceURI: artifactURI });				
	},
	
	executeUnlock: function(rowIds){
		// summary:
		//		Execute the Unlock on selected rows.
		// rowIds:
		//		A map of identifier to rowtype for each of the rows.

		var toUnlock = this._getUniqueKeysFromRowIds(rowIds);
		
		//This is an explicit release, so it can release an explicit lock.
		this._executeUnlockImpl(toUnlock, toUnlock);
	},
	
	
	_executeUnlockImpl: function(rowKeys, rowURIs){
		// summary:
		//		Execute the Unlock on selected rows.
		// rowKeys:
		//		An array of row keys for the grid (this may or may not be the underlying URIs).
		// rowURIs:
		//		An array of corresponding URIs (for the real artifacts, ie base artifacts in modules)

		// We need to check now to see if the team operations discovery
		// service says that the user should be able to release the lock
		// on the artifact(s).
		var toUnlock_rowKeys = [];
		var toUnlock_rowURIs = [];
		var artifactURIs_notMyLocks = [];
		var artifactKeys_notMyLocks = [];
		for (var i=0; i<rowURIs.length; i++) {
			if (!LockManager.isLockedByMe(rowURIs[i])) {
				artifactURIs_notMyLocks.push(rowURIs[i]);
				artifactKeys_notMyLocks.push(rowKeys[i]);
			} else {
				toUnlock_rowURIs.push(rowURIs[i]);
				toUnlock_rowKeys.push(rowKeys[i]);
			}
		}
		
		var continueUnlockDeferred = new dojo.Deferred();
		if (artifactURIs_notMyLocks.length > 0) {
			var resourceDeferreds=[];
			for (var i = 0; i < artifactURIs_notMyLocks.length; i++){
				resourceDeferreds.push(this.getResourceInfo(artifactURIs_notMyLocks[i]));
			}
			
			var resourceDeferredList = new dojo.DeferredList(resourceDeferreds);
			resourceDeferredList.then(dojo.hitch(this, function(resourceInfos) {
				var defList = [];
				var uniqueProcessAreas = {};
				dojo.forEach(resourceInfos, function(resourceInfo) {
					var resourceProcessAreaURI = resourceInfo[1].getProcessAreaURI() || this._projectURI;
					if (!uniqueProcessAreas[resourceProcessAreaURI]) {
						uniqueProcessAreas[resourceProcessAreaURI] = [];
					}
					uniqueProcessAreas[resourceProcessAreaURI].push(resourceInfo[1].getResourceURI());
				}, this);
				
				for (var processAreaURI in uniqueProcessAreas) {
					defList.push(ProcessOperationsCache.isReleaseLocksAllowed(
							processAreaURI, this._projectURI
					).then(dojo.hitch(this, function(isReleaseLocksAllowed) {
						return isReleaseLocksAllowed;
					})));			
				}
				
				var permissionsDeferredList = new dojo.DeferredList(defList);
				permissionsDeferredList.then(dojo.hitch(this, function (results) {
					var releaseLocksNotAllowed = 0;
					var i = 0;
					for (var processAreaURI in uniqueProcessAreas) {
						dojo.forEach(uniqueProcessAreas[processAreaURI], function(resource) {
							if (results[i][1]) {
								toUnlock_rowURIs.push(resource);
								var index = dojo.indexOf(artifactURIs_notMyLocks, resource);
								toUnlock_rowKeys.push(artifactKeys_notMyLocks[index]);
							} else {
								releaseLocksNotAllowed++;
							}
						});
						i++;
					}
					if (releaseLocksNotAllowed > 0) {
						if (releaseLocksNotAllowed == 1) {
							com.ibm.rdm.web.common.MessageHandler.showWarning(7738);
						} else {
							com.ibm.rdm.web.common.MessageHandler.showWarning(7737, [releaseLocksNotAllowed]);
						}
					}
					continueUnlockDeferred.callback();
				}));
			}));
		} else {
			continueUnlockDeferred.callback();
		}
		
		continueUnlockDeferred.then(dojo.hitch(this, function() {
			//This is an explicit release, so it can release an explicit lock.
			var deferred;
			var onDialogClose = dojo.hitch(this, this._onLaunchedDialogHasClosed);
			if (toUnlock_rowURIs.length === 1) {
				deferred = LockManager.releaseLockedResource(
					toUnlock_rowURIs[0], true, false, onDialogClose
				);
			} else {
				deferred = LockManager.releaseLockedResources(
					toUnlock_rowURIs, true, onDialogClose
				);
			}
			
			deferred.addBoth(this, function() {
				var refresh = [];
				dojo.forEach(toUnlock_rowKeys, function(key) {
					refresh.push({
						key: key,
						rowtype: com.ibm.rdm.web.artifactgrid.ArtifactConstants.RESOURCE_ROW
					});
				});
				this.grid.refreshRows(refresh);
			});
		}));
	},

///// End of Lock and Unlock Commands
	
	deleteText: function(num) {
		// This calculates the display text for the command.
		if (num === 1) {
			return this._messages.deleteOne;
		}
		return dojo.string.substitute(this._messages.deleteN, [num]);
	},

	isDeleteEnabled: function(rowIds, count) {
		return false;
	},

	_isDeleteAllowedWithoutConsideringTeamOperations: function(rowIds, count) {
		var isDeletionAllowed = false;
		// Since C/ALM rows do not contain the same methods, disable action.
		if (this.containsCalmRow(rowIds)) {
			isDeletionAllowed = false;
		} else {
			isDeletionAllowed = (
				this._projectOperations.item(ProjectCache.MODIFY_RESOURCE) &&
				!this.baselineURI &&
				!this._isInHistoryMode &&
				!this._isInReviewMode
			);
		}
		return isDeletionAllowed;
	},

	getDeleteEnablementDeferred: function(rowIds, count, /*Boolean?*/doCheckPermissionsForFolderInsteadOfArtifact) {
		// summary:
		//		Returns a deferred which will be called back with a single param
		//		representing whether the "Delete" menu item should be enabled.
		// returns:
		//		A Deferred.
		// doCheckPermissionsForFolderInsteadOfArtifact:
		//		By default the Delete permissions check will be for the process area
		//		associated with the selected artifact.  If you pass this as true,
		//		the Delete permissions check will instead be for the process area
		//		associated with the folder that contains the artifact.
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
		
		projectDeferred.then(dojo.hitch(this, function() {
			
			var isDeleteAllowedWithoutConsideringTeamOperations =
				this._isDeleteAllowedWithoutConsideringTeamOperations(rowIds, count);
			if (count > 1 && !this._baselineURI) {
				// Note: If there's more than one artifact selected no attempt is
				// made to proactively determine if the user should be able to
				// delete all the selected artifacts based on team operation
				// permissions (due to the fact that too many server calls would
				// need to be performed).
				if (this.isUsedInModule(rowIds)) {
					deferred.callback(false);
				}
				else {
					deferred.callback(isDeleteAllowedWithoutConsideringTeamOperations);
				}
			}
			else if (!isDeleteAllowedWithoutConsideringTeamOperations || this._baselineURI) {
				deferred.callback(false);
			}
			else {
				// There's a single row selected for deletion and we know that the
				// user is able to "in general" delete artifacts.  We need to check
				// now to see if the team operations discovery service says that
				// the user should be able to delete this artifact.
				var artifactURIs = [];
				for (var key in rowIds) {
					artifactURIs.push(key);
				}
				var resourceInfo = this._dataCache.getRecordInfo(artifactURIs[0]);
				var resourceURI = resourceInfo.getResourceURI();
				var resourceProcessAreaURI = resourceInfo.getProcessAreaURI() || this._projectURI;
				var artifactTypeURI = this._getArtifactTypeURI(resourceInfo);
				var projectURI = this._projectURI;
				var isDeletedAllowedDeferred =
					ProcessOperationsCache.isDeleteInstanceOfArtifactTypeAllowed(
						resourceURI,
						projectURI,
						artifactTypeURI,
						resourceProcessAreaURI,
						doCheckPermissionsForFolderInsteadOfArtifact,
						true //  Skip cache.
					);
				isDeletedAllowedDeferred.addCallback(
					this,
					function(isDeleteInstanceOfArtifactTypeAllowed){
						deferred.callback(isDeleteInstanceOfArtifactTypeAllowed);
					}
				);
			}
		}));
		return deferred;
	},

	getReleaseLockEnablementDeferred: function(rowIds, count) {
		// summary:
		//		Returns a deferred which will be called back with a single param
		//		representing whether the "Release Lock" menu item should be enabled.
		//		This is for ArtifactCommands, where the rowUri is the one of the artifact itself.
		// returns:
		//		A Deferred.

		return this._getReleaseLockEnablementDeferred(rowIds, count);
	},
	
	_getReleaseLockEnablementDeferred: function(rowIds, count) {
		// summary:
		//		Returns a deferred which will be called back with a single param
		//		representing whether the "Release Lock" menu item should be enabled.
		// returns:
		//		A Deferred.
		var deferred = new dojo.Deferred();
		var isReleaseAllowedDeferred;
		
		//Never allow taking locks in a baseline
		if (this._baselineURI) {
			deferred.callback(false);
		}
		
		// checks whether i can edit the artifacts locked, and whether they are locked by me...
		isReleaseAllowedDeferred = this.getUnlockEnabledDeferred(rowIds, count);
		
		isReleaseAllowedDeferred.then(dojo.hitch(this, function(isReleaseAllowedWithoutConsideringTeamOperations) {
			// if any of the locks are mine and team operations are not considered, Unlock will be enabled
			// we will check to see if user has release lock permissions during execute unlock to determine
			// whether to display a warning to the user or continue with unlock
			deferred.callback(isReleaseAllowedWithoutConsideringTeamOperations);
		}));
		
		return deferred;
	},
	
	copyText: function(num){
		// summary:
		//		Text that should be used for copy menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			return this._messages.copyOne;
		}
		return dojo.string.substitute(this._messages.copyN, [num]);
	},
	
	pasteText: function(num){
		// summary:
		//		Text that should be used for paste menu option
		// num: int
		//		The number of rows that are selected in the grid.
		return this._messages.pasteAsLink;
	},
	
	duplicateText: function(num){
		// summary:
		//		Text that should be used for duplicate menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			return this._messages.duplicateOne;
		}
		return dojo.string.substitute(this._messages.duplicateN, [num]);
	},
	
	createSimilarText: function(num, rowIds){
		// summary:
		//		Text that should be used for copy menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			if (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE) {
				return this._messages.createSimilarArtifact;
			}
			var artifactURI = null;
			for (var key in rowIds) {
				artifactURI = key;
			}
			var resource = this._dataCache.getRecordInfo(artifactURI);
			var objectTypes = resource.getTypes();
			var objectType = null;
			for (var type in objectTypes) {
				objectType = this._typeCache.getProjectObjectType(type);
				break;
			}
			return dojo.string.substitute(this._messages.createNewSimilar, [objectType.getTitle()]);
		}
		return this._messages.createNewSimilarArtifact;
	},

	isCopyEnabled: function(rowIds, count){
		// summary:
		//		Determine if the Copy menu option should be enabled. It should only be enabled if
		//		the current user has the correct permission to move the selected resources.
		// rowIds: int
		//		The identifier of each of the rows that are selected.
		// count: int
		//		The count of the rows that have been selected.
		return false;
	},
	
	isPasteAsLinkEnabled: function(rowIds, count) {
		// summary:
		//		Determine if the Paste as Link menu option should be enabled.
		// rowIds: int
		//		The identifier of each of the rows that are selected.
		// count: int
		//		The count of the rows that have been selected.
		var clipboardData = clipboard.get();
		if(!clipboardData) {
			// Nothing on clipboard...
			return false;
		}
		var resourceURIs = [];
		for (var key in rowIds) {
			resourceURIs.push(key);
		}
		if(!resourceURIs.length) {
			// No source...
			return false;
		}
		if((resourceURIs.length === 1 && clipboardData.uris.length === 1)
				&& resourceURIs[0] === clipboardData.uris[0]) {
			// Attempting to link to self...
			return false;
		}
		return true;
	},
	
	getCopyEnabledDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
			if (count > 0 && !this._baselineURI) {
				//This function decides if copy is allowed.
				//since C/ALM rows do no contain the same methods, disable action.
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}
				//If in a collection, but collection is not in edit mode, cannot create similar
				if (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.COLLECTION &&
						!this._dashboardInterface.isEditMode){
					deferred.callback(false);			
				}

				var projectOperations = this._projectOperations;
				var isCopyAllowed = (projectOperations.item(ProjectCache.MODIFY_RESOURCE) 
									&& !this._isInHistoryMode && !this._isInReviewMode);

				deferred.callback(isCopyAllowed);
			} else {
				deferred.callback(false);
			}
		}));
		
		return deferred;
	},
	
	isCreateSimilarEnabled: function(rowIds, count){
		// summary:
		//		Determine if the Create Similar menu option should be enabled. It should only be enabled if
		//		the current user has the correct permission to move the selected resources. Similar to copy, except we allow it if
		//		there are other objects under edit.
		// rowIds: int
		//		The identifier of each of the rows that are selected.
		// rowType: int
		//		The type of the rows that have been selected.
		// count: int
		//		The count of the rows that have been selected.
		return false;
	},
	
	getCreateSimilarEnabled: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
			if (count === 1 && !this._baselineURI) {
				//This function decides if move is allowed.
				//since C/ALM rows do no contain the same methods, disable action.
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}
				//Disable this for any related (ie drilled into) rows
				if (this.containsRelatedRow(rowIds)) {
					deferred.callback(false);
				}
				//If in a collection, but collection is not in edit mode, cannot create similar
				if (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.COLLECTION &&
						!this._dashboardInterface.isEditMode()){
					deferred.callback(false);			
				}

				var projectOperations = this._projectOperations;
				var isCopyAllowed = projectOperations && (projectOperations.item(ProjectCache.MODIFY_RESOURCE)
									&& !this._isInHistoryMode && !this._isInReviewMode);

				deferred.callback(isCopyAllowed);
			} else {
				deferred.callback(false);
			}
		}));
		
		return deferred;
	},
	
	executeCreateSimilar: function(rowIds, where){
		var statusMessageHandle = StatusDisplay.showStatusMessage(this._messages.creatingArtifact, true);
		var resources = {};
		var resourceURIs = [];
		var count = 0;
		var selectedFolderURI = this._dashboardInterface.getSelectedFolder();
		for (var key in rowIds) {
			var item = this._dataCache.getResourceByUri(key);
			item.parentResourceURI = selectedFolderURI;
			resources[key] = item;
			//Extract parentResource and tags information from existing item in memory where possible
			if (!item.parentResourceURI && !item.__tags){
				resourceURIs.push(key);
			}
			count++;
		}
		
		var descriptionsDeferred = ProjectCache.getBatchResourceDescriptionsMultiFetch(resourceURIs, null, true, null, true);
		descriptionsDeferred.addBoth(this, function(descriptionResponse){
			for (var resourceURI in descriptionResponse) {
					resources[resourceURI].parentResourceURI = descriptionResponse[resourceURI].parentResourceURI;
					resources[resourceURI].__tags = descriptionResponse[resourceURI].tags;
			}
			// returns a deferred, so that descriptionsDeferred blocks until it is resolved.
			var deferred = this._createSimilarArtifacts(resources, where, count, statusMessageHandle);
			StatusDisplay.hideStatusMessage(statusMessageHandle);
			return deferred;
		});
		return descriptionsDeferred;
	},
	
	//TODO abstract this function
	_createSimilarArtifacts: function(resources, where, numResources, statusMessageHandle){
		var deferredsArray = [];
		for (var resourceURI in resources) {
			var resource = resources[resourceURI];
			var projectURI = this._projectURI;
			var folderURI = resource.getParentResourceURI();
			var description = "";
			var artifactObjectTypes = resource.getTypes();
			var typeURI = null;
			for (var type in artifactObjectTypes) {
				typeURI = type;
				break;
			}
			var title = this._messages["new"] + this._typeCache.getProjectObjectType(typeURI).getTitle();
			var contentType = resource.getMimeType();

			var createResourceArgs = {
				format: contentType,
				projectURI: projectURI,
				resourceUriScope: this.resourceURI,
				folderURI: folderURI,
				typeURI: typeURI,
				title: title,
				description: description,
				enableEditInGrid: true
			};
			var onCreateArtifactFormCancelClicked = dojo.hitch(this, function(){
				StatusDisplay.hideStatusMessage(statusMessageHandle);
			});
			var deferred = (
				numResources > 1 ?
				ResourceUtil.createResource(createResourceArgs) :
				this._createResourceAfterPerformingPermissionCheck(
					createResourceArgs, onCreateArtifactFormCancelClicked)
			);
			deferredsArray.push(deferred);
			deferred.then(
				dojo.hitch(this, function(tags, responseResource) {
					responseResource.similarResource = resource;
					if (tags && tags.length > 0){
						var tagsDeferred = TagUtil.setTagsForArtifact([], tags, responseResource.getResourceURI(), this._projectURI, false);
						responseResource.__createSimilarApplyTagsDeferred = tagsDeferred;
					}
					//Set appropriate tags information and parent as handed in from origin
					responseResource.__tags = tags;
					responseResource.parentResourceURI = resource.parentResourceURI;
				}, resource.__tags)	
			);
		}
		var deferredsList = new dojo.DeferredList(deferredsArray);
		deferredsList.promise.always(
			dojo.hitch(this, function(response) {
				StatusDisplay.hideStatusMessage(statusMessageHandle);
			})
		);
		return deferredsList;
	},

	_createResourceAfterPerformingPermissionCheck: function(
		createResourceArgs, onCreateArtifactFormCancelClicked)
	{
		// summary:
		//		Perform a permission check to see if the user should be able to
		//		create an artifact of the specified type in the specified
		//		folder.  If he/she should be allowed to create such an artifact
		//		we make a call to ResourceUtil.createResource.  If he/she should
		//		not be able to create such an artifact we invoke the Create
		//		Artifact dialog, show a relevant error about the permission
		//		issue, and allow for the user to proceed with artifact creation
		//		via this dialog.
		// returns:
		//		This function returns a deferred that contains the creation
		//		information returned by either ResourceUtil.createResource or
		//		NewArtifactForm's onCreate callback.
		// createResourceArgs:
		//		A object representing the single argument that should be passed
		//		in a call to the ResourceUtil.createResource function.
		// onCreateArtifactFormCancelClicked:
		//		If it turns out that a permission issue was found and the Create
		//		Artifact dialog was shown to have the user resolve the issue,
		//		it MAY be the case that the user desides to just Cancel out of
		//		the dialog.  This callback is used to inform the calling function
		//		when this occurs.
		var deferred = new dojo.Deferred();
		var projectURI = createResourceArgs.projectURI;
		var artifactTypeURI = createResourceArgs.typeURI;
		var artifactFormatURI = createResourceArgs.format;
		var folderURI = createResourceArgs.folderURI;
		var moduleContextURI = createResourceArgs.moduleContextURI;
		var canCreateArtifactDeferred =
			ProcessOperationsCache.isCreateInstanceOfArtifactTypeAllowed(
				projectURI, folderURI, artifactTypeURI
			);
		canCreateArtifactDeferred.addCallback(this, function(isCreateInstanceOfArtifactTypeAllowed){
			if (isCreateInstanceOfArtifactTypeAllowed) {
				// We know that the user is authorized to create an artifact
				// of the specified type in the specified folder.  Create the
				// artifact and invoke the callback that's returned by this
				// private function.
				var createResourceDeferred = ResourceUtil.createResource(createResourceArgs);
				createResourceDeferred.then(
					dojo.hitch(this, function(response) {
						var processArea = this._dashboardInterface.getModuleProcessAreaURI();
						if (dojo.isFunction(response.isStub) && response.isStub() && !response.getProcessAreaURI() && processArea) {
							// We have a stub without a processArea... and we have a process area
							response.setProcessAreaURI(processArea);
						}
						deferred.callback(response);
						return response;
					})
				);
			} else {
				// We've determined that the user is not authorized to create
				// an artifact of the specified type in the specified folder.
				// Show the Create Artifact dialog (which will alert the user
				// to this issue and will also enable the user to specify a
				// different folder or artifact type to get around the issue).
				var newArtifactDialog = new NewArtifactForm({
					artifactFormat: artifactFormatURI,
					typeURI: artifactTypeURI,
					projectURI: projectURI,
					resourceURI: this.resourceURI,
					folderURI: folderURI,
					moduleContextURI: moduleContextURI,
					onCancel: onCreateArtifactFormCancelClicked,
					onCreate: dojo.hitch(this, function(response){
						// The user has just successfully created an artifact
						// via the Create Artifact dialog.  Invoke the callback
						// that's returned by this function.
						deferred.callback(response);
						return response;
					})
				});
				newArtifactDialog.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
			}
			return isCreateInstanceOfArtifactTypeAllowed;
		});
		return deferred;
	},
	
	anyModuleTypes: function(descriptionResponses) {
		var key = null,
			entry;
		for (key in descriptionResponses) {
			entry = descriptionResponses[key];
			if (ResourceUtil.isModule(entry.contentType)) {
				return true;
			}
		}
		
		return false;
	},
	
	executeCopy: function(rowIds){
		// summary:
		//		Execute Copy on selected rows.
		// rowIds: collection of URIs -> rowType
		//		The identifiers for each of the rows.
		var copyURIs = this._getUniqueKeysFromRowIds(rowIds);
		var copyData = {
			action: COPY_ACTION,
			uris: copyURIs,
			projectURI: this._projectURI,
			count: copyURIs.length
		};
		// set the clipboard contents
		clipboard.put(copyData);
	},
	
	executePasteAsLink: function(rowIds) {
		// summary:
		//		Execute Copy on selected rows.
		// rowIds: collection of URIs -> rowType
		//		The identifiers for each of the rows.
		var clipboardData = clipboard.get(),
			targetURIs = [],
			sourceURIs = clipboardData.uris;
		for (var key in rowIds) {
			targetURIs.push(key);
		}
		var handleLinkCreated = function(linkTypeURI, createdLink, flipLinks) {
			this.executePostLinkEvent(linkTypeURI, createdLink, flipLinks);
			StatusDisplay.hideStatusMessage();
		};
		
		var dialog = new com.ibm.rdm.web.links.SimpleLinkDialog({
			sourceURIs: sourceURIs,
			targetURIs: targetURIs,
			isModule: (this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE),
			isCrossProject: (clipboardData.projectURI !== this._projectURI),
			isCrossWindow: (clipboardData.window && clipboardData.window !== com.ibm.rdm.web.windowID),
			projectURI: this._projectURI,
			processAreaURI: this._projectURI,
			callBack: dojo.hitch(this, handleLinkCreated)
		});
		dialog.show();
	},
	
	executeDuplicate: function(rowIds){
		// summary:
		//		Execute Duplicate on selected rows.
		// rowIds: collection of URIs -> rowType
		//		The identifiers for each of the rows.

		var copyURIs = this._getUniqueKeysFromRowIds(rowIds);
		var toCopy = this._getRecordInfos(copyURIs),
			anyModules = false;

		var descriptionDeferred,
			picker = null;
		// TODO: Refactor this to use ResourceUtil.copyResourcesToFolder
		if (toCopy.length > 1) {
			descriptionDeferred = ProjectCache.getBatchResourceDescriptionsMultiFetch(copyURIs, null, true);
			descriptionDeferred.addBoth(this, function(descriptionResponses){
				if (this.anyModuleTypes(descriptionResponses)) {
					anyModules = true;
				}
				var copyArtifacts = dojo.hitch(this, function(/*folderURI*/folderURI, /*String*/copyName, /*Boolean*/doDuplicateLinks, moduleCopyPolicy, createLinkType){
					var copyCompleteDeferred = new dojo.Deferred();
					this._selectedFolderURI=folderURI;
					var names = [];
					for (var i = 0; i<toCopy.length; i++) {
						names.push(toCopy[i].getTitle());
					}
					
					var newNamesDeferred = ResourceUtil.makeUniqueNamesAtFolder(toCopy[0].getProjectURI(), folderURI, names);
					newNamesDeferred.addBoth(this, function(newNames){
						var copyOperations = [];
						dojo.forEach(toCopy, dojo.hitch(this, function(item, count) {
							var contentType = null;
							var objectType = null;
							var wrappedResourceURI = null;
	
							if (item.getEntityType() === MimeTypes.getArtifactFormat(MimeTypes.WRAPPER_RESOURCE_NAME, item.getProjectURI())) {
								contentType = MimeTypes.getArtifactFormat(MimeTypes.WRAPPER_RESOURCE_NAME, item.getProjectURI());
								wrappedResourceURI = descriptionResponses[item.getResourceURI()].wrappedResourceURI;
								var objectTypes = item.getTypes();
								for (var type in objectTypes) {
									objectType = type;
									break;
								}
							}
							else {
								contentType = item.getEntityType();
							}
	
							copyOperations.push({
								resourceURI: item.getResourceURI(),
								projectSnapshotURI: item.getSnapshotURI(),
								projectURI: item.getProjectURI(),
								fromArtifactName: item.getTitle(),
								fromArtifactIdentifier: item.getID(),
								copiedName: newNames[count],
								folderURI: folderURI,
								revision: item.getRevision(),
								reviewURI: null,
								contentType: contentType,
								wrappedResourceURI: wrappedResourceURI,
								objectType: objectType,
								resourceUriScope: this.resourceURI,
								delayCallbackUntilDataIndexed: false,
								doDuplicateLinks: doDuplicateLinks,
								moduleCopyPolicy: moduleCopyPolicy,
								createLinkType: createLinkType
							});
						}));
						
						ResourceUtil.copyMultipleToFolder(
							copyOperations,
							dojo.hitch(picker, picker.isCancelled),
							dojo.hitch(picker, picker.updateCopyProgress)
						).then(
							function(response){
								copyCompleteDeferred.callback(response);
							},
							function(response){
								copyCompleteDeferred.errback(response);
							}
						);
					});
					return copyCompleteDeferred;
				});
				picker = new CopyFolderPicker({
					projectURI: toCopy[0].getProjectURI(),
					currentName: toCopy[0].getTitle(),
					isMultiCopy: true,
					numArtifactsToCopy: copyURIs.length,
					onGetCopyResult: dojo.hitch(this, copyArtifacts),
					selectedFolderURI: this._selectedFolderURI,
					currentFolderURI: descriptionResponses[copyURIs[0]].parentResourceURI,
					includesModules: anyModules,
					autoCloseIfSelectedFolderMatchesFolderURI: dojo.hitch(this, function(){return this._dashboardInterface.getSelectedFolder()})
				});
				picker.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
			});
		}
		else {
			
			// just the one artifact to copy.
			// if it's a module we are copying, need to to a ResourceManager.get so that we can 
			descriptionDeferred = com.ibm.rdm.web.model.resources.ResourceManager.getResourceDescribe({
				resourceURI: toCopy[0].getResourceURI()
			});
			
			descriptionDeferred.addBoth(this, function(descriptionResponse){
				
				var copyArtifact = function(/*folderURI*/folderURI, /*String*/copyName, /*Boolean*/doDuplicateLinks, moduleCopyPolicy, createLinkType)
				{
					this._selectedFolderURI=folderURI;
					var contentType = null;
					var objectType = null;
					var wrappedResourceURI = null;

					if (toCopy[0].getEntityType() === MimeTypes.getArtifactFormat(MimeTypes.WRAPPER_RESOURCE_NAME, toCopy[0].getProjectURI())) {
						contentType = MimeTypes.getArtifactFormat(MimeTypes.WRAPPER_RESOURCE_NAME, toCopy[0].getProjectURI());
						wrappedResourceURI = descriptionResponse.wrappedResourceURI;
						var objectTypes = toCopy[0].getTypes();
						for (var type in objectTypes) {
							objectType = type;
							break;
						}
					} else {
						contentType = toCopy[0].getEntityType();
					}

					// if the copy is of a module (the dialog knows this), need to get the copy policy
					// and pass through a flag, ResourceUtil will need to do a "GET" and a "PUT" on the
					// module.
					return ResourceUtil.copyToFolder({
						resourceURI: toCopy[0].getResourceURI(),
						projectSnapshotURI: toCopy[0].getSnapshotURI(),
						projectURI: toCopy[0].getProjectURI(),
						fromArtifactName: toCopy[0].getTitle(),
						fromArtifactIdentifier: toCopy[0].getID(),
						copiedName: copyName,
						folderURI: folderURI,
						revision: toCopy[0].getRevision(),
						reviewURI: null,
						contentType: contentType,
						wrappedResourceURI: wrappedResourceURI,
						objectType: objectType,
						resourceUriScope: toCopy[0].getResourceURI(),
						delayCallbackUntilDataIndexed: false,
						doDuplicateLinks: doDuplicateLinks,
						moduleCopyPolicy: moduleCopyPolicy,
						createLinkType: createLinkType
					});
				};
				var artifactTypeURI = this._getArtifactTypeURI(toCopy[0]);
				var picker = new CopyFolderPicker({
					projectURI: toCopy[0].getProjectURI(),
					currentName: toCopy[0].getTitle(),
					onGetCopyResult: dojo.hitch(this, copyArtifact),
					entityType: toCopy[0].getEntityType(),
					artifactTypeURI: artifactTypeURI,
					selectedFolderURI: this._selectedFolderURI,
					currentFolderURI: descriptionResponse.parentResourceURI,
					autoCloseIfSelectedFolderMatchesFolderURI: dojo.hitch(this, function(){return this._dashboardInterface.getSelectedFolder()}),
					copiedResourceProcessAreaURI: toCopy[0].getProcessAreaURI()
				});
				picker.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
			});
		}
	},

	exportText: function(num){
		// summary:
		//		Text that should be used for export menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			return this._messages.exportOne;
		}
		return dojo.string.substitute(this._messages.exportN, [num]);
	},

	isExportEnabled: function(rowIds, count){
		// summary:
		//		Determine if the Export menu option should be enabled. It should only be enabled if
		//		the current user has the correct permission to move the selected resources.
		// rowIds: int
		//		The identifier of each of the rows that are selected.
		// count: int
		//		The count of the rows that have been selected.
		return false;
	},

	getExportEnabledDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
			if (count > 0) {
				//This function decides if move is allowed.
				//since C/ALM rows do no contain the same methods, disable action.
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}
				//Disable this for any related (ie drilled into) rows
				if (this.containsRelatedRow(rowIds)) {
					deferred.callback(false);
				}
			
				var projectOperations = this._projectOperations;
				var isExportAllowed = projectOperations && projectOperations.item(ProjectCache.MODIFY_RESOURCE);

				deferred.callback(isExportAllowed);
			} else {
				deferred.callback(false);
			}
		}));
		
		return deferred;
	},

	executeExport: function(rowIds){
		// summary:
		//		Execute export on selected rows.
		// rowIds: int
		//		A map of identifiers for each of the rows to their rowtype.
		var toExport = [];
		var objectTypes = [],
			key = null;

		if (this._dataCache.isSecondaryCacheEnabled()) {
			for (key in rowIds) {
				var item = this._dataCache.getResourceByUri(key);
				toExport.push(item.getResourceURI());
				var types = item.getTypes();
				var itemTypes = [];
				for (var type in types) {
					itemTypes.push(type);
				}
				if (dojo.indexOf(objectTypes, itemTypes[0]) === -1) {
					objectTypes.push(itemTypes[0]);
				}
			}	
		} else {
			for (key in rowIds) {
				dojo.forEach(this._dataCache.getArrayListOfResources(), function(item){
					if (item.getResourceURI() === key) {
						toExport.push(item.getResourceURI());
						var types = item.getTypes();
						var itemTypes = [];
						for (var type in types) {
							itemTypes.push(type);
						}
						if (dojo.indexOf(objectTypes, itemTypes[0]) === -1) {
							objectTypes.push(itemTypes[0]);
						}
					}
				});
			}
		}

		var exportRequestParams = this._dashboardInterface.getCurrentViewServiceRequestParams(true);
		exportRequestParams.artifactUris = toExport;
		
		var exportView = ViewServiceClient.composeViewsPostData(exportRequestParams, {gridColumns: this._dashboardInterface.getCurrentConfiguration()});
		
		var exportDeferred = CsvExport.exportXml(
			dojox.xml.parser.parse(exportView),
			this._projectURI,
			(this._viewType === com.ibm.rdm.web.project.DashboardConstants.VIEW_TYPES.MODULE),
			this._baselineURI
		);
		// Note: exportDeferred is actually a Promise (not Deferred) so we can
		// safely use 'always' on it.
		exportDeferred.always(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},

	selectTagsText: function(num) {
		if (num === 1) {
			return this._messages.applyTagsOne;
		}
		return dojo.string.substitute(this._messages.applyTagsN, [num]);
	},

	isSelectTagsEnabled: function(rowIds, count) {
		return false;
	},

	getSelectTagsDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
	
		projectDeferred.then(dojo.hitch(this, function() {
			if (count > 0 && !this._baselineURI) {//This function decides if apply tag is allowed.
				//since C/ALM rows do no contain the same methods, disable action.
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}
			
				var projectOperations = this._projectOperations;
				if (projectOperations) {	
					var isModifyPublicTag = projectOperations.item(com.ibm.rdm.web.common.ProjectCache.MODIFY_PUBLIC_TAG);
					var isModifyPrivateTag = projectOperations.item(com.ibm.rdm.web.common.ProjectCache.MODIFY_PRIVATE_TAG);
					deferred.callback(((isModifyPublicTag || isModifyPrivateTag) && !this._isInHistoryMode && !this._isInReviewMode));
				}
				else {
					deferred.callback(false);
				}
			} else {
				deferred.callback(false);
			}
		}));
		
		return deferred;
	},

	executeSelectTags: function(rowIds) {
		var toApplyTagResourceURIs = this._getUniqueKeysFromRowIds(rowIds);

		// If only one item is selected, allow tags to be set as if the user
		// were specifying tags in the resource viewer.
		if (toApplyTagResourceURIs.length === 1) {
			var toApplyTagResourceInfos = this._getRecordInfos(toApplyTagResourceURIs);
			this.doSelectTags(toApplyTagResourceURIs[0], toApplyTagResourceInfos[0]);
		} else {
			this.doApplyTags(toApplyTagResourceURIs);
		}
	},

	_getModifyArtifactPermissions: function(resourceInfo) {
		// summary:
		//		Returns a deferred that, when fired, will pass a single argument
		//		representing an instance of a ModifyArtifactPermissions object.
		// returns:
		//		A Deferred.
		var deferred = new dojo.Deferred();
		var processAreaURI =
			resourceInfo.getProcessAreaURI() || this._projectURI;
		var artifactTypeURI = this._getArtifactTypeURI(resourceInfo);
		TypesManager.getObjectType({
			typeURI: artifactTypeURI,
			projectURI: this._projectURI
		}).then(
			dojo.hitch(this, function(objectType){
				var artifactTypeActionId = objectType.getActionId();
				ProcessOperationsCache.getModifyArtifactPermissions(
					artifactTypeActionId, processAreaURI, this._projectURI
				).then(
					dojo.hitch(this, function(modifyArtifactPermissions){
						deferred.callback(modifyArtifactPermissions);
					})
				);
			})
		);
		return deferred;
	},

	doSelectTags: function(resourceURI, resourceInfo) {
		// summary:
		//		Present a dialog to enable editing the tags for one artifact.
		var form = new com.ibm.rdm.web.tags.EditTagsForArtifact({
			projectURI: this._projectURI,
			resourceURIs: [resourceURI],
			type: com.ibm.rdm.web.tags.selectTags,
			getModifyArtifactPermissions: dojo.hitch(
				this, this._getModifyArtifactPermissions, resourceInfo
			)
		});
		var self = this;
		form.connectHandler(form._okButtonNode, "onclick", function(event) {
			var savedTagURIs = form.savedTagURIs;
			var startingSelectedTagURIs = form.startingSelectedTagURIs;
			var tagDeferred = TagUtil.setTagsForArtifact(startingSelectedTagURIs, savedTagURIs, resourceURI, self._projectURI, true);
			if (tagDeferred) {
				StatusDisplay.showStatusMessage(this._messages.applyingTags);
				tagDeferred.addCallback(self, function(response){
					StatusDisplay.hideStatusMessage();
					com.ibm.rdm.web.common.UpdateContext.cachedArtifactHasBeenSaved();
					ChangeNotifier.notifyListeners({
						eventTopic: com.ibm.rdm.web.notifications.EventTopics.TAG_ARTIFACTS_UPDATED,
						response: response,
						resourceURIs: [resourceURI],
						projectURI : self._projectURI
					});
				});
				
				tagDeferred.addErrback(self, function(response){
					StatusDisplay.hideStatusMessage();
					return response;
				});
			}
		});
		form.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},

	_onLaunchedDialogHasClosed: function() {
		// summary:
		//		A launched dialog has been closed (via either clicking on one of
		//		its buttons or perhaps by clicking on the Escape key).  Put focus
		//		back in the grid so that the user can continue navigating via
		//		the keyboard.
		this.grid.retakeFocus();
	},

	doApplyTags: function(toApplyTagResourceURIs) {
		// summary:
		//		Apply tags to multiple artifacts.
		var self = this;
		var form;
		// i am not sure how this works, but "this" is actually correct when this
		// function is invoked...
		var applyTags = function() {
			var savedTagURIs = form.savedTagURIs;
			var deferredCompleted = 0;
			if (savedTagURIs.length !== 0) {
				StatusDisplay.showStatusMessage(this._messages.applyingTags);
				for (var i = 0; i < toApplyTagResourceURIs.length; i++) {
					var emptyArr = [];
					var tagDeferred = TagUtil.setTagsForArtifact(emptyArr, savedTagURIs, toApplyTagResourceURIs[i], self._projectURI);
					tagDeferred.addCallback(self, function(response){
						StatusDisplay.hideStatusMessage();
						deferredCompleted++;
						if (deferredCompleted === toApplyTagResourceURIs.length) {
							com.ibm.rdm.web.common.UpdateContext.cachedArtifactHasBeenSaved();
							ChangeNotifier.notifyListeners({
									eventTopic: com.ibm.rdm.web.notifications.EventTopics.TAG_ARTIFACTS_UPDATED,
									response: response,
									resourceURIs: toApplyTagResourceURIs,
									projectURI : self._projectURI
								}
							);
						}
						return response;
					});
					
					tagDeferred.addErrback(self, function(response){
						StatusDisplay.hideStatusMessage();
						return response;
					});
				}
			}
		};

		form = new com.ibm.rdm.web.tags.EditTagsForArtifact({
			projectURI: this._projectURI,
			resourceURIs: toApplyTagResourceURIs,
			type: com.ibm.rdm.web.tags.applyTags,
			onOk: applyTags
		});

		form.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},
	
	removeText: function(num) {
		// This calculates the display text for the command.
		if (num === 1) {
			return this._messages.removeOne;
		}
		return dojo.string.substitute(this._messages.removeN, [num]);
	},

	isRemoveEnabled: function(rowIds, count) {
		// We need to do a permissions check to make sure the user should
		// really be able to perform a remove.
		// Although we return false, a subsequent call to the
		// getRemoveEnablementDeferred function will result in appropriately
		// enabling/disabling the menu item.
		return false;
	},

	getRemoveEnablementDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
		
		projectDeferred.then(dojo.hitch(this, function() {
			var isRemoveAllowedWithoutConsideringTeamOperations =
				this._isRemoveAllowedWithoutConsideringTeamOperations(rowIds, count);
			if (!isRemoveAllowedWithoutConsideringTeamOperations) {
				deferred.callback(false);
			} else {
				var getModifyArtifactPermissions =
					this._dashboardInterface.getModifyArtifactPermissions();
				getModifyArtifactPermissions.addCallback(
					this,
					function(modifyArtifactPermissions){
						deferred.callback(
							modifyArtifactPermissions.isModifyContentAllowed()
						);
					}
				);
			}
		}));
		return deferred;
	},

	_isRemoveAllowedWithoutConsideringTeamOperations: function(rowIds, count) {
		var projectOperations = this._projectOperations;
		var isRemoveAllowed = (
				projectOperations.item(ProjectCache.MODIFY_RESOURCE) &&
				!this._baselineURI &&
				!this._isInHistoryMode &&
				!this._isInReviewMode
			);
		if (count > 0 &&
			!com.ibm.rdm.web.project.EditManager.isAnyUnderEdit(this._projectURI) &&
			isRemoveAllowed
			) {
			return true;
		}
		return false;
	},
	
	executePostLinkEvent: function(linkTypeURI, createdLink, flipLinks) {			
			if(flipLinks) {
				// cross module or cross project... flip link
				if(linkTypeURI.indexOf("/subject") !== -1) {
					linkTypeURI = linkTypeURI.replace("/subject", "/object");
				} else {
					linkTypeURI = linkTypeURI.replace("/object", "/subject");
				}
			}
			
			var linkURI = linkutils.getFormatedLinkURI(linkTypeURI),
				linkName = linkutils.getLinkName(linkTypeURI, createdLink);
			
			var confirmMsg = dojo.string.substitute(this._messages.linkCreateDialogBody, [ linkName ]),
				dialogTitle = this._messages.linkCreateDialogTitle;
			
			var grid = this.grid;
			var colId = linkutils.getColIdByLinkUri(linkURI);
			if(!grid.gridmodel.isVisible(com.ibm.rdm.web.artifactgrid.ArtifactConstants.RESOURCE_ROW, colId)) {
				
				var msgForm = new com.ibm.rdm.web.common.MessageForm({
					confirmContent: confirmMsg,
					dialogTitle: dialogTitle,
					dialogStyle: "yesNo",
					dialogWidth:	"350px",
					onOk: function() {
						grid.addCol(colId);
					}
				});
				msgForm.show();
			}
	},
	
	isClearSuspicionEnabled: function(rowIds, count) {
		return false;
	},
	
	executeClearSuspicion: function(rowIds) {
		var toClear = this._getRecordInfosFromRowIds(rowIds);
		
		var suspicionProfileUris = SuspectLinksUtil.getSuspicionProfileURIsPreference(this._projectURI);
		var clearSuspicionForm = new com.ibm.rdm.web.suspicion.ClearSuspicionForm({
			projectUri: this._projectURI,
			suspicionEntries: [],
			suspicionProfileUris: suspicionProfileUris
		});

		var msgForm = new com.ibm.rdm.web.common.MessageForm({
			confirmContent: clearSuspicionForm.domNode,
			dialogTitle: this._messages.confirmClearSuspicion,
			dialogWidth: "350px",
			dialogStyle: "okCancel",
			okButtonText: this._messages.clear,
			onDialogOpen: dojo.hitch(this, function() {
				dojo.fadeIn({node: clearSuspicionForm.domNode, duration: 400}).play();
			}),
			onOk: dojo.hitch(this, function() {
				var refresh = [];
				var entryUris = [];
				var suspectEntries;
				dojo.forEach(toClear, function(resource) {
					suspectEntries = resource.suspectEntries;
					if (suspectEntries && suspectEntries.length > 0) {
						entryUris = entryUris.concat(suspectEntries);
						resource.suspectEntries = [];
						refresh.push({
							key: resource.getResourceURI(),
							rowtype: com.ibm.rdm.web.artifactgrid.ArtifactConstants.RESOURCE_ROW
						});
					}
				});
				SuspectLinksUtil.clearSuspicion(entryUris).then(
					dojo.hitch(this, function(response) {
						this.grid.refreshRows(refresh);
						return response;
					})
				);
			})
		});
		msgForm.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},
	
	getClearSuspicionEnablementDeferred: function(rowIds, count) {
		var clearSuspicionEnabledDfd = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
		
		projectDeferred.then(dojo.hitch(this, function() {
			var isAllowed = true;
			isAllowed = this._isClearSuspicionEnabled(rowIds, count);
			if (!isAllowed) {
				clearSuspicionEnabledDfd.callback(false);
			} else if (count > 1) {
				// no permission check if more than one link selected
				clearSuspicionEnabledDfd.callback(true);
			} else {
				// TODO: Real clear suspicion permission check
				clearSuspicionEnabledDfd.callback(true);
			}
		}));
		return clearSuspicionEnabledDfd;
	},
	
	_isClearSuspicionVisible: function(rowIds, count) {
		var visible = false;
		var toClear = this._getRecordInfosFromRowIds(rowIds);
		
		dojo.some(toClear, function(resource) {
			var suspectEntries = resource.suspectEntries;
			if (suspectEntries && suspectEntries.length > 0) {
				visible = true;
				return false;
			}
		});
		
		return visible;
	},
	
	_isClearSuspicionEnabled: function(rowIds, count){
		var isEditAllowed = false;
		
		isEditAllowed = this._projectOperations.item(com.ibm.rdm.web.common.ProjectCache.CLEAR_SUSPICION) 
				&& !this._baselineURI && !this._isInHistoryMode;

		return isEditAllowed;
	},
	
	isOpenEnabled: function(rowIds, count) {
		// summary:
		//		returns true if the open artifact menu function should be enabled.
		return count == 1; 
	},	
	
	_getKeyComboForOperation: function(operationId){
		var keyCombo = null;

		var getKeyCombo = dojo.hitch(this, function(keyHandlers){
			var keyCombo = null,
				i;
			for (i in keyHandlers){
				if (keyHandlers[i].operationName === operationId){
					keyCombo = this._buildKeyCombo(keyHandlers[i]);
					break;
				}
			}
			return keyCombo;
		});
		
		keyCombo = (
			getKeyCombo(this.grid.getKeyPressHandlers())
			|| getKeyCombo(this.grid.getKeyDownHandlers())
		);
			
		return keyCombo;
	},

	_buildPreexistingBrowserBasedKeyCombo: function(operationId, count) {
		// summary:
		//		Add shortcut info for some of the preexisting browser based
		//		shortcuts.
		// description:
		//		The UX team wants us add info for some of the preexisting browser
		//		based shortcuts, such as Ctrl-click (to open in new tab) and
		//		Shift-click (to open in new window).  We do that here.

		var keyString = "";
		if (count === 1) {
			switch (operationId) {
			case ".openInNewTab":
				if (dojo.isMac) {
					keyString = dojo.string.substitute(
						this._messages.clickLinkShortcut,
						["&#8984;"] // Apple command key
					);
				} else {
					keyString = dojo.string.substitute(
						this._messages.clickLinkShortcut,
						["Ctrl"]
					);
				}
				break;
			case ".openInNewWindow":
				if (dojo.isMac && dojo.isSafari) {
					keyString = dojo.string.substitute(
						this._messages.clickLinkShortcut,
						[
							"&#8984;" // Apple command key
							+ "-"
							+ "&#8997;" // Apple option key
						]
					);
				} else {
					keyString = dojo.string.substitute(
						this._messages.clickLinkShortcut,
						["Shift"]
					);
				}
				break;
			}
		}
		return keyString;
	},
	
	_buildKeyCombo: function(keyPressHandler){
		var keyString = "";
		if (keyPressHandler.ctrl){
			if (dojo.isMac && dojo.isSafari && keyPressHandler.copyKey){
				keyString += "&#8984;"; // Apple command key
			}
			else if (dojo.isMac && dojo.isSafari && !keyPressHandler.copyKey){
				keyString += "&#8963;";
			}
			else{
				keyString += "Ctrl+";
			}
		}
		if (keyPressHandler.alt){
			if (dojo.isMac){
				keyString += "&#8997;"; // Apple option key
			}
			else{
				keyString += "Alt+";
			}
		}
		if (keyPressHandler.shift){
			if (dojo.isMac){
				keyString += "&#8679;"; // Apple shift key 
			}
			else{
				keyString += "Shift+";
			}
		}
		if (keyPressHandler.keyChar){
			keyString += keyPressHandler.keyChar.toUpperCase();
		}
		if (keyPressHandler.keyCode){
			if (keyPressHandler.keyChar){
				keyString += "+";
			}
			switch (keyPressHandler.keyCode){
				case dojo.keys.DELETE:
					if(dojo.isMac) {
						keyString += "&#8998;"; // Apple erase right (delete)
					}
					else {
						keyString += "Del";
					}
					break;
				case dojo.keys.ENTER:
					if (dojo.isMac) {
						keyString += "&#8617;"; // Apple Return key
					} else {
						keyString += "Enter";
					}
					break;
				case dojo.keys.INSERT:
					keyString += "Ins";
					break;
				case dojo.keys.RIGHT_ARROW:
					keyString += "&rarr;";
					break;
				case dojo.keys.LEFT_ARROW:
					keyString += "&larr;";
					break;
				case dojo.keys.F1:
					keyString += "F1";
					break;
				case dojo.keys.F2:
					keyString += "F2";
					break;
				case dojo.keys.F3:
					keyString += "F3";
					break;
				case dojo.keys.F4:
					keyString += "F4";
					break;
				case dojo.keys.F5:
					keyString += "F5";
					break;
				case dojo.keys.F6:
					keyString += "F6";
					break;
				case dojo.keys.F7:
					keyString += "F7";
					break;
				case dojo.keys.F8:
					keyString += "F8";
					break;
				case dojo.keys.F9:
					keyString += "F9";
					break;
				case dojo.keys.F10:
					keyString += "F10";
					break;
				case dojo.keys.F11:
					keyString += "F11";
					break;
				case dojo.keys.F12:
					keyString += "F12";
					break;
			}
		}
		return keyString;
	},
		
	/////////////////////////////////////////////////////////
	//// Parsing Commands
		
	parseRequirementText: function(num) {
		if (num === 1) {
			return this._messages.parseRequirement;
		}
		return dojo.string.substitute(this._messages.parseRequirements, [num]);
	},
	
	isParseEnabled: function(rowIds, count) {
		// summary:
		//		Determine if the parse requirements menu option should be enabled. At the moment,
		//		it should only be enabled if one resource is selected, the user is able to
		//		modify the resource, and it is one of the following artifact types: RM-RTF,
		//		doc, docx, rtf, odt.
		// rowIds: int
		//		The identifier of each of the rows that are selected.
		// count: int
		//		The count of the rows that have been selected.
		return false;
	},
	
	getParseEnabledDeferred: function(rowIds, count) {
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
		
		projectDeferred.then(dojo.hitch(this, function() {
			var resourceInfo, richTextType, mimeType, wrappedResourceMimeType = null;
			var validMimeType = false;
			if (!this._baselineURI){
				//since C/ALM rows do no contain the same methods, disable action.
				if (this.containsCalmRow(rowIds)) {
					deferred.callback(false);
				}
			
				for (var key in rowIds) {
					resourceInfo = this._dataCache.getRecordInfo(key);
			
					if (resourceInfo.getWrappedResourceURI && resourceInfo.getWrappedResourceURI()){
						wrappedResourceMimeType = resourceInfo.getWrappedResourceMimeType();
					}
				
					if (wrappedResourceMimeType !== null) {
						mimeType = wrappedResourceMimeType;
					}
					else {
						mimeType = resourceInfo.getMimeType();
						richTextType = MimeTypes.getArtifactFormat(MimeTypes.REQUIREMENT_NEW_NAME, this._projectURI);
					}
				
					if(mimeType === MimeTypes.DOC.contentType ||
							mimeType === MimeTypes.DOCX.contentType ||
								mimeType === MimeTypes.ODT.contentType ||
									mimeType === richTextType) {
						validMimeType = true;
					}
					else {
						validMimeType = false;
						break;
					}
				}

				var projectOperations = this._projectOperations;	
				deferred.callback(projectOperations && projectOperations.item(ProjectCache.MODIFY_RESOURCE) && validMimeType);
			}
			else {
				deferred.callback(false);
			}
		}));
		
		return deferred;
	},
	
	doParseRequirements: function(rowIds) {
		var resourceInfo = null;
		var resourceURI = null;
		var resourceURIs = [];
		for (var key in rowIds) {
			resourceInfo = this._dataCache.getRecordInfo(key);
			resourceURI = (resourceInfo) ? resourceInfo.getResourceURI() : null;
			resourceURIs.push(resourceURI);
		}
		
		var wizard = new com.ibm.rdm.web.external.ImportWizard({
			projectURI: this._projectURI
		});
		wizard.addWizardSteps(com.ibm.rdm.web.external.ImportWizard.wizardType.REQ_PARSING, 
			{
				sourceURIs: resourceURIs
			}
		);
		wizard.show(dojo.hitch(this, this._onLaunchedDialogHasClosed));
	},
	
	moduleChildText: function(num, rowIds, key){
		// summary:
		//		Text that should be used for copy menu option
		// num: int
		//		The number of rows that are selected in the grid.
		if (num === 1) {
			var artifactURI = key;
			var resource = this._dataCache.getRecordInfo(artifactURI);
			var objectTypes = resource.getTypes();
			var objectType = null;
			for (var type in objectTypes) {
				objectType = this._typeCache.getProjectObjectType(type);
				break;
			}
			return dojo.string.substitute(this._messages.createModuleChildArtifact, [objectType.getTitle()]);
		}
		return this._messages.createModuleChildArtifacts;
	},

	_getArtifactTypeURI: function(resourceInfo) {
		var artifactTypeUris = resourceInfo.getTypes();
		// Note: we only ever have one URI found in the object returned by getTypes() ...
		for (var artifactTypeURI in artifactTypeUris) {
			return artifactTypeURI;
		}
		return null;
	},

	getEditAttributesEnablementDeferred: function(rowIds, count) {
		// summary:
		//		Returns a deferred which will be called back with a single
		//		param representing whether the "Edit Attributes" menu
		//		item should be enabled.
		// returns:
		//		A Deferred.
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
		
		projectDeferred.then(dojo.hitch(this, function() {
			if (count === 1 && this._editInterface.isEditAttribsEnabled(rowIds, count)){
				var resourceInfos = this._getRecordInfos(
					this._getUniqueKeysFromRowIds(rowIds)
				);
				this._getModifyArtifactPermissions(resourceInfos[0]).then(
					dojo.hitch(this, function(modifyArtifactPermissions) {
						deferred.callback(
							//We need the dialog if we can edit attributes OR we can edit the artifact itself -
							//as we can now change the type.
							modifyArtifactPermissions.isModifyCustomAttributesAllowed() 
							|| modifyArtifactPermissions.isModifyArtifactAllowed()
						);
					})
				);
			} else {
				deferred.callback(count > 1 && this._editInterface.isEditAttribsEnabled(rowIds, count));
			}
		}));
		return deferred;
	},

	getEditRowEnablementDeferred: function(rowIds, count) {
		// summary:
		//		Returns a deferred which will be called back with a single
		//		param representing whether the "Edit Artifact" menu
		//		item should be enabled.
		// returns:
		//		A Deferred.
		var deferred = new dojo.Deferred();
		var projectDeferred = ProjectCache.getProjectPermissions(this._projectURI);
		
		projectDeferred.then(dojo.hitch(this, function() {
			if (count === 1) {
				var resourceInfos = this._getRecordInfos(
					this._getUniqueKeysFromRowIds(rowIds)
				);
				this._getModifyArtifactPermissions(resourceInfos[0]).then(
					dojo.hitch(this, function(modifyArtifactPermissions) {
						if (modifyArtifactPermissions.isModifyArtifactAllowed() &&
							this._editInterface.isEditRowEnabled(rowIds, count)
							) {
							deferred.callback(true);
						} else {
							deferred.callback(false);
						}
					})
				);
			} else {
				deferred.callback(false);
			}
		}));
		return deferred;
	},
	
	getLockRowEnablementDeferred: function(rowIds, count) {
		// summary:
		//		Checks whether the user has manual lock permission for the selected artifacts process areas
		//		and whether the user is able to edit those artifacts.
		//		For multiple rows, only checks manual lock permission to keep server calls down

		var deferred = new dojo.Deferred(), // return this
			idx, // loop counter
			resourceInfos = this._getRecordInfos(
				this._getUniqueKeysFromRowIds(rowIds)
			),
			processAreaUri,
			uniqueProcessAreas = [],
			processAreasChecked = 0,
			manualLockAllowedOnAllProcessAreas = true;
		
		//Never allow taking locks in a baseline
		if (this._baselineURI) {
			deferred.callback(false);
		}
		
		if (count === 1) {
			processAreaUri = resourceInfos[0].getProcessAreaURI() || this._projectURI;
			ProcessOperationsCache.isManualLockAcquireAllowed(processAreaUri, this._projectURI).addCallback(this, function(manualLockAllowed) {
			   if (!manualLockAllowed) {
				   deferred.callback(false);
			   } else {
				   this.getEditRowEnablementDeferred(rowIds, count).addCallback(function (val) {
					  deferred.callback(val); 
				   });
			   }
			});
			
		} else {
			// to try and reduce calls to the ProcessOperationsCache, create an array
			// the contents of which will be unique process area URIs in the selection.
			// A simple object would be easier for this, but we need to reliably (i.e on IE8)
			// check the number of unique process areas and cannot use Object.keys() there.
			for (idx =0; idx < resourceInfos.length; idx++) {
			   processAreaUri = resourceInfos[idx].getProcessAreaURI() || this._projectURI;
			   if (dojo.indexOf(uniqueProcessAreas, processAreaUri) === -1) {
				  uniqueProcessAreas.push(processAreaUri);
			   }
			}
			// for each process area, check whether the user has manual lock permission
			dojo.forEach(uniqueProcessAreas, function(processArea) {	
					
				ProcessOperationsCache.isManualLockAcquireAllowed(processAreaUri, this._projectURI).addCallback(this, function(manualLockAllowed) {
					processAreasChecked++;
					if (!manualLockAllowed) {
						manualLockAllowedOnAllProcessAreas = false;
					}
					
					if (processAreasChecked === uniqueProcessAreas.length) {
						// callback with the result, once we have checked them all.
						// potentially slight improvement here as we could have bailed
						// out earlier.
						deferred.callback(manualLockAllowedOnAllProcessAreas);
					}
					
				});
			}, this);
			
		}
		
		return deferred;
	},
	
	isUsedInModule: function(rowIds) {
		for (var resourceURI in rowIds) {
			var resource = this._dataCache.getRecordInfo(resourceURI);
			if (resource && resource.getModuleHostURI && resource.getModuleHostURI()) {
				return true;
			}
		}
		return false;
	}
});

})();
