/*******************************************************************************
 * Licensed Materials - Property of IBM
 * (c) Copyright IBM Corporation 2007, 2014. All Rights Reserved.
 *
 * Note to U.S. Government Users Restricted Rights:
 * Use, duplication or disclosure restricted by GSA ADP Schedule
 * Contract with IBM Corp.
 *******************************************************************************/

(function () {
dojo.provide("com.ibm.rdm.web.grid.Module");

dojo.require("dijit._Widget");
dojo.require("dijit._Templated");
dojo.require("dijit.form.NumberTextBox");

dojo.require("com.ibm.rdm.web.grid.List");
dojo.require("com.ibm.rdm.web.grid.EditMenu");
dojo.require("com.ibm.rdm.web.util.domutils");
dojo.require("com.ibm.rdm.web.util.keyutils");
dojo.require("com.ibm.rdm.web.common.ErrorHandler");
dojo.require("com.ibm.rdm.web.util.bidiutils"); //bidi
dojo.require("jazz.ui.toolbar.Button");
dojo.require("com.ibm.rdm.web.common.dnd.DragZone");
dojo.require("com.ibm.rdm.web.common.dnd.DragDropFactory");
dojo.require("com.ibm.rdm.web.artifactgrid.ArtifactConstants");
dojo.require("dojo.NodeList-traverse");
dojo.requireLocalization("com.ibm.rdm.web.grid", "Module");


var RESOURCE_ROW = com.ibm.rdm.web.artifactgrid.ArtifactConstants.RESOURCE_ROW;

// This function is made to ensure that the supplied node can
// safely be detached form the DOM and then reattached.
var _nodeCanBeDetached = function (node) {
	return dojo.query(".cke_skin_rational", node).length === 0;
};

var DomUtils = com.ibm.rdm.web.util.domutils;
var INDENT_CHUNK = 5;

// time to wait after a refresh before attempting to set focus.
var PRE_FOCUS_DELAY_MS = 100;

// Root part of an aria-describedby attribute value for each cell.  Has the
// id of the row added so that it refers to a unique description element for
// each row.
var ARIA_ROW_DESCRIPTION_PREFIX = "ariaRowDescription-";

var BidiUtils = com.ibm.rdm.web.util.bidiutils;  //bidi

dojo.declare(
	"com.ibm.rdm.web.grid.Module",
	[dijit._Widget, dijit._Templated],
{
	// summary:
	//		Provides a widget which displays grid-like content.

	widgetsInTemplate: true,

	templateString: '<div dojoAttachPoint="containerNode" class="com-ibm-rdm-web-Module" ></div>',

	selectedObject: null,
	
	selectedRows: null,
	
	//The ordered list of rows in this module.
	rowOrder: null,
	
	totalRowCount: 0,
	
	// Added by ebcraig -
	// used in case of lightweight grid such as in reporting wizard
	// to prevent normal keypress/click handlers from interfering with function
	// as intended
	skipHandlerHookup: false,
	
	// May be set on construction.
	// If false, tabbing into and out of the grid is not managed, but tabbing follows
	// the normal browser behaviour, moving between input and other elements that have
	// tabIndex set on them.
	// If true (default), the grid catches tabs into it and clicks on cells, so that
	// the grid can be navigated with arrow keys, rows placed in ActionableMode etc.
	// In this case, all cell content must not be focusable by default, and methods must
	// be provided in the gridmodel column definitions to handle editing.
	manageFocusNavigation: true,
	
	// May be set on construction.
	// On an instance with manageFocusNavigation==false, this flag may be set to
	// allow the rows to be tab focus targets, and thereby allowing keyboard selection
	// of rows.
	// It is only valid for this to be true when manageFocusNavigation is false.  When manageFocusNavigation
	// is true, rows are always selectable.
	manageRowSelection: false,

	//A collection of objectIds to actual nodes.
	allRows: null,

	dojoConnectList: null,
	
	// A function that, if defined, is passed event information for keypresses.  Should return
	// false or undefined if the key has been handled.  Return true to indicate that the
	// key has not been handled, in which case the key event is allowed to propagate.
	keyPressHandler: null,
	// A function that, if defined, is passed event information for keydowns.  Should return
	// false or undefined if the key has been handled.  Return true to indicate that the
	// key has not been handled, in which case the key event is allowed to propagate.
	keyDownHandler: null,

	//fetchFallback: Boolean
	//flag sent as part of a getPage request to return the first page if the target object is deleted.
	fetchFallback: false,
	
	//The grid model
	gridmodel: null,
	
	// doTrackFocusForCells: Boolean
	//		Optionally passed into constructor.
	//		If true, rows will have the class com.ibm.rdm.web.grid.Module.ClassNames.ROW_FOCUS
	//		as long as a cell in that row has focus.
	doTrackFocusForCells: false,
	
	// An array of collapsed artifact uris.
	// Used for re-apply the collapsed state of the artifact after refresh.
	collapsedArtifactUris: [],
	
	// An array of expanded artifact uris.
	// Used for re-apply the expanded state of the artifact after refresh.
	expandedArtifactUris: [],
	
	// The structure of the module.
	// Used for toggle the collapsed state of the artifact after refresh
	moduleStructure: null,
	
	// An array of selected artifact uris.
	// Used for re-select the selected artifacts after refresh.
	selectedArtifactsUri: [],
	
	//Functions used to calculate highlight classes
	_highlightFunctions: null,
	
	//Keep track of the id->row key mapping, so we can know what objects to destroy
	_idToKey: null,
	
	// Identifies the row or cell that is the target for tabbing into or out of the grid.
	// Must contain rowKey and columnIdMap.  columnIdMap is a hash of rowType to columnId.
	// We keep a separate columnId for each rowType so that when focus moves vertically
	// through different row types, for each row type we return to the same column.
	_navigationTabTarget: null,
	
	// A node whose purpose is to catch keyboard focus when focus is currently not within the
	// grid.  Provides a tabtarget that can catch tabs that would otherwise skip the grid.
	// Only set and used when this.manageFocusNavigation is true.
	_focusCatcherNode: null,
	
	// Flag to cotrol delayed focus.  Required so that when focus is given to the grid
	// through a mouse click, the initial onfocus event does not cause fucus to jump to
	// the old focus node (so scrolling the grid) before the new focus point has been
	// found and set by the click handler.
	_allowPendingFocusToContinue: false,
	
	// An array of objects inidcating key combinations and matching handler functions.
	_internalKeyPressHandlers: null,
	// An array of objects inidcating key combinations and matching handler functions for keydown.
	_internalKeyDownHandlers: null,
		
	_toDestroy: null,
	
	//Array of domnodes to be cleared selectively
	_nodeBucket: null,
	
	_linkDropEvent: null,
	_linkBoxEvent: null,
	_linkColEvent: null,

	_wasInEditModeWhenOpenEditMenuWasRequested: false,
	_wasInStubRowWhenOpenEditMenuWasRequested: false,
	

	constructor: function () {

		this._highlightFunctions = {};
		this.allRows = {};
		this._idToKey = {};
		this.rowOrder = [];
		this._toDestroy = [];
		this._nodeBucket = [];
		this._linkColParents = {};
		
		this.selectedRows = {};
		this.totalRowCount = 0;
		this._navigationTabTarget = {columnIdMap: {}};
		this.dojoConnectList = new com.ibm.rdm.web.grid.List();
		this.messages = dojo.i18n.getLocalization("com.ibm.rdm.web.grid", "Module");
	},

	onNodeFocused: function(/*DOMNode*/node) {
		// summary:
		//		User defined function to handle the case of a new node becoming
		//		selected in the module.
		// tags:
		//		callback
	},
		
	getSortFunction: function (rowType, columnId) {
		return this.gridmodel.getSortFunction(rowType, columnId);
	},
	
	removeCol: function(columnId) {
		dojo.query("."+columnId, this.domNode).orphan();
	},
	
	addCol: function (rowTypeId, columnId, position) {
		// starting at the first child of the div containing the server
		// HTML data...
	
		var node = this.content.firstChild;
		var functionsToCall = [];
		
		while (node) {
			var thisRowTypeId = dojo.attr(node, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			
			//If the right row type, or a child of the row type.
			if (thisRowTypeId === rowTypeId || this.gridmodel.isChildRowTypeOf(thisRowTypeId, rowTypeId)) {
			
				//get the key
				var key = dojo.attr(node, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
				
				//get the backing record
				
				var contentRow = this.gridmodel.getRowData(thisRowTypeId, key);
				
				if (!contentRow) {
					console.warn("No ContentRow Data for", thisRowTypeId, key);
				}

				//create the div
				var cellDiv = this._buildCell(thisRowTypeId, columnId, functionsToCall, contentRow, key, 0);
				
				var width = this.gridmodel.getWidth(thisRowTypeId, columnId)+"px";
				var isAutoWidth = this.gridmodel.isAutoWidth(thisRowTypeId, columnId);
				if (isAutoWidth) {
					dojo.style(cellDiv, "width", "100%");
					dojo.query(columnId, cellDiv).style("width", "");
				} else {
					dojo.style(cellDiv, "width", width);
					dojo.query(columnId, cellDiv).style("width", width);
				}

				//place it
				var td = node.rows[0].insertCell(position);
				dojo.place(cellDiv, td, "replace");
				//if (columnId === com.ibm.rdm.web.grid.GridModel.COL_EXPANDO) {
					
				//} else {}
				//dojo.place(expandoCell, colTd, "first");
				
				//For IE, we need to reapply the selection checkbox... sigh.
				if ((dojo.isIE > 0) && this.IE_reselect_hack) {
					this.IE_reselect_hack.checked = true;
				}
			}
			node = node.nextSibling;
		}
	
		//Now, lets makes sure we get the deferred values we want
		this._callDeferredFunctions(functionsToCall);
	},
	
	_getValueFromRecord: function (rowTypeId, columnId, contentRow) {
	
		try {
			if (!contentRow) {
				console.error("Trying to get value from a null contentRow for:", rowTypeId, columnId);
			}
			//Calculate the value (as a string) to be displayed.
			var valueKey = this.gridmodel.getValueKey(rowTypeId, columnId);
				
			if (valueKey) {
				//Caller expects to use the value in the record given by this key
				var value = dojo.getObject(valueKey, false, contentRow);

				if (this.gridmodel.getValueFunction(rowTypeId, columnId)) {
					//caller has also provided a function to apply to the value from the record.
					return this.gridmodel.getValueFunction(rowTypeId, columnId)(value);
				}
				//no function, just use the value in the record.
				return value;
			}
			
			if (this.gridmodel.getRecordFunction(rowTypeId, columnId)) {
				//Caller has provided a synchronous call which will return the value, given the record.
				return this.gridmodel.getRecordFunction(rowTypeId, columnId)(contentRow);
			}
		} catch (e) {
			com.ibm.rdm.web.common.ErrorHandler.error("Error Calculating Row, Col", rowTypeId, columnId, e);
			return "ERROR";
		}
	},
	
	_mergeCell: function (args) {

		var left = args.left,
			right = args.right,
			wrapper;
			
		dojo.place(left, right, "first");

		wrapper = dojo.create("div", {
			"class": args.wrapperClass || ""
		});
					
		while (right.firstChild) {
			wrapper.appendChild(right.firstChild);
		}
		// FireFox will always grow a <td> to accomodate
		// the width of it's children, we therefore require
		// a div of the same width with overflow: hidden
		// to prevent this from happening.  Because this
		// overflowWrapper div gets the same class as the
		// column, it will be sized to the column width
		// but its children will not cause the grid to
		// appear jagged.
		var overflowWrapper = dojo.create("div", {
			"class": dojo.attr(right, "class"),
			style: {
				"overflow": "hidden"
			}
		});
		
		overflowWrapper.appendChild(wrapper);
		right.appendChild(overflowWrapper);
		
		dojo.addClass(right, "expandoContainer");
	},
	
	_createRow: function(rowTypeId, key, contentRow, functionsToCall) {
		var contentsDiv;
		
		var rowDivOuter = document.createElement('table');
		dojo.attr(rowDivOuter, "objectId", key);
		dojo.attr(rowDivOuter, "rowtype", rowTypeId);
		dojo.addClass(rowDivOuter, rowTypeId);
		if (this.manageRowSelection) {
			// We are allowing the browser to handle focus through normal tab behaviour.
			// However, we do need allow selection of a row, so make the row focusable.
			// Allow the overall row to be a focus target so that the row can be
			// keyboard selected.
			dojo.attr(rowDivOuter, "tabIndex", "0");
		}

		if (this.selectedRows[key]) {
			//Add selected class.
			dojo.addClass(rowDivOuter, com.ibm.rdm.web.grid.Module.ClassNames.ROW_CHECKED);
		}
		
		//Apply any highlighting due
		this._applyHighlightClasses(rowTypeId, key, rowDivOuter);
		
		//work out and apply indent
		var indent = this.gridmodel.getIndentMultiple(rowTypeId, key);
		
		if(typeof indent.level === "number") {
			dojo.addClass(rowDivOuter, "indent"+indent.level);
		}
		
		var rowIndentToApply = (indent.row*INDENT_CHUNK);
		var cellIndentToApply = (indent.cell*INDENT_CHUNK);
		
		var rowClassFunctions = this.gridmodel.getRowClassFunctions(rowTypeId);
		if (rowClassFunctions) {
			for (var j=0; j < rowClassFunctions.length; j++) {
				var rowClasses = rowClassFunctions[j](key);
				if (rowClasses) {
					for (var k=0; k < rowClasses.length; k++) {
						dojo.addClass(rowDivOuter, rowClasses[k]);
					}
				}
			}
		}
		var temp = document.createElement('tbody');
		dojo.place(temp, rowDivOuter, "only");
		
		var rowDivInner = document.createElement('tr');
		dijit.setWaiRole(rowDivOuter, "row");	// ARIA
		dojo.place(rowDivInner, temp, "only");
		
		// Add a second hidden row in the table to contain the ARIA description for the row
		var hiddenRow = dojo.create("tr", {style: "display:none;"}, temp);
		var descriptionCell = dojo.create("td", {id: ARIA_ROW_DESCRIPTION_PREFIX+key,"class":com.ibm.rdm.web.grid.Module.AttributeNames.ROW_DESCRIPTION}, hiddenRow);
		// Register a function to populate this cell and other ARIA info on the row when the recsource is available.
		var ariaContentDescriptor =  {
			func: dojo.hitch(this, this._updateAriaRowInfo),
			setContentFn: function(){},
			row: contentRow,
			key: key,
			containingNode: rowDivOuter
		};
		functionsToCall.push(ariaContentDescriptor);

		//Build the row column by column
		var cols = this.gridmodel.getColumnIds(rowTypeId);
		
		var compIndent = 0;

		
		var expandoCell, colTd;
		for (var j = 0; j < cols.length; j++) {
			if (this.gridmodel.isVisible(rowTypeId, cols[j])) {

				colTd = this._buildCell(rowTypeId, cols[j], functionsToCall, contentRow, key, compIndent, cellIndentToApply);

				if (expandoCell) {
					this.gridmodel.EXPANDO_CONTAINER = colTd;
					this._mergeCell({
						left: expandoCell,
						right: colTd,
						wrapperClass: "expandoWrapper"
					});
					// set ARIA property to indicate that this cell contains and expando.  Description
					// will then take into account aria-expanded property.
					dijit.setWaiRole(colTd, "range");
					expandoCell = null;
				}
				
				if (cols[j] === com.ibm.rdm.web.grid.GridModel.COL_EXPANDO) {
					expandoCell = colTd;
				}
				
				//The first (non selection) col width is increased to provide the indent
				if (cols[j] != "_selector" && rowIndentToApply > 0) {
					dojo.style(colTd,  "paddingLeft", ""+(rowIndentToApply+2)+"px");
					rowIndentToApply = 0;
				}

				dojo.place(colTd, rowDivInner, "last");
				
				//For IE, we need to reapply the selection checkbox... sigh.
				if ((dojo.isIE > 0) && this.IE_reselect_hack) {
					this.IE_reselect_hack.checked = true;
				}
			}
		}
		

		
		// Attach dragover/dragleave events directly to the table rows, otherwise
		// we need to constantly look up the parent table row whenever we get
		// one of these events..
		if (this.dragDropArgs) {
			
			
			
			if (this.dragDropArgs.dropTarget && this.dragDropArgs.isDropEnabled()) {
				var dropZone = new com.ibm.rdm.web.common.dnd.DropZone({
					node: rowDivOuter,
					accepts: this.dragDropArgs.validDropTypes,
					onDrop: dojo.hitch(this, function(ev) {
						
						if(this._linkDropEvent) {
							// Our own hacky stopPropagation
							this._linkDropEvent = null;
							return;
						}
						
						var dropTarget = rowDivOuter,
							beforeFirst = false,
							overTopThird = this._isInTopThird(ev, rowDivOuter);

						// we now have the table row the drop took place on
						if (overTopThird) {
							// if we were over the top 1/3 of the row, get
							// the previous sibling
							dropTarget = rowDivOuter.previousSibling;
							if (dropTarget === this._dndWrapper) {
								// accounting for the fact that the previous sibling
								// might be the drag pointer itself.
								dropTarget = dropTarget.previousSibling;
								if (!dropTarget || !dojo.attr(dropTarget, "objectid")) {
									dropTarget = this._dndWrapper.nextSibling;
									beforeFirst = true;
								}
							}
						}
						this.dragDropArgs.onDrop(ev, dojo.attr(dropTarget, "objectid"), beforeFirst);
					}),
					onDragOver: dojo.hitch(this, function(ev){
							this.dragOverCommon(ev, rowDivOuter, false);
					})
				});
				
				this._toDestroy.push(dropZone);
			} else if (this.dragDropArgs.dragTarget && !this.dragDropArgs.dropTarget) {
				var dragZone = new com.ibm.rdm.web.common.dnd.DragZone({
					node: rowDivOuter,
					accepts: this.dragDropArgs.validDropTypes,
					onDragOver: dojo.hitch(this, function(ev){
						this.dragOverCommon(ev, rowDivOuter, false);
					})
				});
				
				this._toDestroy.push(dragZone);
			}
		}
		
		return rowDivOuter;
	},
	
	dragOverCommon: function(ev, rowDivOuter, linkEvent) {
		if(ev.dataTransfer && dojo.indexOf(ev.dataTransfer.types, "Files") !== -1) {
			// We don't want to decorate the grid when performing a file upload drag
			return;
		}
		var node = rowDivOuter,
			overTopThird = (linkEvent) ? false : this._isInTopThird(ev, rowDivOuter),
				isDropEnabled = this.dragDropArgs.isDropEnabled(),
				linkBoxClass = "link";
			
			if (overTopThird) {
				node.parentNode.insertBefore(this._dndWrapper, node);
			} else {
				node.parentNode.insertBefore(this._dndWrapper, node.nextSibling);
			}
			
			if(this.dragDropArgs.isModule) {
				// Modules get special treatment
				if (!this._dndWrapper.previousSibling && isDropEnabled) {
					linkBoxClass += " first book";
				} else if(!this._dndWrapper.previousSibling && !isDropEnabled) {
					linkBoxClass += " first";
				} else if(isDropEnabled) {
					linkBoxClass += " book";
				}
			}
			this._linkBox.className = linkBoxClass;

			if(typeof this.dragDropArgs.onDragOver === "function") {
				this.dragDropArgs.onDragOver(ev, dojo.attr(node, "objectid"), linkEvent);
			}
			
			if(this._linkBoxEvent) {
				// Linkbox styles lingered somehow...
				// usually because the user was going reeeeally fast and skipped the dragLeave event
				dojo.removeClass(this.domNode, "link-box-active");
				dojo.removeClass(this._linkBox, "active");
				this._linkBoxEvent = null;
			}
			if(this._linkColEvent) {
				// clean up any leftover link events
				dojo.query("td.link-active", this.domNode).removeClass("link-active");
				this._linkColEvent = null;
			}
	},
	
	_isInTopThird: function(ev, node) {
		// TODO Move me
		// Note: this function fires frequently...
		// dojo.position is substantially slower than the
		// native equivalent
		var pos = node.getBoundingClientRect();
		if (pos.top + (pos.height/3) > ev.pageY) {
			return true;
		}
		return false;
	},
	
	_updateAriaRowInfo: function (functionToCall, resource, key, rowNode) {
		// summary:
		//		Sets various values on a row that provide the ARIA descriptions for the row.
		//		Specifically:
		//			- sets a row description value that is used as part of the row
		//			  description when focus is on a cell.  This is the value in an
		//			  element referenced by an aria-describedby attribute.
		//			- sets a rowLabel attribute to a value that describes the row.  This is not
		//			  an ARIA value directly, but is used to build a row label as the
		//			  row is selected and deselected.
		//			- sets the aria-label on the row taking into account the rowLabel value set
		//			  above, and whether the row is selected.  This provides a description of
		//			  the row when focus is on the whole row.
		var rowTypeId = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
		var rowLabelFunction = this.gridmodel.getRowLabelFunction(rowTypeId);
		var rowLabel = rowLabelFunction ? rowLabelFunction(resource) : "";
		dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROW_LABEL, rowLabel);
		var rowDescriptionFunction = this.gridmodel.getRowDescriptionFunction(rowTypeId);
		if (rowDescriptionFunction) {
			var rowDescription = rowDescriptionFunction(resource);
			var ariaRowDescriptionNode = dojo.query(".rowDescription", rowNode)[0];
			//If this is undefined, we do not want to add innerHTML.  It will cause IE
			//		to freak out, and not display entire artifact row.
			if(ariaRowDescriptionNode){
				ariaRowDescriptionNode.innerHTML = rowDescription;
			}
		}
		this._updateAriaSelectionLabels(key, rowNode);
		functionToCall();
	},
	
	_updateAriaSelectionLabels: function (key, rowNode) {
		// summary:
		//		Sets the aria-label on the row taking into account the rowLabel value that is
		//		already set on a row, and taking into account whether or not the row is
		//		selected.  This provides a description of the row when focus is on the whole row.
		var ariaRowLabel = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROW_LABEL);
		ariaRowLabel = ariaRowLabel || "";
		// OK to concatenate these strings because it not really one string.  It is several separate
		// messages that we want to put into the same ARIA label.
		var isSelected = this.selectedRows[key];
		ariaRowLabel = isSelected ? dojo.string.substitute(this.messages.ariaSelected, [ariaRowLabel]) :
			dojo.string.substitute(this.messages.ariaNotSelected, [ariaRowLabel]);
		dijit.setWaiState(rowNode, "label", ariaRowLabel);
		
		// Also set the label on the checkbox cell if there is one.  This message includes additional information
		var checkboxCell = dojo.query("td."+com.ibm.rdm.web.grid.GridModel.COL_SELECTOR, rowNode)[0];
		if (checkboxCell) {
			var ariaSelectCheckboxLabel = ariaRowLabel + " " +
				(isSelected ? this.messages.ariaSelectedAction :  this.messages.ariaNotSelectedAction) + " " +
				this.messages.ariaNavigation;
			dijit.setWaiState(checkboxCell, "label", ariaSelectCheckboxLabel);
		}
	},
	
	_buildCell: function (rowTypeId, columnId, functionsToCall, contentRow, key, compIndent, indentToApply) {
		if (!compIndent) {
			compIndent = 0;
		}

		var widthVal;
		if (columnId === com.ibm.rdm.web.grid.GridModel.COL_MENU) {
			return this._buildMenuCell(rowTypeId, contentRow, functionsToCall);
		} else if (columnId === com.ibm.rdm.web.grid.GridModel.COL_ACTION) {
		} else if (columnId === com.ibm.rdm.web.grid.GridModel.COL_SELECTOR) {
			return this._buildSelectorCell(key, rowTypeId);
		} else if (columnId === com.ibm.rdm.web.grid.GridModel.COL_EXPANDO) {
			widthVal = (this.gridmodel.getWidth(rowTypeId, columnId));
			//An expando not on the primary column must grow in width as the indent grows
			var primary = this.gridmodel.getPrimaryRowTypeId();
			if (rowTypeId !== primary && !this.gridmodel.isChildRowTypeOf(rowTypeId, primary) && this.gridmodel.isExpandoResizable(rowTypeId)) {
				widthVal = widthVal + indentToApply;
			}
			
		} else {
			widthVal = (this.gridmodel.getWidth(rowTypeId, columnId)-compIndent);
		}
			
		
		if (widthVal < 5) {
			widthVal = 5;
		}
		
		var width = widthVal+"px";
		
		var isAutoWidth = this.gridmodel.isAutoWidth(rowTypeId, columnId);

		if (isAutoWidth) {
			width = "100%";
		}
		
		var colTd;
		/* [SW] - this could be generalised, here if we are building
		 *  an expando "cell" - don't actually build a TD but just a
		 *  <div> instead since we want this to actually go in the next
		 *  cell we build
		 */
		if (columnId === com.ibm.rdm.web.grid.GridModel.COL_EXPANDO) {
			colTd = document.createElement('div');
			width = "auto";
		}
		else {
			colTd = document.createElement('td');
			dijit.setWaiRole(colTd, "gridcell"); // ARIA
			var colDescribedByElementId = this.gridmodel.getColumnNameId(rowTypeId, columnId);
			if (colDescribedByElementId) {
				var describedByAttrVal = ARIA_ROW_DESCRIPTION_PREFIX+key + " " + colDescribedByElementId;
				dijit.setWaiState(colTd, "describedby", describedByAttrVal);
			}
		}
		
		//Add the column class
		dojo.addClass(colTd, columnId);
		dojo.style(colTd, "width", width);
		dojo.attr(colTd, "colid", columnId);
		
		//Mark as editable
		var contentIsEditableValueOrPromise = this.gridmodel.isEditable(rowTypeId, columnId, key);
		dojo.when(contentIsEditableValueOrPromise, function(contentIsEditable){
			if (contentIsEditable) {
				dojo.addClass(colTd, com.ibm.rdm.web.grid.Module.ClassNames.CELL_EDITABLE);
			}
		});
		//console.log("M _buildCell", rowTypeId, columnId, width, isAutoWidth);
		
		//Hover effect for addComment button
		if (columnId === com.ibm.rdm.web.artifactgrid.ArtifactConstants.COMMENTS_COL) {
				this.connect(colTd, "onmouseenter", dojo.hitch(this, function(event){
					var buttonNode = dojo.query('.addCommentButton', colTd)[0];
					if (buttonNode) {
						dojo.addClass(buttonNode, "show");
					}
				}));
				this.connect(colTd, "onmouseleave", dojo.hitch(this, function(event){
					var buttonNode = dojo.query('.addCommentButton', colTd)[0];
					if (buttonNode) {
						dojo.removeClass(buttonNode, "show");
					}
				}));
		
		}
		
		var cellClasses = this.gridmodel.getCellClasses(rowTypeId, columnId);
		if (cellClasses) {
			for (var k=0; k < cellClasses.length; k++) {
				dojo.addClass(colTd, cellClasses[k]);
			}
		}
		var temp = document.createElement('div');
				
		//Add the column class inside as well
		dojo.addClass(temp, columnId);
		if (cellClasses) {
			for (var k=0; k < cellClasses.length; k++) {
				dojo.addClass(temp, cellClasses[k]);
			}
		}
		if (this.gridmodel.mayHaveScrollBars(rowTypeId, columnId)) {
			dojo.style(temp, "overflow", "auto");
		} else {
			dojo.style(temp, "overflow", "hidden");
		}
		dojo.style(temp, "width", width);
		dojo.addClass(temp, "gridCellWrapper");
		dojo.place(temp, colTd, "only");
		
		var temp2 = document.createElement('div');
		dojo.place(temp2, temp, "only");
		
		var innerCellClasses = this.gridmodel.getInnerCellClasses(rowTypeId, columnId);
		if (innerCellClasses) {
			dojo.addClass(temp2, innerCellClasses);
		}
		
		var contents;
		var dcf = this.gridmodel.getDeferredContentFunction(rowTypeId, columnId);
		if (dcf) {
			//Caller has an asynchronous function, ie one that may require a trip to the server to resolve.
			//Therefore we put a marker div in the html, and then give the caller a function
			//which they can call with the actual result which will REPLACE this div (in fact, will
			//replace all the innerHTML of its parent).
			//We also give them the entire current row.
			
			//If we have some text to use as a placeholder, get it.
			var dpf = this.gridmodel.getDeferredPlaceholderFunction(rowTypeId, columnId);
			var placeholder = "";
			if (dpf) {
				placeholder = dpf(rowTypeId, columnId, contentRow);
			}
			
			var uniqueId = "UNIQ"+com.ibm.rdm.web.grid.Module.UNIQUE_ID++;
			this._idToKey[uniqueId]=key;
			contents= '<div id="'+uniqueId+'">'+placeholder+'</div>';
			//console.log("M _buildCell", uniqueId, rowTypeId, columnId, key);
		} else if (this.gridmodel.isVariable(rowTypeId, columnId)) {
			//We need an editable control
			var type = this.gridmodel.getVariableType(rowTypeId, columnId);
			if (type === com.ibm.rdm.web.grid.Grid.VARIABLE_CHECKBOX) {
				//Its a checkbox
				contents = document.createElement('input');
				dojo.attr(contents, "type", "checkbox");
				dojo.attr(contents, "focusable", true);
				var value = this._getValueFromRecord(rowTypeId, columnId, contentRow);
				var isDisabled = value.disabled;
				var isChecked = value.checked;
				contents.checked = isChecked;
				contents.disabled = isDisabled;
				//Work around for IE - checked is cleared when node is appended!
				var checkAgain = isChecked;
			} else if (type === com.ibm.rdm.web.grid.Grid.VARIABLE_NUMBERBOX) {
				contents = document.createElement('input');
				dojo.attr(contents, "type", "text");
				dojo.attr(contents, "focusable", true);
				var width = (this.gridmodel.getWidth(rowTypeId, columnId)-30)+"px";
				dojo.style(contents, "width", width);
				
				var value = this._getValueFromRecord(rowTypeId, columnId, contentRow);
				var isDisabled = value.disabled;
				var amount = value.value;
				contents.value = amount;
				contents.disabled = isDisabled;
				
				
				var self=this;
				var handle = dojo.connect(contents, "onkeyup", function(event) {self._textBox(event.currentTarget, rowTypeId, columnId, key);});
				this.dojoConnectList.addItem(handle);
				if (dojo.isIE > 6) {
					//Ensure text is selected on a click in IE
					var handle = dojo.connect(contents, "onclick", function(event) {
						event.currentTarget.focus();
						event.currentTarget.select();
					});
					this.dojoConnectList.addItem(handle);
				}
			}
		} else {
			contents = this._getValueFromRecord(rowTypeId, columnId, contentRow);
		}
		
		var contentsDiv;
		if (contents && dojo.isObject(contents)) {
			contentsDiv = contents;
		} else {
			contentsDiv = document.createElement('div');
			if (contents) {
				contentsDiv.innerHTML = contents;
				BidiUtils.applyTextDirSetting(contentsDiv, contents);
			}
		}

		if (dcf) {
			//now let the caller know - but call back in a little while
			var isFocusCell = this._isFocusCell(key, columnId);
			functionsToCall[functionsToCall.length] =  {
					func: dcf,
					setContentFn: dojo.hitch(this, this.applyDeferredContent, contentsDiv,  this.domNode, uniqueId, contentIsEditableValueOrPromise),
					row: contentRow,
					key: key,
					/* The node that will contain the contents passed back in function supplied to
					 * the deferred callback function ("functionToCall").  This is required by the
					 * Rich Text Editor so that it can give itself an initial height based on the
					 * height of the cell it is going into - so the tables doesn't appear to collapse
					 * and then expand again.  I suppose we could just pass the dimensions as these
					 * are all that are required.
					 */
					containingNode: contentsDiv,
					indent: indentToApply,
					columnId: columnId,
					isFocusCell: isFocusCell
				};
				
		}
		
		dojo.place(contentsDiv, temp2, "only");
		
		if(checkAgain && dojo.isIE) {
			contents.checked = true;
		}
		
		
		
		if( (columnId.substring(0, 2) === "lk") && this.dragDropArgs ) {
			this.makeLinkColDropTarget(colTd);
		}
		
		
		return colTd;
	},
	
	_textBox: function (target, rowTypeId, columnId, key) {
		var value = target.value;
		this.gridmodel.setValue(rowTypeId, columnId, key, value);
	},
	
	applyDeferredContent: function(cellNode, mainNode, uniqueId, contentIsEditableValueOrPromise, content, destroyObject) {
		// summary:
		//		When a deferred function is used, it calls back on this function.
		// uniqueId:
		//		The id of the node (this is added by the module code, rather than the deferred function caller).
		// mainNode:
		//		The parent div node - added by the module code.
		// content:
		//		The content for the cell with this id. This is added by the caller of the module.
		// destroyObject:
		//		If provided, this object will have 'destroyRecursive' called on it when the row is destroyed.

		//console.log("M applyDeferredContent", id, this._idToKey[id], mainNode, content, !destroyObject, destroyObject);
		//var divToReplace = dojo.byId(id);

		var divToReplace = cellNode,
			editConnection;//dojo.byId(id);
		if (divToReplace && divToReplace.parentNode) {
			if (content === null || content === undefined) {
				content = dojo.doc.createTextNode("");
			} else if (dojo.isString(content)) {
				var text = content;
				content = dojo.doc.createTextNode(content);
				//bidi
				BidiUtils.applyTextDirSetting(divToReplace.parentNode, text, true);
			}
			var parent = divToReplace.parentNode;
			dojo.destroy(divToReplace);
			dojo.place(content, parent);
			dojo.when(contentIsEditableValueOrPromise, dojo.hitch(this, function(contentIsEditable){
				if (contentIsEditable) {
					var editDiv = dojo.create("a", {
						"class": "editButton rdm-button-link",
						"title": this.messages.edit
					}, parent, "first");
					dojo.create("span", {
						className: "editDescriptionCommand"
					}, editDiv);
					editConnection = dojo.connect(editDiv, "onclick", this, function(event){
						this._startEdit(null, content);
					});
				}
			}));
		} else {
			console.error("Could Not find id", uniqueId, content);
		}

		var key = this._idToKey[uniqueId];
		var row = this.allRows[key];
		
		if (destroyObject) {
			if (row) {
				row.destroy.push(destroyObject);
			}
			delete this._idToKey[uniqueId];
		}
		if (row) {
			if (editConnection){
				row.connections.push(editConnection);
			}
		}
		
	},

	_buildMenuCell: function(rowTypeId, resource, functionsToCall) {
		var widthVal = this.gridmodel.getWidth(rowTypeId, com.ibm.rdm.web.grid.GridModel.COL_MENU);
		var width = widthVal+"px";
		//Menu Cells are presumed not to be 'isAutoWidth'

		var colTd = document.createElement('td');
		dojo.addClass(colTd, com.ibm.rdm.web.grid.GridModel.COL_MENU);
		dojo.style(colTd, "width", width);
		
		
		dojo.attr(colTd, "colid", com.ibm.rdm.web.grid.GridModel.COL_MENU);
		
		var temp = document.createElement('div');
		dojo.addClass(temp, com.ibm.rdm.web.grid.GridModel.COL_MENU);
		dojo.style(temp, "width", width);
		dojo.place(temp, colTd, "only");
		
		var temp2 = document.createElement('img');
		dojo.attr(temp2, "src", this._blankGif);
		dojo.attr(temp2, "focusable", "true");
		dojo.attr(temp2, "menuType", "1");
		dojo.addClass(temp2, "menuLauncher");
		dojo.place(temp2, temp, "last");
		
		// place ARIA labels on the menu cell for audio description
		dijit.setWaiRole(colTd, "gridcell");

		if (this.gridmodel.getMenuCellLabelFunction(rowTypeId)) {
			var ariaLabelUpdateDescriptor =  {
				func: dojo.hitch(this, this._updateAriaMenuCellLabel, rowTypeId),
				setContentFn: function(){},
				row: resource,
				containingNode: colTd
			};
			functionsToCall.push(ariaLabelUpdateDescriptor);
		}
		
		return colTd;
	},
	
	_updateAriaMenuCellLabel: function (rowTypeId, functionToCall, resource, key, containingNode) {
		// summary:
		//		Populate the aria-label attribute on a menu cell, so giving the cell a label that
		//		describes the cell.
		var cellLabelFunction = this.gridmodel.getMenuCellLabelFunction(rowTypeId);
		if (!cellLabelFunction) {
			return; // should never get here - callback should not have been set if no such function.
		}
		var label = cellLabelFunction(resource);
		dijit.setWaiState(containingNode, "label", label);
		functionToCall();
	},
	
	_buildSelectorCell: function(key, rowTypeId) {
		
		var widthVal = this.gridmodel.getWidth(rowTypeId, com.ibm.rdm.web.grid.GridModel.COL_SELECTOR);
		var width = widthVal+"px";

		var colTd = document.createElement('td');
		dojo.addClass(colTd, com.ibm.rdm.web.grid.GridModel.COL_SELECTOR);
		dojo.style(colTd, "width", width);
		
		
		dojo.attr(colTd, "colid", com.ibm.rdm.web.grid.GridModel.COL_SELECTOR);
		
		var temp = document.createElement('div');
		dojo.addClass(temp, com.ibm.rdm.web.grid.GridModel.COL_SELECTOR);
		dojo.style(temp, "width", width);
		dojo.place(temp, colTd, "only");
			
		var temp2 = document.createElement('input');
		dojo.attr(temp2, "type", "checkbox");
		dojo.attr(temp2, "focusable", true);
		dojo.attr(temp2, "tabIndex", "-1");
		dojo.addClass(temp2, com.ibm.rdm.web.grid.Module.ClassNames.MAIN_SELECT);
		this.IE_reselect_hack = null;
		if (this.selectedRows[key]) {
			temp2.checked = true;
			//IE hack so that this can be reapplied!
			this.IE_reselect_hack = temp2;
		}
		
		if (this.dragDropArgs && this.dragDropArgs.dragSource && this.dragDropArgs.dragHandleInSelector) {
			
			var dragHandle = dojo.create("div", {
				"class": "dragHandle"
			}, temp);
			
			this.connect(dragHandle, "mousedown", function() {
				var nl = new dojo.NodeList(dragHandle);
				var parent = nl.parents("table")[0];
				this.addToSelection([dojo.attr(parent, "objectid")]);
			});
			
				
			var dragSource = new com.ibm.rdm.web.common.dnd.DragSource({
				getDragImage: this.dragDropArgs.getDragImage,
				node: dragHandle,
				types: this.dragDropArgs.dragTypes,
				onDragStart: dojo.hitch(this, function(args) {
					if (typeof this.dragDropArgs.onDragStart === "function") {
						this.dragDropArgs.onDragStart(args);
					}
					this._dragInProgress = true;
				}),
				onDragEnd: dojo.hitch(this, function() {
					this._dragInProgress = false;
				})
			});
			
			this._toDestroy.push(dragSource);
		}
		
		temp.appendChild(temp2);
		
		return colTd;
	},
	
	_reorderColsInRow: function(rowTypeId, row, columnIds) {
		//Get key
		var key = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		
		//Build list of existing column cells
		var cols = {},
			newCols = [];
		var self = this;
		dojo.forEach(row.rows[0].cells, function(item) {
			var colId = dojo.attr(item, "colId");
			if (colId) {
				cols[colId] = item;
			}
		});
		
		//work out and apply indent
		var indent = this.gridmodel.getIndentMultiple(rowTypeId, key);
		var indentToApply = (indent*INDENT_CHUNK);
				
		var functionsToCall = [];
		var currCol = 0;
		var compIndent = 0;
		var expandoCell;
		dojo.forEach(columnIds, function(columnId, index) {
			var isVisible = self.gridmodel.isVisible(rowTypeId, columnId);
			if (isVisible) {
				
				//The first (non selection) col width is increased to provide the indent
				if (index == 1) {
					compIndent = -(indent*INDENT_CHUNK);
				} else {
					compIndent = 0;
				}
				//Compensate the indent from the first non fixed width cell.
				if (!self.gridmodel.isFixedWidth(rowTypeId, columnId)) {
					compIndent = (indent*INDENT_CHUNK);
					indent = 0;
				}

				
				var col = cols[columnId];
				
	
				//Recalculate col if there is no existing col OR a compensating indent needs to be applied
				// [SW] - again specific checking for the expando column as that one is merged. could be
				// generalised
				if ((!col || compIndent > 0) && columnId !== com.ibm.rdm.web.grid.GridModel.COL_EXPANDO) {
					var contentRow = self.gridmodel.getRowData(rowTypeId, key);
					col = self._buildCell(rowTypeId, columnId, functionsToCall, contentRow, key, compIndent);
					newCols.push(col);
				}

				//The first (non selection) col width is increased to provide the indent
				if (columnId != "_selector" && indentToApply > 0) {
					dojo.style(col,  "paddingLeft", ""+(indentToApply+2)+"px");
					indentToApply = 0;
				}
				

				if (col && dojo.hasClass(col, "expandoContainer")) {
					var width = self.gridmodel.getWidth(rowTypeId, columnId),
						toWiden = dojo.query("."+columnId, col);
					
					toWiden.push(col);
					toWiden.style("width", width+"px");
				}
			
				
				//For IE, we need to reapply the selection checkbox... sigh.
				if ((dojo.isIE > 0) && self.IE_reselect_hack) {
					self.IE_reselect_hack.checked = true;
				}
				
			}
		});

		if (newCols.length > 0) {
			//Now rebuild
			row.insertRow(0);
			dojo.forEach(newCols, function (col) {
				row.rows[0].appendChild(col);
			});
			//And remove the original
			row.deleteRow(1);
		}
		
		//And mark up the new one
		//Now, lets makes sure we get the deferred values we want
		this._callDeferredFunctions(functionsToCall);
	},
	
	refreshRows: function(rowkeys, /*Boolean?*/ noFocus, deferDestroy) {
		// summary:
		//		Takes the given node details and refreshes them
		//		in the grid. If the node is not found, it is ignored.
		// rowkeys:
		//		An array of {key:... rowtype: ... } entries.
		// noFocus:
		//		If set and true, no attempt is made to restore focus to the grid
		//		following the refresh.
		// deferDestroy:
		//		If set and true, old rows are not destroyed and are put into a bucket.
		//		Solves an issue that can occur in IE when editing multiple attributes on the same row.

		//console.log("M refreshRows", rowkeys);

		//Place to hold any deferred population calls
		var functionsToCall = [];
		
		var currentFocusElement = document.activeElement;
		var focusRow;		// Set to a new row if the old row had focus
		
		if (dojo.isArray(rowkeys)) {
			dojo.forEach(rowkeys, dojo.hitch(this, function(item, index) {

				if (this.allRows && this.allRows[item.key]) {
					var oldNode = this.allRows[item.key];
					var node = oldNode.node;
					//Refresh Node
					var contentRow = this.gridmodel.getRowData(item.rowtype, (item.newKey ? item.newKey: item.key));
					var newNode = this._createRow(item.rowtype, (item.newKey ? item.newKey: item.key), contentRow, functionsToCall);
					var newConnections = this.prepareNodeForInsert(newNode);
					
					var allNodeRecord = {node: newNode, rowtype: item.rowtype, destroy: [], connections: newConnections};
					
					//Keep the version if necessary
					var version = this.gridmodel.getVersion(item.rowtype, item.key);
					if (version) {
						allNodeRecord.version = version;
					}
					if (item.newKey){
						delete this.allRows[item.key];
						if (this._navigationTabTarget.rowKey === item.key){
							this._navigationTabTarget.rowKey = item.newKey;
						}
					}
					this.allRows[(item.newKey ? item.newKey : item.key)] = allNodeRecord;
					
					// before removing existing row remember if we have to restore focus
					if (node === currentFocusElement || dojo.isDescendant(currentFocusElement, node)) {
						focusRow = newNode;
					}
					this._doDisconnects(oldNode.connections);
					//Call destroyRecursive on all the replaced rows
					this._callDestroys(oldNode.destroy);
					//Replace the nodes row
					dojo.place(newNode, node, "replace");
					if(deferDestroy){
						this._nodeBucket.push(node);
					}
					else{
						dojo.destroy(node);
						this._clearNodeBucket();
					}
										
					//For IE, we need to reapply the selection checkbox... sigh.
					if ((dojo.isIE > 0) && this.IE_reselect_hack) {
						this.IE_reselect_hack.checked = true;
					}
				}

			}));
			
			//Now, lets make sure we get the deferred values we want
			var updatesCompleteDeferred = this._callDeferredFunctions(functionsToCall);
			
			// restore focus if it was on a row we have refreshed.
			updatesCompleteDeferred.addCallback(this, function(){
				return this._selectObject(focusRow);
			});
			if (!noFocus) {
				updatesCompleteDeferred.addCallback(this, function() {
					if (!this._wasInEditModeWhenOpenEditMenuWasRequested) {
						// Retake focus after a short delay, because on IE something happens later in the refresh
						// that takes focus away from an input control back to the document body.
						window.setTimeout(dojo.hitch(this, function(){
							this.retakeFocus();
						}), 100);
					}
					if (this._wasInStubRowWhenOpenEditMenuWasRequested) {
						var rowInfo = this._getCurrentRowInfo();
						if (!this._isStubRow(rowInfo.node)) {
							// Note: To address Defect 85588 we've deferred opening the
							// edit menu until the stub row has successfully saved ...
							// NOW we can open edit menu (as the row is now associated
							// with a real resource URI).
							this._onOpenMenuKey();
						}
					}
				});
			}
			return updatesCompleteDeferred;
		}
	},
	
	runRowClassFunctions: function (rowKeys) {

		var rowClassFunctions = this.gridmodel.getRowClassFunctions(this.gridmodel.getPrimaryRowTypeId());
		for (var i =0; i<rowKeys.length; i++) {
			var key = rowKeys[i];
			if (rowClassFunctions && this.allRows[key]) {
				// make sure we remove any existing lock classes, otherwise we might end up
				// in a situation where we have both (it was locked by another and now is
				// locked by me for instance)
				dojo.removeClass(this.allRows[key].node, ["permLockedByMe", "permLockByOther"]);
				for (var j=0; j < rowClassFunctions.length; j++) {
					var rowClasses = rowClassFunctions[j](key);
					if (rowClasses) {
						for (var k=0; k < rowClasses.length; k++) {
							dojo.addClass(this.allRows[key].node, rowClasses[k]);
						}
					}
				}
			}
		}
	},
	
	refreshExpandos: function(rowkeys) {
		// summary:
		//		Takes the given node details and refreshes their expando cells
		//		in the grid. If the node is not found, it is ignored.
		// rowkeys:
		//		An array of {key:... rowtype: ... } entries.
			
		//Place to hold any deferred population calls
		var functionsToCall = [];
		var self=this;

		if (dojo.isArray(rowkeys)) {
			dojo.forEach(rowkeys, function(item) {

				if (self.allRows && self.allRows[item.key]) {
					var node = self.allRows[item.key].node;
					
					//Get the right col - if it exists
					// TODO: formalise this...
					var col = dojo.query("div."+com.ibm.rdm.web.grid.GridModel.COL_EXPANDO, node);
					if (col[0]) {
						//Refresh the expando cell
						var contentRow = self.gridmodel.getRowData(item.rowtype, item.key);
						var indent = self.gridmodel.getIndentMultiple(item.rowtype, item.key);
						var cellIndentToApply = (indent.cell*INDENT_CHUNK);
						var newCell = self._buildCell(item.rowtype, com.ibm.rdm.web.grid.GridModel.COL_EXPANDO, functionsToCall, contentRow, item.key,
														0, cellIndentToApply, false);
						//preserve the padding
						var paddingLeft = dojo.style(col[0], "paddingLeft")+"px";
						dojo.style(newCell, "paddingLeft", paddingLeft);
						//Replace the expando cell, if we have one
						col[0].parentNode.replaceChild(newCell, col[0]);

					}
				}

			});
			
			// Now, lets makes sure we get the deferred values we want
			// There is no need to set a timeout before calling the functions
			// to render the expando.  Infact, doing so cause the display to
			// flicker.
			
			self._callDeferredFunctions(functionsToCall, true);
		}
	},
	
	reapplyCols: function(rowTypeId) {
	
		//This reorders and changes the width of the cols with this RowType.
		var columnIds = this.gridmodel.getColumnIds(rowTypeId);
		var self = this;
		
		//Get all the children of this rowtype
		var childRowTypeIds = this.gridmodel.getChildRowTypeIds(rowTypeId);
		
		//Fix the original rows
		dojo.query("table."+rowTypeId, this.domNode).forEach(function(row) {
			self._reorderColsInRow(rowTypeId, row, columnIds);
		});
		
		//And now the child rows
		for (var t=0; t < childRowTypeIds.length; t++) {
			dojo.query("table."+childRowTypeIds[t], this.domNode).forEach(function(row) {
				self._reorderColsInRow(childRowTypeIds[t], row, columnIds);
			});
		}
		
	},
	
	setGridmodel: function(gridmodel) {
		this.gridmodel = gridmodel;
	},

	startup: function () {
	},
	
	_startEdit: function(event, node) {
		//Remove all Selected Rows
		var uris = [];
		for (var x in this.selectedRows) {
			uris.push(this.selectedRows[x].uri);
		}
		this.removeFromSelection(uris);
		
		node = node || event.target;
		
		var ids = this.getRowAndColumnIds(node);
		var columnId = ids.columnId;
		var rowTypeId = ids.rowTypeId;
		
		if (rowTypeId && columnId) {
			if (dojo.hasClass(ids.colNode, com.ibm.rdm.web.grid.Module.ClassNames.CELL_EDITABLE)) {
				var objectId = dojo.attr(ids.rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
				this._moveGridFocus(objectId, rowTypeId, columnId);
				this._onStartActionModeKey();
			}
		}
		if (event) {
			dojo.stopEvent(event);
		}
	},
	
	_handleDoubleClick: function(event) {
		DomUtils.clearSelection();
		if (!this.gridmodel.onDoubleClick()) {
			// if double click event was not handled externally, call _startEdit
			this._startEdit(event);
		} else {
			dojo.stopEvent(event);
		}
	},
	
	getSelectedObjectId: function () {
		if (this.selectedObject !== null) {
			return dojo.attr(this.selectedObject, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		}
	},
	
	getSelectedRows: function(/*?Boolean*/ includeFocusRow){
		// summary:
		//		Return an array of all selected rows.
		// includeFocusRow:
		//		If present and true, the focus row (or the row containing the focus cell)
		//		is included in the result, regardless of whether it is selected or not.
		// returns:
		//		An array, in which each entry is an object with uri and rowTypeId fields,
		//		to identify the rows.
		var currentSelectedRows =[],
			focusRowInfo,
			focusRowId,
			selectedRow;
		if (includeFocusRow) {
			focusRowInfo = this._getCurrentRowInfo();
			if (focusRowInfo) {
				focusRowId = focusRowInfo.key;
			}
		}
		for (var x in this.selectedRows) {
			selectedRow = this.selectedRows[x];
			currentSelectedRows.push(selectedRow);
			if (focusRowId && (focusRowId === selectedRow.uri)) {
				// the focus row is one of the selected ones we already
				// have, so avoid adding it a second time.
				includeFocusRow = false;
			}
		}
		if (includeFocusRow && focusRowInfo) {
			// focus row has been requested, and we have not already added it.
			currentSelectedRows.push({uri: focusRowId, rowTypeId: focusRowInfo.type});
		}
		return currentSelectedRows;

	},
	
	getFocusRow: function() {
		// summary:
		//		Gets identity for the row containing the current focus.
		// returns:
		//		An object with key and rowType entries for the focus row,
		//		or undefined if there is no identifiable focus row.
		var focusRowInfo = this._getCurrentRowInfo();
		if (focusRowInfo) {
			return {key: focusRowInfo.key, rowType: focusRowInfo.type};
		}
	},

	getFocusRowIfRowWasRemoved: function(rowId) {
		// summary:
		//		Determine which row should get focus if the row represented
		//		by rowId was removed.
		// returns:
		//		If a sensible "new row to use for focus" can be found an
		//		an object with 'key' and 'rowType' will be returned;
		//		otherwise null will be returned.

		var rowNode = dojo.query("[objectId='" + rowId + "']", this.domNode)[0];
		if (rowNode) {
			var newFocusRowNode = (
				this._findNextSiblingObject(rowNode)
				|| this._findPreviousSiblingObject(rowNode)
			);
			if (newFocusRowNode) {
				return {
					key: dojo.attr(newFocusRowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID),
					rowType: dojo.attr(newFocusRowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID)
				};
			}
		} else {
			return null;
		}
	},

	setHighlightFunction: function (highlightClass, highlightFunction) {
		if (this._highlightFunctions[highlightClass] !== highlightFunction) {
			this._highlightFunctions[highlightClass] = highlightFunction;
			this._reapplyHighlightFunctions();
		}
		if (highlightFunction === null) {
			delete this._highlightFunctions[highlightClass];
		}
	},
	
	_applyHighlightClasses: function(rowTypeId, key, item) {
		// summary:
		//		Applies the highlight classes to a node.
	
		for (var className in this._highlightFunctions) {
			//See if we need to add any
			if (this._highlightFunctions[className] && this._highlightFunctions[className](key, rowTypeId)) {
				dojo.addClass(item, className);
			} else {
				dojo.removeClass(item, className);
			}
		}
	},
	
	_reapplyHighlightFunctions: function() {
		//simple reapply of the function
		var self = this;
	
		dojo.query("["+com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID+"]", this.domNode).forEach(function (item) {
			//iterate over all, removing existing and putting in new.
			var id = dojo.attr(item, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
			var rowTypeId = dojo.attr(item, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			self._applyHighlightClasses(rowTypeId, id, item);
		});
	},

	postCreate: function () {
		this._createContentNode();
		// hookup top level event listeners once.
		this._initializeKeyPressHandlers();
		this._hookupEventListeners();
		this._initializeFocusCatcher();
	},

	/* Create the content div that will contain the HTML
	* returned by the server.
	*/
	_createContentNode: function () {
		var content = document.createElement("div");
		//dojo.style(content, "margin", "1px");
		this.content = content;
		this.domNode.appendChild(this.content);
	},
	
	_hookupEventListeners: function () {
		if (!this.skipHandlerHookup){
			this.connect(this.domNode, "onclick", this._handleClick);
			this.connect(this.domNode, "onkeypress", dojo.hitch(this, this._handleKeyPress, false));
			this.connect(this.domNode, "onkeydown", dojo.hitch(this, this._handleKeyPress, true));
			this.connect(this.domNode, "ondblclick", this._handleDoubleClick);
			this._connectHandleKeyDown();
		}
	},
	
	_initializeFocusCatcher: function() {
		// summary:
		//		Adds the focus catcher node and hooks up to events on it so that the
		//		grid can manage its own focus.  This must only be done if we are configured
		//		to manage focus, rather than relying on the browser default tabbing and
		//		clicking behaviour for focus on cell content.
		if (this.manageFocusNavigation) {
			// Create a node that is used to catch focus
			this._focusCatcherNode = dojo.create("div", {
				tabIndex: "0",
				width: "0px",
				height: "0px",
				"class": "gridFocusCatcher"
			}, this.domNode, "after");
			this.connect(this._focusCatcherNode, "onfocus", this._onGridFocus);
		}
	},

	_handleMouseOverRow: function (event, node) {
		// summary:
		//		React to mouse over events.
		if (this._dragInProgress) {
			return;
		}
		
		var gridmodel = this.gridmodel;
		dojo.addClass(node, com.ibm.rdm.web.grid.Module.ClassNames.ROW_HOVER);
		
		// Also add a class to the menu launcher.  For some reason Chrome does not reliably remove
		// the displayed menu launcher icon when the rowHover class is removed from an ancestor DOM
		// node of the one with the image.  Other browsers seem fine, so this looks like it may be
		// a defect in Chrome's processing of CSS rules (or at least a difference from other browsers).
		// So we add/remove a class directly on the node that has the image, which does seem to work.
		// See work item 59928.
		dojo.query(".menuLauncher", node).forEach(function(menuLauncherNode){
			dojo.addClass(menuLauncherNode, com.ibm.rdm.web.grid.Module.ClassNames.ROW_HOVER);
		});
		
		// // Get the current row and inform the model about mouseOver event.  Grid may
		// // be configured to be notified about mouse over row events.
		node = node || event.target;
		var ids = this.getRowAndColumnIds(node);
		var objectId = dojo.attr(ids.rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		var rowTypeId = dojo.attr(node, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
		
		gridmodel.mouseOverRow(rowTypeId, objectId);
	},

	// Hide edit buttons when not hovering over row.
	_handleMouseOutRow: function (event, node) {
		// summary:
		//		React to mouse out events.
		var gridmodel = this.gridmodel;
		dojo.removeClass(node, com.ibm.rdm.web.grid.Module.ClassNames.ROW_HOVER);

		// Also remove a class from the menu launcher.  For some reason Chrome does not reliably remove
		// the displayed menu launcher icon when the rowHover class is removed from an ancestor DOM
		// node of the one with the image.  Other browsers seem fine, so this looks like it may be
		// a defect in Chrome's processing of CSS rules (or at least a difference from other browsers).
		// So we add/remove a class directly on the node that has the image, which does seem to work.
		// See work item 59928.
		dojo.query(".menuLauncher", node).forEach(function(menuLauncherNode){
			dojo.removeClass(menuLauncherNode, com.ibm.rdm.web.grid.Module.ClassNames.ROW_HOVER);
		});

		// Get the current row and inform the model about mouseOut event.  Grid may
		// be configured to be notified about mouse out row events.
		node = node || event.target;
		var ids = this.getRowAndColumnIds(node);
		var objectId = dojo.attr(ids.rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		var rowTypeId = dojo.attr(node, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
		gridmodel.mouseOutRow(rowTypeId, objectId);
	},
	
	_handleCellOnFocus: function(evt, rowNode) {
		// Only called when this.doTrackFocusForCells is true
		dojo.addClass(rowNode, com.ibm.rdm.web.grid.Module.ClassNames.ROW_FOCUS);
	},
	
	_handleCellOnBlur: function(evt, rowNode) {
		// Only called when this.doTrackFocusForCells is true
		dojo.removeClass(rowNode, com.ibm.rdm.web.grid.Module.ClassNames.ROW_FOCUS);
	},

	// Keyboard navigation functions
	_handleKeyPress: function (isKeyDownEvent, event) {
		//handling of action menu key navigation, if menu is visible.
		var editMenuQueryResult = dojo.query(".editmenu");
		
		var consume = false;
		

		if (editMenuQueryResult.length > 0) {
			if (!isKeyDownEvent) {
				// keypress event
				var editMenuNode = editMenuQueryResult[0];
				var editMenuWidget = dijit.getEnclosingWidget(editMenuNode);
				editMenuWidget.handleKeyPress(event.keyCode);
			}
			dojo.stopEvent(event);
			return;
		}

		// Handle keypress events.  There can be two sets of handlers.  First, those
		// that are defind internally, for built-in grid keyboard functionality.  Second,
		// keypress handlers that are contributed through a keyPressHandler method by our
		// caller.
		consume = !this._internalKeyEventHandler(event, isKeyDownEvent);

		if (!consume && this.keyPressHandler && !isKeyDownEvent) {
			consume = !this.keyPressHandler(event);
		}
		if (!consume && this.keyDownHandler && isKeyDownEvent) {
			consume = !this.keyDownHandler(event);
		}

		if (consume) {
			//console.debug("Module._handleKeyPress stopping event:", event);
			dojo.stopEvent(event);
		}
	},
	
	_connectHandleKeyDown: function() {
		// summary:
		//		If required, connects a handler for the keydown event.  This is not used to provide any real
		//		functionality, but as a workaround for some browser-specific "features".
		var internalKeyDownHandler;
		if (dojo.isFF) {
			// We use shift-F10 to open row menu, but on FF this also opens the browser
			// context menu.  We can prevent this by cancelling the onkeydown event
			// for that key combination.
			internalKeyDownHandler = com.ibm.rdm.web.util.keyutils.getKeyHandler({
				keyCode: dojo.keys.F10,
				shift: true,
				onKeyPress: function(event){
					/* do nothing, but event is cancelled */
				}
			});
		}
		
		// We can also be constructed with a keyDownHandler function.  If that is defined we must
		// call it for keyDown events.
		if (this.keyDownHandler || internalKeyDownHandler) {
			// We have at least one function that wants to handle keyDown events.
			this.connect(this.domNode, "onkeydown", function(event){
				var consumed;
				if (internalKeyDownHandler) {
					consumed = internalKeyDownHandler(event);
					// This handler stops the event internally, so no need to do that here.
				}
				if (!consumed && this.keyDownHandler) {
					consumed = !this.keyDownHandler(event);
					if (consumed) {
						dojo.stopEvent(event);
					}
				}
			});
		}
	},
	
	_findNextSiblingObject: function (fromWhere) {
		var node = fromWhere.nextSibling;
		return this._findNextOrPrevious(node, "nextSibling", fromWhere.parentElement || fromWhere.parentNode );
	},

	_findNextOrPrevious: function (node, nextOrPrevious, parent) {
		do {
			if (node) {
				if (dojo.hasAttr(node, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID)) {
					return node;
				} else {
					node = node[""+nextOrPrevious+""];
				}
			}
			
			if (!node) {
				//Check parent
				parent = parent[""+nextOrPrevious+""];
				if (parent) {
					if (nextOrPrevious == "nextSibling") {
						node = parent.firstChild;
					} else if (nextOrPrevious == "previousSibling") {
						node = parent.lastChild;
					}
				} else {
					return null;
				}
			}
			
		} while (node);
	},
	
	_findPreviousSiblingObject: function (fromWhere) {
		var node = fromWhere.previousSibling;
		return this._findNextOrPrevious(node, "previousSibling", fromWhere.parentElement || fromWhere.parentNode);
	},
	
	addToSelection: function(keys , /*optional*/ disableOnChange) {
		var informOnChange = typeof(disableOnChange) != "undefined" ? disableOnChange : true;
		var selectionChanged = false;
		var self=this;
		dojo.forEach(keys, function(key) {
			if (!self.selectedRows[key]) {
				var row = self.allRows[key];
				if (row) {
					var rowTypeId = dojo.attr(row.node, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
					if (self.gridmodel.isSelectable(rowTypeId)) {
						self.selectedRows[key] = {rowTypeId: rowTypeId, uri: self.gridmodel.mapKeyToUri(rowTypeId, key)};
						self.applyCheckRow(row.node, key, rowTypeId);
						selectionChanged = true;
					}
				}
			}
		});
		
		if (selectionChanged && informOnChange) {
			this._informSelectionChanged();
		}
	},
	
	removeFromSelection: function(keys) {
		var selectionChanged = false;
		var self=this;
		dojo.forEach(keys, function(key) {
			if (self.selectedRows[key]) {
				delete self.selectedRows[key];
				var row = self.allRows[key];
				if (row) {
					self.unapplyCheckRow(row.node, key);
				}
				selectionChanged = true;
			}
		});
		if (selectionChanged) {
			this._informSelectionChanged();
		}
	},


	applyCheckRow: function(row, objectId, rowTypeId, skipRowSelectableCheck) {
		if (!this.gridmodel.allowsMultiSelect(rowTypeId)) {
			//No multi select, so ensure only this row selected.
			this._selectAll(false);
		}

		if (skipRowSelectableCheck || this.gridmodel.isSelectable(rowTypeId) || dojo.hasClass(row, "underEdit")) {
			dojo.query("."+com.ibm.rdm.web.grid.Module.ClassNames.MAIN_SELECT, row).forEach(function (item) {item.checked = true; });
			this.selectedRows[objectId]={rowTypeId: rowTypeId, uri: this.gridmodel.mapKeyToUri(rowTypeId, objectId)};
			dojo.addClass(row, com.ibm.rdm.web.grid.Module.ClassNames.ROW_CHECKED);
			this._updateAriaSelectionLabels(objectId, row);
		}
	},
	
	unapplyCheckRow: function(row, objectId) {
		dojo.query("."+com.ibm.rdm.web.grid.Module.ClassNames.MAIN_SELECT, row).forEach(function (item) {item.checked = false; });
		delete this.selectedRows[objectId];
		dojo.removeClass(row, com.ibm.rdm.web.grid.Module.ClassNames.ROW_CHECKED);
		this._updateAriaSelectionLabels(objectId, row);
		this.setHeaderCheckbox(false);
	},
	
	onCheck: function(event, row, objectId, rowTypeId, columnId, checked) {
		// summary:
		//		A row check mark has been selected.

		row.focus();
		window.setTimeout(
			dojo.hitch(this, function(){
				if (checked) {
					//Deal with it, handling the shift keys
					this._selectObjectByEvent(event, row, rowTypeId, columnId, true);
					
				} else {
					this.unapplyCheckRow(row, objectId, rowTypeId);
				}
				if (!checked) {
					this._informSelectionChanged();
				}
			}),
			1
		);
	},
	
	setHeaderCheckbox: function(state) {
		// summary:
		//		Called to set or clear the header checkbox.  Our container connects
		//		to this method to pass the event to the header.
		// state:
		//		true to set the checkbox.  false to clear it.
	},
	
	selectAll: function(onOrOff) {
		this._selectAll(onOrOff);
		
		this._informSelectionChanged();
	},
	
	_selectAll: function(onOrOff) {
		// summary:
		//		Allows the selecting or deselecting of all the (selectable) rows in the grid.
		// onOrOff:
		//		true if they are all to be switched on, false if all are to be switched off
		dojo.query("."+com.ibm.rdm.web.grid.Module.ClassNames.MAIN_SELECT, this.domNode).forEach(function (item) {item.checked = onOrOff; });
		var that = this;
		if (onOrOff) {
			dojo.query("["+com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID+"]", this.domNode).forEach(function (item) {
				var objectId = dojo.attr(item, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
				var rowTypeId = dojo.attr(item, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
				if (that.gridmodel.isSelectable(rowTypeId)) {
					that.selectedRows[objectId]={rowTypeId: rowTypeId, uri: that.gridmodel.mapKeyToUri(rowTypeId, objectId)};
					dojo.addClass(item, com.ibm.rdm.web.grid.Module.ClassNames.ROW_CHECKED);
					that._updateAriaSelectionLabels(objectId, item);
				}
			});
		} else {
			that.selectedRows = {};
			dojo.query("["+com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID+"]", this.domNode).forEach(function (item) {
				dojo.removeClass(item, com.ibm.rdm.web.grid.Module.ClassNames.ROW_CHECKED);
				var objectId = dojo.attr(item, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
				that._updateAriaSelectionLabels(objectId, item);
			});
		}
		this.setHeaderCheckbox(onOrOff);
	},
	
	notifyEditStartedOnCell: function(rowKey, columnId) {
		// summary:
		//		Tells the grid that a cell within it is going into
		//		edit mode.  Moves focus to that cell and puts the
		//		row into actionable mode.
		//		Does not, in itself, start any editors on the
		//		row.  May be used by a widget contained in the row
		//		to indicate that it has internally switched to
		//		edit mode.
		// rowKey:
		//		Identifier for the row that is being edited.
		// columnId:
		//		Identifies the cell within the row.
		
		// If this is a move to a different row, save any rows that are
		// in edit.
		var rowInfo = this._getCurrentRowInfo();
		if (!rowInfo || rowKey !== rowInfo.key) {
			// If we have no rowInfo, we have never had focus in the grid, so there should not be any
			// other rows under edit.  But go through the save procedure anyway, just in case.  Better to
			// be safe and not risk losing any unsaved data.
			this.onSaveKey();
		}
		// move to the new cell.
		this._moveGridFocus(rowKey, null, columnId, /*actionMode*/true);
	},
	
	isInEditMode: function() {
		// returns:
		//		true if the grid is currently in edit mode (Actionable Mode) on a row.
		var rowInfo = this._getCurrentRowInfo();
		return (rowInfo && rowInfo.actionMode);
	},
	
	onEditorSaveKey: function() {
		// summary:
		//		Called when a save operation is invoked (typically by a hotkey) from a contained editor, or a widget that
		//		can contain an editor.  Notifies the grid that outstanding grid edits should be saved.
		this._saveAllRows();
	},
	
	onEditorCancelKey: function() {
		// summary:
		//		Called when a edit cancel operation is invoked  (typically by a hotkey) from a contained editor, or a widget that
		//		can contain an editor.  Notifies the grid that outstanding grid edits should be saved.
		this._cancelAllEdits();
	},


	/* public */
	clear: function() {
		if (this.domNode) {
			this._clearNodeBucket();
			this._clearHandles();
			this._removeAllRows();
			dojo.destroy(this.content);
			this._createContentNode();
			this.setHeaderCheckbox(false);
		}
	},
	
	uninitialize: function(){
		dojo.forEach(this._toDestroy, function(object){
			if (object.destroyRecursive && !object._beingDestroyed){
				object.destroyRecursive();
			}
			else if (object.destroy && !object._beingDestroyed){
				object.destroy();
			}
			delete (object);
		});
		this._clearNodeBucket();
		if (this.domNode) {
			this._clearHandles();
			this._removeAllRows();
			dojo.destroy(this.content);
		}
	},
	
	hardClear: function() {
		this.clear();
	},
	
	_removeAllRows: function(/*boolean*/keepSelection) {
		// summary:
		//		Clears selection, deletes and garbage collects all rows.
		for (var key in this.allRows) {
			if (!keepSelection) {
				if (this.selectedRows[key]) {
					delete this.selectedRows[key];
				}
			}
			//Destroy the entries for the rows we need to
			this._doDisconnects(this.allRows[key].connections);

			var toDestroy = this.allRows[key].destroy;
			if (toDestroy.length > 0) {
				//We need to tidy up
				this._callDestroys(toDestroy);
			}
		}
		this.allRows = {};
		this.rowOrder = [];
		this.totalRowCount = 0;
	},

	_doDisconnects: function(connections){
		if (connections){
			dojo.forEach(connections, dojo.disconnect);
		}
	},
	
	_callDestroys: function(toDestroyArray) {
		//We need to tidy up
		for (var i=0; i<toDestroyArray.length; i++) {
			if (toDestroyArray[i] && toDestroyArray[i].destroyRecursive) {
				toDestroyArray[i].destroyRecursive();
			} else if (dojo.isArray(toDestroyArray[i])) {
				dojo.forEach(toDestroyArray[i], function (toDestroy) {
					toDestroy.destroyRecursive();
				});
			} else {
				console.warn("Module: undestroyable artifact:", typeof toDestroyArray[i]," ", toDestroyArray[i]);
			}
		}
	},
	
//	addRow: function (/*Unique Id*/objectId, /*rowtype*/rowType, /*Actual Row*/node) {
//		this.allRows[objectId] = {rowType: rowType, node: node};
//	},

	getContentNode: function () {
		return this.content;
	},

	prepareNodeForInsert: function (/*Element*/node) {
		// summary:
		//		This function makes sure that the supplied node is prepared
		//		for inclusion in the modules DOM.  So it makes sure the mouse over
		//		listeners are added.
		var that = this;
		var connectionsCreated = [];
		dojo.query("td", node).forEach( function (newCell) {
			connectionsCreated = connectionsCreated.concat(that._setMenuLauncherHoverEvents(newCell));
			if (that.doTrackFocusForCells) {
				connectionsCreated.push(dojo.connect(newCell, "onfocus", function(evt){
					that._handleCellOnFocus(evt, node);
				}));
				connectionsCreated.push(dojo.connect(newCell, "onblur", function(evt){
					that._handleCellOnBlur(evt, node);
				}));
			}
		});
		
		var handle;
		handle = dojo.connect(node, "onmouseover", function(event){
			that._handleMouseOverRow(event, node);
		});
		connectionsCreated.push(handle);
		handle = dojo.connect(node, "onmouseleave", function(event){
			that._handleMouseOutRow(event, node);
		});
		connectionsCreated.push(handle);

		that.dojoConnectList.addItems(connectionsCreated);
		return connectionsCreated;
	},

	_handleClick: function (event, node) {
		// summary:
		//		Generic event handler determines whether
		//		a node representing an object has been clicked
		//		(i.e. row selection) or a menu icon has been clicked.
		//		dojo.connect'ed to onClick on the parent div.

		var isInEditMode = this.isInEditMode();
		
		if (this._dragInProgress) {
			return;
		}
		
		node = node || event.target;
		var ids = this.getRowAndColumnIds(node),
			columnId = ids.columnId,
			rowTypeId = ids.rowTypeId,
			row = ids.rowNode,
			column = ids.colNode;


		// If an edit button click, simply return - this has already been handled
		// by _startEdit.
		if (node && node.parentNode && dojo.hasClass(node.parentNode, "editButton")) {
			return;
		}
		
		// if the node is a resource link, don't process the event (and select the row 56613)
		if (node && dojo.hasClass(node, "jazz-ui-ResourceLink")) {
			dojo.removeClass(row, "rowHover");
			return;
		}
		
		
		//Check to see if this clicked element should simply do class manipulation of its parent
		if (dojo.attr(node, "classToggle")) {
			var classToggle = dojo.attr(node, "classToggle");
			dojo.toggleClass(node.parentNode, classToggle);
			dojo.stopEvent(event);
			return;
		}
		
		
		
		if (this.manageFocusNavigation) {
			// The grid is managing focus of its cells.  Set the grid focus to the clicked cell
			// Cancelling any delayed focus prevents the grid from scrolling out from under the
			// place the user has clicked.
			this._cancelPendingFocus();
			if(row) {
				var rowKey = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
				this._moveGridFocus(rowKey, rowTypeId, columnId);
			}
		}
		
		var menuType = dojo.attr(node, com.ibm.rdm.web.grid.Module.AttributeNames.MENU_TYPE);
		var that = this;
		if (menuType === "1") {
			this._openRowActionMenu(event, row, rowTypeId, node, isInEditMode);
			return;
		} else if (menuType === "2") {
			//Expando pressed
			var objectId = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
			this.applyExpando(rowTypeId, objectId);
			return;
		}
		
		//Deal with EditCommands
		if (rowTypeId && columnId) {
			var isVariable = this.gridmodel.isVariable(rowTypeId, columnId);
			if (isVariable)	{
				if (node !== null && node.nodeName === "INPUT" && node.type==="checkbox") {
					var objectId = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
					this.gridmodel.setValue(rowTypeId, columnId, objectId, node.checked);
				}
				return;
			}
		}
		
		var isSelector = dojo.hasClass(node, "_selector");
		var isEditable = dojo.hasClass(node, com.ibm.rdm.web.grid.Module.ClassNames.CELL_EDITABLE);
		//Any click on row is a selection click
		var isRowSelector = this.gridmodel.rowIsSelector(rowTypeId);
		var isInSelector =  isSelector || isEditable || isRowSelector;
		
		if (isInSelector || (node !== null && node.nodeName === "INPUT") &&
			rowTypeId && this.gridmodel.hasSelectionBox(rowTypeId)) {
			var checked = false;
			if (isInSelector) {
				
				if (!isSelector && isRowSelector) {
					//Need to make node the selector
					var checkNodeQuery = dojo.query("._selector", row);
					var checkNode = checkNodeQuery && checkNodeQuery[0];
				} else {
					//We need to work out if the node is checked or not
					var checkNode = node && node.childNodes && node.childNodes[0] && node.childNodes[0].childNodes &&
											node.childNodes[0].childNodes[0];
				}
				if (checkNode) {
					checked = !checkNode.checked;
					dojo.attr(checkNode, "checked", checked);
				}
			} else {
				checked = node.checked;
			}
			var objectId = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
			this.onCheck(event, row, objectId, rowTypeId, columnId, checked);
			return;
		}
		
		//Is entire row an Expando?
		if (rowTypeId && this.gridmodel.rowIsExpando(rowTypeId)) {
			var objectId = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
			this.applyExpando(rowTypeId, objectId);
			return;
		}
		
		//Mop up any event not already dealt with
		this._selectObjectByEvent(event, row, rowTypeId, columnId, false);
	},

	_isStubRow: function(rowNode) {
		var objectId = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		return (objectId && objectId.indexOf("STUB_") !== -1);
	},
	
	_openRowActionMenu: function(event, rowNode, rowType, menuButtonNode, isInEditMode) {
		var showEditMenu = dojo.hitch(this, function(){
			var objectId = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
			
			var rowPreviouslySelected = dojo.hasClass(rowNode, "row-checked");
			if(!rowPreviouslySelected){
				//If the row was not previously selected, we should deselect all other selected rows
				var uris = [];
				for (var x in this.selectedRows) {
					uris.push(this.selectedRows[x].uri);
				}
				this.removeFromSelection(uris);
			}
			this.applyCheckRow(rowNode, objectId, rowType);
			this._informSelectionChanged();
			
			var count = 0;
			var rows = {};
			for (var x in this.selectedRows) {
				count++;
				var uri = this.selectedRows[x].uri;
				var rowTypeId = this.selectedRows[x].rowTypeId;
				rows[uri] = rowTypeId;
			}
			
			var operations = this.gridmodel.getOperations(rowType);
			var menuLayout = this.gridmodel.getMenuLayout(rowType);

			var editMenu = new com.ibm.rdm.web.grid.EditMenu({
				node: menuButtonNode,
				objectId: objectId,
				launchGesture: event,
				menuType: "1",		// Always "1" for artifact menu
				module: this,
				row: rowNode,
				rows: rows,
				count: count,
				onCancel: dojo.hitch(this, function(){
					var tabTarget = this._navigationTabTarget;
					if (tabTarget) {
						this.moveFocusToCellOnRow(tabTarget.rowKey);
					}
				}),
				destroy: dojo.hitch(this, function() {
					this._wasInEditModeWhenOpenEditMenuWasRequested = false;
					this._wasInStubRowWhenOpenEditMenuWasRequested = false;
					editMenu.inherited("destroy", arguments);
				})
			}, operations, menuLayout, dojo.moduleUrl("com.ibm.rdm.web", "grid/images/grid-rsrce-dd-hover.gif").toString(), 22, 30);
			this._toDestroy.push(editMenu);

			this._editMenu = editMenu;

			editMenu.show();
		});
		if (isInEditMode) {
			this._wasInEditModeWhenOpenEditMenuWasRequested = true;
			rowNode.focus();
			if (this._isStubRow(rowNode)) {
				// Note: To address Defect 85588 we're explicitly NOT calling
				// showEditMenu() in this 'if' block.
				// Instead, as a result of the below onSaveKey() call, eventually
				// the this class's 'refreshRows' method will be invoked and that
				// method will ask for the edit menu to be displayed AFTER the row
				// has been saved and an appropriate URI can be used for the
				// actions associated with the edit menu's menu items.
				// While waiting for the save to occur, the pencil icon appropriately
				// turns into a spinning icon to cue the user into the fact that the
				// edit menu will eventually appear.
				this._wasInStubRowWhenOpenEditMenuWasRequested = true;
			} else {
				showEditMenu();
			}
			this.onSaveKey();
		} else {
			rowNode.focus();
			showEditMenu();
		}
	},

	applyExpando: function(rowTypeId, objectId) {
		this.gridmodel.applyExpando(rowTypeId, objectId);
	},

	findContainingRow: function (node) {
		// summary:
		//		Given a node (on which some event has taken place)
		//		walk back up the DOM until we find a parent that has
		//		a OBJECTID attribute.

		return DomUtils.findAncestorByAttr(node, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
	},
	
	getRowAndColumnIds: function (node) {
		// summary:
		//		Given a node (on which some event has taken place)
		//		get the row and column Ids applicable

		var colNode = DomUtils.findAncestorByAttr(node, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
		if (colNode) {
			var columnId = dojo.attr(colNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
			//Go from colnode
			var rowNode = DomUtils.findAncestorByAttr(colNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
		} else {
			//Start again for rownode
			var rowNode = DomUtils.findAncestorByAttr(node, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
		}
		if (rowNode) {
			var rowTypeId = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
		}
		
		return {columnId: columnId, rowTypeId: rowTypeId, rowNode: rowNode, colNode: colNode};
	},

	_selectObjectByEvent:  function(event, row, rowTypeId, columnId, checkInBox) {
		// Here to allow subclasses to override the arguments.  see VirtualGrid
		//
		this._selectObjectByEventImpl(event, row, rowTypeId, columnId, checkInBox);
	},

	_selectObjectByEventImpl: function(event, row, rowTypeId, columnId, checkInBox) {
		// summary:
		//		Handles object selection from an event, including taking into account the modifier keys.
		//		none - only this row is selected.
		//		shift - this row and all inbetween are selected.
		//		ctrl - this row is added from the selection.
		//		Combinations using shift and ctrl are not supported.
		
		if (row) {
			var objectId = dojo.attr(row, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
			if (this.gridmodel.allowsMultiSelect(rowTypeId)) {
				var ctrl = event.ctrlKey || event.metaKey;
				var shift = event.shiftKey;
				
				if (!(ctrl || (checkInBox && !shift))) {
					//Clear all previous rows, unless - this was a ctrl click, or it was in a checkbox and wasn't shifted.
					if (!shift) {
						this._selectAll(false);
					}
					this.setHeaderCheckbox(false);
				}
				
				if (shift) {
					var previousId;
					if (this._previousSelection) {
						//Use the first shift-key clicked.
						previousId = this._previousSelection;
					} else {
						previousId = this.getSelectedObjectId();
						//Preserve the previousSelection
						this._previousSelection =previousId;
					}
					if (previousId) {
						// select between ids will deselect everything
						// if the selection is allowed otherwise it wont,
						// preserving the the current selection
						this.selectBetweenIds(previousId, objectId);
					} else {
						//No previous row - just ignore the shift
						this._selectAll(false);
						this.applyCheckRow(row, objectId, rowTypeId);
					}
				} else {
					this.applyCheckRow(row, objectId, rowTypeId);
					this._previousSelection = null;
				}

				if (this.gridmodel.isSelectable(rowTypeId)) {
					this._selectObject(row);
				}
			} else {
				if (this.gridmodel.isSelectable(rowTypeId)) {
					this.applyCheckRow(row, objectId, rowTypeId);
					this._selectObject(row);
				}
			}
			
			this._informSelectionChanged();
		}
	},
	
	_informSelectionChanged: function () {
		//This notifies the grids' owner if that has been set up.
		this.gridmodel.selectionChanged(this.getSelectedRows());
	},
	
	
	isLoaded: function (key) {
		// summary:
		//		Returns whether or not the given key is loaded in the grid.  The key
		//		is loaded if it is currently displayed.
		return true;
	},
	
	enableFisheye: function() {
		// no-op
	},
	
	disableFisheye: function() {
		// no-op
	},

	scrollIntoView: function (key) {
		var details = this.allRows[key];
		if (details && details.node) {
			dijit.scrollIntoView(details.node);
		}
	},
	
	getCellHtml: function(key, colId) {
		var row = this.allRows[key];
		if (row && row.node) {
			var results = dojo.query("div."+colId, row.node);
			if (results.length > 0) {
				return results[0].firstChild;
			}
		}
		return null;
	},
	
	_selectObject: function (/* element */node, /*boolean ?*/scroll, /*boolean ?*/noPublish, /* boolean ?*/ skipBrowserFocus) {
		// summary:
		//		Selects the node, giving it keyboard focus and adding the
		//		row-selected CSS class.  Scroll is an optional parameter
		//		which if passed causes this function to remember then scroll
		//		back to the x coordinate of the view before the node was
		//		given focus.  This is only required on IE and is needed
		//		because IE likes to make sure the left most edge of the dom
		//		node with the focus is visible.
		// noPublish:
		//		true if the selection of this node should not be published
		// skipBrowserFocus:
		//		true if browser focus should not be set for the row
		// returns:
		//		A deferred that resolves once the node has been changed and
		//		given focus.
		
		var focusDeferred;

		if (node) {
			// we set the CSS style in here, but then just forward the event
			// on.  If this widget is being used outside of CAF, whatever
			// listens to this event would do something different than when
			// it is being used within CAF for example.
		
			if (this.selectedObject !== null) {
				dojo.removeClass(this.selectedObject, com.ibm.rdm.web.grid.Module.ClassNames.ROW_SELECTED);
				//dojo.removeAttr(this.selectedObject, "tabIndex");
				dojo.query("["+com.ibm.rdm.web.grid.Module.AttributeNames.FOCUSABLE+"]",
							this.selectedObject).forEach(function (item) {
								//dojo.removeAttr(item, "tabIndex");
				});
			}
			this.selectedObject = node;
			//dojo.attr(this.selectedObject, "tabIndex", "0");
			dojo.addClass(node, com.ibm.rdm.web.grid.Module.ClassNames.ROW_SELECTED);

			var objectid = dojo.attr(node, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);

			var that = this;
			dojo.query("["+com.ibm.rdm.web.grid.Module.AttributeNames.FOCUSABLE+"]", node).forEach(function (item) {
				//dojo.attr(item, "tabIndex", "0");

				if (!item.hasKeyboardListener) {

					var handle = dojo.connect(item, "onkeypress", function (event) {

						if (event.keyCode === dojo.keys.ENTER) {
							//if edit menu is already displayed do not launch another.
							var editMenuQueryResult = dojo.query(".editmenu", this.domNode);
							if (editMenuQueryResult.length > 0) {
								return;
							}

							dojo.stopEvent(event);
							that._handleClick(null, item);
						}

					});
					that.dojoConnectList.addItem(handle);
				}

				item.hasKeyboardListener = true;

			});

			var scrollX = this.domNode.parentNode.scrollLeft;

			if (skipBrowserFocus) {
				focusDeferred = new dojo.Deferred();
				focusDeferred.callback();
			} else {
				focusDeferred = this._setFocusToNode(node);
			}

			if (scroll) {
				this.domNode.parentNode.scrollLeft = scrollX;
			} else if (scrollX !== this.domNode.parentNode.scrollLeft) {
				this.domNode.parentNode.scrollLeft = 0;
			}
		}
		return focusDeferred;
	},
	
	deleteRows: function (rowKeys, /*boolean*/dontMoveFocus, /*boolean?*/retainSelection) {
		// summary:
		//		Deletes the rows mentioned.
		// rowKeys:
		//		An array of row keys.
		var self = this,
			underDrag = (dojo.query('.com-ibm-rdm-web-artifact-ModuleArtifactWidget.dragging')[0]) ? true : false;
		dojo.forEach(rowKeys, function(rowKey) {
			self.deleteRow(rowKey, true, dontMoveFocus, retainSelection, underDrag);
		});
	},
	
	deleteRow: function (/*Object*/key, /*Boolean*/doNotAnimate, /*Boolean*/dontMoveFocus, /*Boolean*/retainSelection, /*Boolean*/underDrag) {
		// summary:
		//		Allows a row to be deleted.

		//Remove from selection if its there
		var selectionChanged = false;
		if (!retainSelection) {
			if (this.selectedRows[key]) {
				delete this.selectedRows[key];
				selectionChanged = true;
			}
		}
		
		var deletedNode = this._findNodeByObjectId(key);
		this.totalRowCount--;
		
		//Remove the row from the rowOrder
		var newOrder = [];
		for (var i = 0; i < this.rowOrder.length; i++) {
			var thisOne = this.rowOrder[i];
			if (thisOne.key != key) {
				newOrder.push(thisOne);
			}
		}
		this.rowOrder = newOrder;
		
		if (deletedNode) {
			var removalFunction,
				nextFocusNodeKey,
				nextFocusNode;
				
			// Find the key of the next node to focus on
			nextFocusNode = DomUtils.getNextElementSibling(deletedNode);
			if (!nextFocusNode) {
				nextFocusNode = DomUtils.getPreviousElementSibling(deletedNode);
			}
			if (nextFocusNode) {
				nextFocusNodeKey = dojo.attr(nextFocusNode, "objectId");
			}
			
			
			if (!doNotAnimate) {
				removalFunction = function (n) {
					var anim = dojo.fx.wipeOut({node : n, duration : 500}),
					connection = dojo.connect(anim, "onEnd", function () {
						if (n.parentNode) {
							n.parentNode.removeChild(n);
						}
						dojo.disconnect(connection);
					});
					anim.play();
				};
			}
		
			//this._removeObjectsBetween(deletedNode, deletedNode, true, removalFunction);

			if(underDrag) {
				// If the dragSource is removed from DOM the necessary drag events are not fired.
				var dragHandleNode = dojo.query(".dragHandle.active", deletedNode)[0];
				if(dragHandleNode) {
					dragHandleNode.className = "orphan-dragHandle";
					document.body.appendChild(dragHandleNode);
				}
			}

			if (doNotAnimate) {
				deletedNode.parentNode.removeChild(deletedNode);
			} else {
				removalFunction(deletedNode);
			}
			if (nextFocusNodeKey && !dontMoveFocus) {
				//console.log("deleteRow focussing on: " + nextFocusNodeKey);
				// this.selectObjectById(nextFocusNodeKey);
				this._navigationTabTarget.rowKey = nextFocusNodeKey;
				var rowInfo = this._getCurrentRowInfo();
				if (rowInfo) {
					this._moveFocusToFirstCell(rowInfo);
				}
			}
		}
		
		if (selectionChanged) {
			this._informSelectionChanged();
		}
		
		delete this.allRows[key];
	},
	
	_removeObjectsBetween: function(/*Node*/startObject,
									/*Node*/endObject,
									/*Boolean?*/inclusive,
									/*Function?*/removalFunction) {
		// summary:
		//		Remove all the DOM nodes between startObject and endObject.
		//		if inclusive is true then remove the start and end nodes as well.
		//		Start and end are expected to be contiguous, if they are not
		//		then this function will throw an error. If endObject is undefined
		//		then this function will remove all siblings after startObject.
		
		var nextSibling = inclusive ? startObject : startObject.nextSibling,
				removalList = [], // the list of dom nodes we are going to remove.
				holdingDiv; // a DIV to group all the nodes that are going to be removed.

		while (nextSibling && nextSibling !== endObject) {
				removalList.push(nextSibling);
				nextSibling = nextSibling.nextSibling;
		}

		if (nextSibling === endObject) {
				if (inclusive) {
					removalList.push(endObject);
				}
				// good, we found all the nodes between start & end.
				// now take them out.
				holdingDiv = document.createElement("div");
				if (removalList.length > 0) {

					var ignore = true;
					try {
						this.content.replaceChild(holdingDiv, removalList[0]);
						ignore = false;
					} catch (e) {
						//If we have an error here, ignore it
					}
					if (!ignore) {
						// add all the nodes to be removed into a single parent
						// we can animate the removal of this node more smoothly
						for (var i=0; i<removalList.length; i++) {
							holdingDiv.appendChild(removalList[i]);
						}

						if (!removalFunction) {
							holdingDiv.parentNode.removeChild(holdingDiv);
						} else {
							removalFunction(holdingDiv);
						}
					}
				} else {
					console.error("Found nothing to remove between ", startObject, " and ", endObject);
				}

				return removalList.length;

		} else {
				throw {
					msg : "MODULE COULD NOT MAKE CONTIGUOUS NODE LIST FROM " + startObject + " TO " + endObject
				};
		}
	},

	/*
	* Explicitly select the DOM node representing Id.
	* Expects that any DOM element reprsenting an Object
	* has an attribute "objectId".  If skipBrowserFocus is true,
	* browser focus is not given to the row.
	*/
	selectObjectById: function (id, skipBrowserFocus) {
		var node = dojo.query("["+com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID+"='"+id+"']", this.domNode)[0];
		if (node) {
			var rowTypeId = dojo.attr(node, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			var isSelectable = this.gridmodel.isSelectable(rowTypeId);
			
			if (isSelectable) {
				this._selectObject(node, false, false, skipBrowserFocus);
				return true;
			}
			return false;
		} else {
			return false;
		}
	},
	
	selectBetweenIds: function (firstId, secondId) {
		//This goes through all objects, and selects those between the first and second.
		//It is unlikely to perform well on large lists - we will need a better data model to support them.
		//I hope this does it in order!
		
		this._selectAll(false);
		
		var selecting = false;
		var that = this;
		dojo.query("["+com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID+"]", this.domNode)
			.forEach(function(item) {
				//We want the selection to be inclusive.
				var selectThis = selecting;
				var thisRowId = dojo.attr(item, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
				var thisRowTypeId = dojo.attr(item, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
				if (thisRowId === firstId || thisRowId === secondId) {
					selecting = !selecting;
				}
				if (selecting || selectThis) {
					that.applyCheckRow(item, thisRowId, thisRowTypeId);
				}
			});
	},

	/*
	* this gets called for each column when view header info is received
	* and whenever the user resizes the column
	*/
	setColumnWidth: function (/* String */ rowTypeId, /* String */column,
								/* Integer */newWidth,
								/* Integer */newRowWidth) {
		//console.log("M setColumnWidth", rowTypeId, column, newWidth, newRowWidth);
		
		this.gridmodel.setWidth(rowTypeId, column, newWidth);

		dojo.query("."+column, this.domNode).style("width", newWidth+"px");
		
		this.updateLinkBoxPosition();
	},

	/*
	* Invoked when the module has been closed.
	*/
	close: function () {
		//Stop listening for events.
		this._clearHandles();
	},
	
	_clearHandles: function () {
		//Stop listening for events.
		this.dojoConnectList.forEach(function f(handle){
				dojo.disconnect(handle);
		});
		this.dojoConnectList.clear();
	},
	
	_clearNodeBucket: function(){
		while(this._nodeBucket.length > 0){
			dojo.destroy(this._nodeBucket[0]);
			this._nodeBucket.splice(0, 1);
		}
	},

	_findNodeByObjectId: function (/*string*/objectId) {
		// summary:
		//		Return the DOM node that has an objectid attribute matching the parameter.
		return dojo.query("table["+com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID+"='"+objectId+"']", this.domNode)[0];
	},

	getFirstObjectId: function () {
		// summary:
		//		Return the objectid of the first object in this module DOM.
		if (this.content.firstChild) {
			return dojo.attr(this.content.firstChild, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		}
	},

	getLastObjectId: function () {
		// summary:
		//		return the objectid of the last object in this module DOM.
		if (this.content.lastChild) {
			return dojo.attr(this.content.lastChild, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		}
	},

	_updateCellButtonImageCursors: function(cell, cursorType) {
		// summary:
		//		Updates the mouse cursor of a given cells image buttons.
		// cell:
		//		The cell to update.
		// cursorType:
		//		The type of cursor.
		if (cell) {
			var icons = dojo.query(".cellButtonImage", cell);
			if (icons) {
				dojo.forEach(icons, function (icon) {
					dojo.style(icon, "cursor", cursorType);
				});
			}
		}
	},
	
	sortBy: function(columnId, order) {
		var primaryRowTypeId = this.gridmodel.getPrimaryRowTypeId();
		var sortFunction = this.getSortFunction(primaryRowTypeId, columnId);
		
		sortFunction(order, dojo.hitch(this, this.applyOrder));
	},

	insertRow: function(insertAfterHereKey, newKey, rowTypeId) {
		// summary:
		//		Allows a row to be inserted.
		// insertAfterHereKey:
		//		The key for the row after which the row will be inserted.
		//		If this is null, the row is inserted at the very beginning of the document.
		// newKey:
		//		The key of the row to insert.
		// rowTypeId:
		//		The rowType of the row to insert.
		// returns:
		//		A deferred that resolves when the row has been inserted and displayed.
		return this.insertRows(insertAfterHereKey, [{key: newKey, rowtype: rowTypeId}]);
	},

	insertRows: function(insertAfterHereKey, rows, noFocus) {
		// summary:
		//		allows a row to be inserted.
		// insertAfterHereKey:
		//		The key for the row after which the rows will be inserted.
		//		If this is null, the row is inserted at the very beginning of the document.
		// rows:
		//		Array of {key: ..., rowtype: ...} entries.
		
		//Place to hold any deferred population calls
		var functionsToCall = [],
			self = this,
			insertDeferred,
			firstInsertedRow;
		
		if (rows.length > 0) {
			//Go backwards as the inserts are all after the same 'insertAfterHereKey', and so are reversed.
			for( var x = rows.length-1; x>= 0; x--) {
				var row = rows[x],
					key = row.key,
					rowtype = row.rowtype;
				var insertedRow = self._insertRow(insertAfterHereKey, key, rowtype, functionsToCall);
				if (!firstInsertedRow) {
					firstInsertedRow = insertedRow;
				}
				this.totalRowCount++;
			}
			
			//Now, lets make sure we get the deferred values we want
			insertDeferred = self._callDeferredFunctions(functionsToCall);
		}
		
		//Now add rows to the rowOrder
		var newOrder = [];
		var inserted = false;
		for (var i = 0; i < this.rowOrder.length; i++) {
			var thisOne = this.rowOrder[i];
			newOrder.push(thisOne);
			if (thisOne.key == insertAfterHereKey) {
				newOrder = newOrder.concat(rows);
				inserted = true;
			}
		}
		
		if (!inserted) {
			//insert at top
			newOrder = this.rowOrder.concat(rows);
		}
		this.rowOrder = newOrder;
		
		var focusDeferred = this._setFocusToNode(firstInsertedRow, insertDeferred);
		
		// place focus on the first inserted row - unless caller told us not to.
		if (!noFocus) {
			insertDeferred.addCallback(this, function(){
				var focusDeferred = this._selectObject(firstInsertedRow);
				return focusDeferred;
			});
		}
		return insertDeferred;
	},
	
	_insertRow: function(insertAfterHereKey, newKey, rowTypeId, functionsToCall) {
		// summary:
		//		Allows a row to be inserted.
		// insertAfterHereKey:
		//		The key for the row after which the row will be inserted.
		//		If this is null, the row is inserted at the very beginning of the document.
		// newKey:
		//		The key of the row to insert.
		// rowTypeId:
		//		The rowType of the row to insert.
		// returns:
		//		The inserted row node.

		//Build the actual row

		var self=this;
		//var insertionFunction = function (n) {
		//	var anim = dojo.fadeIn({node : n, duration : 500});
		//	anim.play();
		//};
		
		var contentRow = self.gridmodel.getRowData(rowTypeId, newKey);
		var row = self._createRow(rowTypeId, newKey, contentRow, functionsToCall);
		var newConnections = self.prepareNodeForInsert(row);
		//dojo.style(row, "opacity", "0");
		
		//Insert it into the AllRows
		this.allRows[newKey]={node: row, rowtype: rowTypeId, destroy: [], connections: newConnections};
	
		//Insert it into the dom.
		if (insertAfterHereKey) {
			//Put the node after the target
			var target = this.allRows[insertAfterHereKey];
			if (target) {
				dojo.place(row, target.node, "after");
			} else {
				com.ibm.rdm.web.common.ErrorHandler.warn("insertAfterHereKey not found", insertAfterHereKey);
			}
		} else {
			// Insert as first row.  Rows are in groups of a few rows, within parent
			// nodes, so create a new parent node to hold all the added nodes at the
			// top if necessary.
			var insertParent = dojo.query(">.topInsertParent", this.content)[0];
			if (!insertParent) {
				insertParent = dojo.create("div", {"class":"topInsertParent"}, this.content, "first");
			}
			dojo.place(row, insertParent, "first");
		}
		
		//insertionFunction(row);
		return row;
	},

	applyOrder: function(newResourceOrder) {
		// summary:
		//		Applies the give row order to the grid.
		//		If rows are not available, they will be asked for,
		//		so this can be used to initialise or refresh the grid,
		//		as well as re-order.
		// newResourceOrder:
		//		An array of {key:... rowtype: ... } entries.

		var self = this;
		
		//remove all the nodes...
		if ( self.content.hasChildNodes()) {
			while ( self.content.childNodes.length >= 1 ) {
				self.content.removeChild( self.content.firstChild );
			}
		}
		
		var rowCount = 0;
		var fragment = document.createDocumentFragment();
		
		//Place to hold any deferred population calls
		var functionsToCall = [];
		var self=this;

		var newAllRows = {},
			oldRows = this.allRows;
		
		this.allRows = newAllRows;
		
		var divOfTen = document.createElement("div");
		var count = 0;
		
		if (dojo.isArray(newResourceOrder)) {
			dojo.forEach(newResourceOrder, function(item, index) {

				if (oldRows && oldRows[item.key]) {
					var differentVersion = false;
					var current = oldRows[item.key];
					var node = null;
					
					//Check the version if necessary
					var newVersion = self.gridmodel.getVersion(item.rowtype, item.key);
					if (newVersion) {
						var currentVersion = current.version;
						differentVersion = (newVersion != currentVersion);
					}
					if (!differentVersion) {
						node = current.node;
						newAllRows[item.key]= current;
						//Remove from current - as those remaining in here will be destroyed at end
						delete oldRows[item.key];
					}
				}
				
				if (node && _nodeCanBeDetached(node) && !differentVersion) {
					self.prepareNodeForInsert(node);
					divOfTen.appendChild(node);
					rowCount++;
				} else {
					//Node doesn't exist - we must build it
					var contentRow = self.gridmodel.getRowData(item.rowtype, item.key);
		
					var allNodeRecord = {rowtype: item.rowtype, destroy: [], connections: []};
					newAllRows[item.key]=allNodeRecord;
					var row = self._createRow(item.rowtype, item.key, contentRow, functionsToCall);
					if (index === 0){
						dojo.addClass(row, "row1");
					}
					var newConnections = self.prepareNodeForInsert(row);
					
					allNodeRecord.node = row;
					allNodeRecord.connections = allNodeRecord.connections.concat(newConnections);
					//Keep the version if necessary
					var version = self.gridmodel.getVersion(item.rowtype, item.key);
					if (version) {
						allNodeRecord.version = version;
					}
					
					
					divOfTen.appendChild(row);
					rowCount++;
				}
				if (count++ > 9) {
					//We have 10 - add as a separate tree
					fragment.appendChild(divOfTen);
					divOfTen = document.createElement("div");
					count = 0;
				}
			});
			
			//Add the remainder
			if (count > 0) {
				fragment.appendChild(divOfTen);
			}
			
			//Now, lets makes sure we get the deferred values we want
			self._callDeferredFunctions(functionsToCall);
		}
		
		self.totalRowCount = rowCount;
		
		self.content.appendChild(fragment);
		
		var deletedRows = [];
		var selectionChanged = false;
		
		for (var key in oldRows) {
			deletedRows.push(key);
			if (this.selectedRows[key]) {
				delete this.selectedRows[key];
				selectionChanged = true;
			}
			//Destroy the entries for the rows we need to
			this._doDisconnects(oldRows[key].connections);
			var toDestroy = oldRows[key].destroy;
			if (toDestroy.length > 0) {
				//We need to tidy up
				this._callDestroys(toDestroy);
			}
		}
		if (deletedRows.length > 0) {
			this.gridmodel._notifyRowDeletes(deletedRows);
			if (selectionChanged) {
				this._informSelectionChanged();
			}
		}
		
		this.rowOrder = newResourceOrder;
		
		if (this.dragDropArgs) {
			this.setDragDropArgs(this.dragDropArgs);
		}
	},
	
	getRowsOrder: function() {
		return this.rowOrder;
	},
	
	getRowCount: function() {
		return this.totalRowCount;
	},
	
	getRowByKey: function(key) {
		return this.allRows[key];
	},
	
	_callDeferredFunctions: function(functionsToCall, dontYield) {
		// summary:
		//		Calls a set of functions that are used to populate cells in the grid.
		// functionsToCall:
		//		An array of descriptors, each describing a function to call, and its arguments.
		// dontYield:
		//		Flag to indicate that calls should not be batched and separated by yields.
		// returns:
		//		A deferred that resolves when all the functions have been called, and their
		//		setContentFn functions called.

		//Lets call deferred functions in batches of 20 to improve perceived performance?
		var BATCH_SIZE = 20;
		var start = 0;
		var length = functionsToCall.length;
		var allDeferreds = [];
		
		// Add a deferred to each function descriptor, so that we can know when it is complete.
		for (var i = 0 ; i < length ; i++) {
			var deferred = new dojo.Deferred();
			functionsToCall[i].deferred = deferred;
			allDeferreds.push(deferred);
		}
		
		if (length > 0) {
			//Divide the calling of functions into batches
			do {
				var limit = Math.min(start+BATCH_SIZE, length);
				this._callFunctionBatch(functionsToCall, start, limit, dontYield);
				start = start+BATCH_SIZE;
			} while (limit < length);
		}
		
		return new dojo.DeferredList(allDeferreds);
	},
	
	_callFunctionBatch: function (functionsToCall, start, limit, dontYield) {
		
		var doSingleCall = function(functionDescriptor, index) {
				var setContentFunction = function() {
					try {
						functionDescriptor.setContentFn.apply(this, arguments);
						functionDescriptor.deferred.callback();
					} catch (e) {
						functionDescriptor.deferred.errback(e);
					}
				};
				functionDescriptor.func(setContentFunction, functionDescriptor.row, functionDescriptor.key,
										functionDescriptor.containingNode, functionDescriptor.indent, functionDescriptor.columnId, functionDescriptor.isFocusCell);
			
		};
		
		var doBatchCall = function () {
			for (var j = start; j < limit; j++) {
				doSingleCall(functionsToCall[j], j);
			}
		};

		//if (!dontYield) {
		//	window.setTimeout(doBatchCall, 0);
		//} else {
			doBatchCall();
		//}
	},

	_setMenuLauncherHoverEvents: function(cell) {
		// summary:
		//		Sets up the hover over events for a cell's editor launcher icon.
		// cell:
		//		The cell to update.
		var that = this;
		var connectionsCreated = [];
		if (cell) {
			var launcherIcons = dojo.query(".menuLauncher", cell);
			if (launcherIcons) {
				dojo.forEach(launcherIcons, function (launcherIcon) {
					var handle;
					handle = dojo.connect(launcherIcon, "onmouseover", function() {
						if (!that._dragInProgress) {
							dojo.addClass(cell, com.ibm.rdm.web.grid.Module.ClassNames.EDITOR_LAUNCHER_HOVER);
						}
					});
					connectionsCreated.push(handle);

					handle = dojo.connect(launcherIcon, "onmouseleave", function()
						{dojo.removeClass(cell, com.ibm.rdm.web.grid.Module.ClassNames.EDITOR_LAUNCHER_HOVER);});
					connectionsCreated.push(handle);
				});
			}
		}
		return connectionsCreated;
	},
	
	_setFocusToNode: function(focusNode, /* optional */ deferredOrPromise){
		// summary:
		//		Sets focus to a given node.
		// focusNode:
		//		A node to set focus on.
		// deferredOrPromise:
		//		An optional deferred or promise.  If provided focus is set once this
		//		deferred or promise has resolved.
		// returns:
		//		A deferred or promise that resolves once the focus has been set.  If a
		//		promise was provided, a promise is returned.  Otherwise a deferred is
		//		returned.
		
		// TODO - BGM - Probably need to remove this when keyboard focus work is more stable.
		// Until then it effectively does nothing, but we leave it here, because its callers
		// are expecting a deferred back, so removing it is non-trivial.
		
		if (!focusNode) {
			return;
		}
		
		if (!this.focusCount) {
			this.focusCount = 0;
		}
		this.focusCount++;
		
		
		var setFocusDelayed = function(focusNode) {
			var focusDeferred = new dojo.Deferred();
			window.setTimeout(function(){
				//console.log("setFocusDelayed: initial focus:");
				//console.log(document.activeElement);
				//dojo.attr(focusNode, "tabIndex", "0");
				//focusNode.focus();
				//console.log("setFocusDelayed: after focus:");
				//console.log(document.activeElement);
				focusDeferred.callback();
			}, PRE_FOCUS_DELAY_MS);
			return focusDeferred;
		};
		
		if (deferredOrPromise) {
			// We've been given a promise of some sort - chain into its callback list.
			if (deferredOrPromise.addCallback) {
				// it's a deferred
				return deferredOrPromise.addCallback(function(){
					return setFocusDelayed(focusNode);
				});
			}
			// It's a promise
			return deferredOrPromise.then(function(){
				return setFocusDelayed(focusNode);
			});
		} else {
			// return a deferred that resolves when focus is done
			return setFocusDelayed(focusNode);
		}
	},
	
	/////////////////////////
	// KEYBOARD NAVIGATION //
	/////////////////////////
	
	_onGridFocus: function(/*Object?*/event) {
		// summary:
		//		Called when the grid gains focus.  Controls which cell within the grid gains focus.
		// description:
		//		We use event on a higher node do add/remove tabindex and switch focus to some lower node.
		//		This is better than leaving tabindex on a single node within the grid, because then it would
		//		be too easy to lose the ability to tab into the grid at all if the one node that has focus were
		//		removed, changed etc.  This way, at all that is localised to this one place.
		if (event && event.target !== this._focusCatcherNode) {
			return;
		}
		var rowNode, rowType, cellNode, focusNode;
		//console.log("_onGridFocus needs to transfer focus to a row/cell");
		// transfer focus to the row/cell described in tabTarget
		focusNode = this._findCurrentFocusNode();
		
		// If we've got column or row node, we can focus on it.  Otherwise need to find the first cell in the grid to focus on.
		if (!focusNode) {
			//		if (!rowNode) {
			// Find first cell in the grid.
			var tabTarget = this._navigationTabTarget;
			rowNode = dojo.query("table", this.domNode)[0];
			tabTarget.rowKey = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
			tabTarget.columnIdMap[rowType] = null;
			if (rowNode) {
				rowType = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
				cellNode = dojo.query("td", rowNode)[0]; // cell focus is on the td
				tabTarget.columnIdMap[rowType] = dojo.attr(cellNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
			}
			focusNode = cellNode ? cellNode : rowNode;
		}
		if (focusNode) {
			// delay moving focus, in case there is a mouse click.  The delay allows us to
			// cancel and override this focus on a mouse click, which prevents the
			// grid from scrolling out from under the mouse click.
			this._allowPendingFocusToContinue = true;
			window.setTimeout(dojo.hitch(this, function(){
				if (!this._allowPendingFocusToContinue) {
					return;
				}
				this._focusOnNode(focusNode);
				this._allowPendingFocusToContinue = false;
			}), 100);
		} else {
//			console.log("Module._onGridFocus has nothing to focus on");
		}
	},
	
	_cancelPendingFocus: function() {
		this._allowPendingFocusToContinue = false;
	},
	
	_initializeKeyPressHandlers: function() {
		// summary:
		//		There are two arrays of handlers that are created
		//		(_internalKeyPressHandlers and _internalKeyDownHandlers),
		//		one for keydown events and one for keypress events.
		//		The purpose is to use keydown for non-printable key handling
		//		and keypress for printable key handling.
		
		// Each entry is a hash with:
		//
		// keyCode, keyChar			Provides either one to identify a key.
		// ctrl, shift, alt			Provide any combination of modifiers.  Absent
		//							is considered false.
		// keyPress:				Flag to determine which handler array the key
		//							event will go into.
		// handlerControlsConsumption:
		//							If set true, the handler is called and the return
		//							value determines whether the search continues for
		//							further handlers for the same key.  A handler should
		//							return true if it chooses not to handle the key.  It
		//							is expected that there will be another handler for
		//							the same key later in the chain, and it will
		//							generally consume the key.
		//
		//							Normally when a key is passed to its handler it is
		//							considered consumed.  If this flag is present
		//							and true, the return value from the handler determines
		//							whether it is consumed.  If the handler returns true
		//							the key is considered not consumed, and search continues
		//							for further handlers for it.  Thus some handlers do
		//							not need to worry about consumption.  If a handler
		//							can return a true value you should consider whether
		//							or not this flag needs to be set.
		//							All this allows us to have a key represented multiple
		//							times in the chain with different handlers.  Also
		//							A single handler can be shared between different
		//							keys.
		//
		//							When keys are overloaded in this way, the order of
		//							entries in this table is important.
		//
		// handlerFunction:			A function that is called to handle the keypress.
		//							If it returns true, the key is not consumed.  If
		//							if returns false-ish (including null or undefined),
		//							the key is considered consumed.  Thus most handlers
		//							can just return nothing.
		// keyPolicies:				An integer that is a set of bit flags indicating
		//							which focus the key is valid for.
		//
		
		// Some values for a bitmask that indicates which focus policy each handler
		// should be used for.
		var MANAGED_FOCUS = 1, UNMANAGED_FOCUS = 2, MANAGE_ROW_SELECTION = 4;

		var allKeyPressHandlers = [
				{	/* SAVE */
					keyChar: 's',
					ctrl: true,
					keyPolicies: MANAGED_FOCUS | UNMANAGED_FOCUS,
					keyPress: true,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onSaveKeypress)
				},
				{	/* Cancel edit (Escape) */
					keyCode: dojo.keys.ESCAPE,
					keyPolicies: MANAGED_FOCUS | UNMANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onCancelKeypress)
				},
				{	/* SPACE - select row for the unmanaged focus case.  If allowRowSelection is set, allow row to be keyboard selected. */
					keyChar: ' ',
					keyPolicies: MANAGE_ROW_SELECTION,
					keyPress: true,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onUnmanagedFocusSelectRowKey)
				},
				{	/* SHIFT-SPACE - select row for the unmanaged focus case.  If allowRowSelection is set, allow row to be keyboard selected. */
					keyChar: ' ',
					shift: true,
					keyPolicies: MANAGE_ROW_SELECTION,
					keyPress: true,
					handlerFunction: dojo.hitch(this, this._onUnmanagedFocusSelectRowKey)
				},

				{	/* LEFT */
					keyCode: dojo.keys.LEFT_ARROW,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onLeftRightArrowKey, true)
				},
				{	/* RIGHT */
					keyCode: dojo.keys.RIGHT_ARROW,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onLeftRightArrowKey, false)
				},
				{	/* UP */
					keyCode: dojo.keys.UP_ARROW,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerFunction: dojo.hitch(this, this._onUpDownArrowKey, true)
				},
				{	/* DOWN */
					keyCode: dojo.keys.DOWN_ARROW,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerFunction: dojo.hitch(this, this._onUpDownArrowKey, false)
				},
				{	/* HOME */
					keyCode: dojo.keys.HOME,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onHomeEndKey, true)
				},
				{	/* END */
					keyCode: dojo.keys.END,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onHomeEndKey, false)
				},
				{	/* CTRL-HOME */
					keyCode: dojo.keys.HOME,
					ctrl: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onCtrlHomeEndKey, true)
				},
				{	/* CTRL-END */
					keyCode: dojo.keys.END,
					ctrl: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onCtrlHomeEndKey, false)
				},
				{	/* PAGE UP */
					keyCode: dojo.keys.PAGE_UP,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onPageUpDownKey, true)
				},
				{	/* PAGE DOWN */
					keyCode: dojo.keys.PAGE_DOWN,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onPageUpDownKey, false)
				},
				{	/* SPACE - Invoke cell action */
					keyChar: ' ',
					keyPolicies: MANAGED_FOCUS,
					keyPress: true,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onCellActionKey)
				},
				{	/* Invoke cell action - ENTER */
					keyCode: dojo.keys.ENTER,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					meta: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onCellActionKey)
				},
				{	/* SPACE - select row */
					keyChar: ' ',
					keyPolicies: MANAGED_FOCUS,
					keyPress: true,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onSelectRowKey)
				},
				{	/* SHIFT-SPACE - select row */
					keyChar: ' ',
					shift: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: true,
					handlerFunction: dojo.hitch(this, this._onSelectRowKey)
				},
				{	/* SHIFT-UP - select and move up */
					keyCode: dojo.keys.UP_ARROW,
					shift: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerFunction: dojo.hitch(this, this._onSelectAndMoveKey, true)
				},
				{	/* SHIFT-DOWN - select and move down */
					keyCode: dojo.keys.DOWN_ARROW,
					shift: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerFunction: dojo.hitch(this, this._onSelectAndMoveKey, false)
				},
				{	/* F2 - switch to actionable mode */
					keyCode: dojo.keys.F2,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerFunction: dojo.hitch(this, this._onStartActionModeKey)
				},
				{	/* ENTER - switch to actionable mode */
					keyCode: dojo.keys.ENTER,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerFunction: dojo.hitch(this, this._onStartActionModeKey)
				},
				{	/* TAB - navigate cells in action mode */
					keyCode: dojo.keys.TAB,
					keyPolicies: MANAGED_FOCUS,
					keyPress: true,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onActionModeTabKey, false)
				},
				{	/* SHIFT-TAB - navigate cells in action mode */
					keyCode: dojo.keys.TAB,
					shift: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: true,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onActionModeTabKey, true)
				},
				{	/* SELECT ALL */
					keyChar: 'a',
					ctrl: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: true,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onSelectAllKey, false)
				},
				{	/* SHIFT-F10 */
					keyCode: dojo.keys.F10,
					shift: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onOpenMenuKey)
				},
				{	/* CTRL-DOWN */
					keyCode: dojo.keys.DOWN_ARROW,
					ctrl: true,
					keyPolicies: MANAGED_FOCUS,
					keyPress: false,
					handlerControlsConsumption: true,
					handlerFunction: dojo.hitch(this, this._onStartIntraCellNavigationKey)
				}
		];
		
		var length = allKeyPressHandlers.length,
			activeKeyPressHandlers = [],
			activeKeyDownHandlers = [],
			currentFocusPolicy;

		currentFocusPolicy = this.manageFocusNavigation ? MANAGED_FOCUS : UNMANAGED_FOCUS;
		if (this.manageRowSelection) {
			currentFocusPolicy |= MANAGE_ROW_SELECTION;
		}
		
		// Select only those key handlers that match the focus policy for the grid.
		for (var i = 0 ; i < length ; i++) {
			var handler = allKeyPressHandlers[i];
			if (handler.keyPolicies & currentFocusPolicy) {		// bitmask check so use single &
				if (handler.keyPress) {
					activeKeyPressHandlers.push(handler);
				} else {
					activeKeyDownHandlers.push(handler);
				}
			}
		}
		
		this._internalKeyPressHandlers = activeKeyPressHandlers;
		this._internalKeyDownHandlers = activeKeyDownHandlers;
	},
	
	_internalKeyEventHandler: function(event, isKeyDownEvent) {
		// summary:
		//		Generic handler for keypress and keydown, which calls handlers for keys as defined in the _internalKeyPressHandlers table.
		// returns:
		//		true to indicate that the key has *not* been handled and so should be
		//		propagated.
		var len, handler, notConsumed;
		var keyHandlers = isKeyDownEvent ? this._internalKeyDownHandlers : this._internalKeyPressHandlers;
			
		//console.log("Module._internalKeyEventHandler() keyChar: "+ event.keyChar + " keyCode: "+ event.keyCode+ " synthetic: " + event.synthetic + " keyIdentifier: " + event.keyIdentifier +
		//			 " ctrl: " + event.ctrlKey + " shift: " + event.shiftKey + " alt: " + event.altKey);
		len = keyHandlers.length;
		for (var i = 0 ; i < len ; i++) {
			handler = keyHandlers[i];
			if (
				(
						// Look for a matching keyChar or keyCode.  Only look at keyCode if there is no
						// keyChar in the event because there are keyCodes that can appear both with and
						// without a keyChar.  e.g. "q" and F2 both have keyCode == 113.
						(handler.keyChar && (handler.keyChar === event.keyChar)) ||
						(!event.keyChar && handler.keyCode && (handler.keyCode === event.keyCode))
				) &&
				(
						(!!handler.ctrl === !!event.ctrlKey) &&
						(!!handler.shift === !!event.shiftKey) &&
						(!!handler.alt === !!event.altKey) &&
						(!!handler.meta === !!event.metaKey)
				)
				
			) {
				notConsumed = handler.handlerFunction(event);
				if (handler.handlerControlsConsumption && notConsumed) {
					// continue looking for a handler for the key.
					continue;
				}
				// key was given to a handler function, so is considered consumed.
				return false;
			}
		}
		return true;
	},
	
	_onSaveKeypress: function() {
		// summary:
		//		Handler for the save key.
		var deferred = this.onSaveKey();
		// record that the row is no longer in action mode.
		this._navigationTabTarget.actionMode = false;
		return deferred;
	},
	
	_onCancelKeypress: function() {
		// summary:
		//		Handler for the Cancel (Escape) key.
		this.onEscKey();
		// record that the row is no longer in action mode.
		this._navigationTabTarget.actionMode = false;
	},
	
	_onUnmanagedFocusSelectRowKey: function(event) {
		// summary:
		//		Handler for the case where the grid is not managing focus - the browser is managing
		//		focus through normal tab key behaviour - but we are allowing the row to take focus
		//		and be keyboard selected.
		// event:
		//		The key event that brought us here.
		var objectId = dojo.attr(event.target, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		if (!objectId) {
			// keypress was not on the top-level table element for the row - probably an input field that has focus.
			// So ignore this keypress, and allow it to propagate.
			return true;
		}
		
		// toggle selection of the row.
		if (this.selectedRows[objectId]) {
			this.removeFromSelection([objectId]);
		} else {
			this.addToSelection([objectId]);
		}
	},
	
	_onLeftRightArrowKey: function(/*Boolean*/goLeft) {
		// summary:
		//		Moves focus to the next cell horizontally.
		// goLeft:
		//		true to move left, false to move right
		var rowInfo = this._getCurrentRowInfo(),
			cellInfo = this._getCurrentCellInfo(rowInfo);
		if (rowInfo.actionMode) {
			// don't move using arrows when row in action mode - let the arrow
			// key bubble up.
			return true;
		}
		if (rowInfo && cellInfo) {
			if (rowInfo.rowFocus) {
				// We are focussed on a row. Move to the first or last cell - same as
				// Home/End keys
				return  this._onHomeEndKey(!goLeft);
			}
			var nextColNode = goLeft ? cellInfo.node.previousSibling : cellInfo.node.nextSibling;
			if (nextColNode) {
				var nextColId = dojo.attr(nextColNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
				this._moveGridFocus(rowInfo.key, rowInfo.type, nextColId);
			}
		}
	},
	
	_onUpDownArrowKey: function(goUp) {
		// summary:
		//		Moves the cell focus up or down
		// goUp:
		//		true to go up, false to go down
		var nextRowNode,
			currentRowInfo = this._getCurrentRowInfo(),
			currentRowNode = currentRowInfo.node;
		if (currentRowInfo.actionMode) {
			// in action mode - cannot move vertically
			return;
		}
		nextRowNode = goUp ? this._findPreviousSiblingObject(currentRowNode) : this._findNextSiblingObject(currentRowNode);
		if (!nextRowNode) {
			// at top or bottom of the current grid display.  This could be the end of the grid, or could
			// be a page boundary in a virtually-paged grid.  Call a handler to allow virtual grid to determine
			// the next move.
			this.onKeyUpDownAtGridEnd(goUp, currentRowInfo.key);
			return;
		}
		
		this.focusCorrespondingCellInNewRow(nextRowNode);
	},
	
	focusCorrespondingCellInNewRow: function(nextRowNode) {
		// summary:
		//		Moves focus to a new row with due regard for the cell that is currently in focus on
		//		the currently focused row.  Attempts to focus on the same cell in the new row.  If the row type
		//		is different, selects the focus cell based on the last accessed cell for that row type.
		var nextCellNode, nextColId, nextRowType,
			currentRowInfo = this._getCurrentRowInfo();
		if (!currentRowInfo.rowFocus) {
			nextRowType = dojo.attr(nextRowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			var tabTarget = this._navigationTabTarget;
			// currently on a cell, find a cell in the row below
			if (nextRowType === currentRowInfo.type) {
				// Find the corresponding cell on the next row
				nextCellNode = dojo.query("td[colid='" + tabTarget.columnIdMap[nextRowType] + "']", nextRowNode)[0];
				nextColId = dojo.attr(nextCellNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
			} else {
				// find a row/cell in the row below which is of a different row type
				nextCellNode = dojo.query("td[colid='" + tabTarget.columnIdMap[nextRowType] + "']", nextRowNode)[0];
				if (!nextCellNode) {
					// just use the first cell in the row
					nextCellNode = dojo.query("td", nextRowNode)[0];
				}
				nextColId = dojo.attr(nextCellNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
			}
		} // else focus is currently on a row, so leave nextRowType and nextColId undefined so we get row focus
		var nextRowKey = dojo.attr(nextRowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		this._moveGridFocus(nextRowKey, nextRowType, nextColId);
		
	},
	
	onKeyUpDownAtGridEnd: function(currentRowKey) {
		// summary:
		//		Called when there is an key up/down event at the extreme top/bottom row in the grid.  Such a
		//		keypress may need to move focus across a virtual page boundary.
		//		*** Sub-class should override this to handle the virtual grid scenarios. ***
		//		*** This default implementation does nothing, so leaving the focus unchanged. ***
	},

	_moveFocusToFirstCell: function(rowInfo) {
		var cells = dojo.query("td[colid]", rowInfo.node);
		var cellNode = cells[0];
		var nextColId = dojo.attr(cellNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
		this._moveGridFocus(rowInfo.key, rowInfo.type, nextColId);
	},

	_moveFocusToLastCell: function(rowInfo) {
		var cells = dojo.query("td[colid]", rowInfo.node);
		var cellNode = cells[cells.length-1];
		var nextColId = dojo.attr(cellNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
		this._moveGridFocus(rowInfo.key, rowInfo.type, nextColId);
	},
	
	_onHomeEndKey: function(goHome) {
		// summary:
		//		Handler for the home or end keys.  Moves the focus to the start or end of the
		//		current row.
		// goHome:
		//		true to go the the home (leftmost) cell.  false to go to the end (rightmost) cell.
		var rowInfo = this._getCurrentRowInfo();
		if (rowInfo.actionMode) {
			// don't move using home/end when row in action mode - let the
			// key bubble up.
			return true;
		}
		if (rowInfo) {
			if (goHome) {
				this._moveFocusToFirstCell(rowInfo);
			} else {
				this._moveFocusToLastCell(rowInfo);
			}
		}
	},
	
	_onCtrlHomeEndKey: function(goHome) {
		// summary:
		//		Handler for Ctrl-Home and Ctrl-End keys.  Moves focus to the extreme top or bottom
		//		row.
		this.setFocusForCtrlHomeEnd(goHome);
	},
	
	setFocusForCtrlHomeEnd: function(goHome) {
		// summary:
		//		General grid handler for the Ctrl-Home and Ctrl-End keys.  May be overridden by
		//		subclass to provide behaviour in virtual grid.
		var rowCount = this.rowOrder.length;
		if (rowCount === 0) {
			return;
		}
		var targetRowKey = this.rowOrder[goHome ? 0 : (rowCount - 1)].key;
		var newRowNode = this.allRows[targetRowKey].node;
		this.focusCorrespondingCellInNewRow(newRowNode);
	},
	
	_onPageUpDownKey: function(goUp) {
		// summary:
		//		Handler for page up and down keys, causing focus to move to a corresponding
		//		cell/row about one-viewport distance up or down.
		var	currentRowInfo = this._getCurrentRowInfo();
		this.setFocusForPageUpDown(goUp, currentRowInfo.key);
	},
	
	setFocusForPageUpDown: function(goUp, currentRowKey) {
		// summary:
		//		Find the row that is approximately one viewport distance up/down from the given row,
		//		and set focus to it.
		//		*** Sub-class should override this to handle the virtual grid scenarios. ***
		// goUp:
		//		Indicates whether we are doing a page up or page down from the current position.
		// currentRowKey:
		//		Key for the current row position.
		
		// In the distance calculations we take into account the height of only the target *or* the start row, depending
		// on the direction we are going.  This ensures that page up and page down move corresponding rows, even if the
		// rows are differing heights, and a pageUp followed by pageDown will bring focus back to the same row (unless
		// it hits the top or bottom).
		
		var curNode, prevNode, targetNode,
			currentRowInfo = this._getCurrentRowInfo(),
			distanceToMove = this.domNode.parentNode.offsetHeight;
			
		prevNode = currentRowInfo.node;
		if (goUp) {
			// going up - exclude current node, and include target node in measurement
			while (true) {
				curNode = this._findPreviousSiblingObject(prevNode);
				if (!curNode) {
					// at top of available rows - stick at the top row.
					targetNode = prevNode;
					break;
				}
				distanceToMove -= curNode.offsetHeight;
				if (distanceToMove <= 0) {
					// reached a viewport distance away from start - so found our target
					targetNode = curNode;
					break;
				}
				prevNode = curNode;
			}
		} else {
			// going down - include current node, and exclude target node from measurement
			while (true) {
				distanceToMove -= prevNode.offsetHeight;
				if (distanceToMove <= 0) {
					// target is the next node (if there is one).
					targetNode = this._findNextSiblingObject(prevNode);
					if (!targetNode) {
						targetNode = prevNode;
					}
					break;
				}
				curNode = this._findNextSiblingObject(prevNode);
				if (!curNode) {
					// at bottom of available rows - stick at the bottom row
					targetNode = prevNode;
					break;
				}
				prevNode = curNode;
			}
		}
		if (targetNode !== currentRowInfo.node) {
			this.focusCorrespondingCellInNewRow(targetNode);
		}
	},
	
	_onSelectRowKey: function() {
		// summary:
		//		Handle keypress that select the row that is currently in focus.
		//		Either selects or deselects the row.
		var tabTarget = this._navigationTabTarget;
		if (!tabTarget.rowKey) {
			return;
		}
		if (tabTarget.actionMode) {
			// don't select/deselect when row in action mode - let the
			// key bubble up.
			return true;
		}
		var currentRowNode = dojo.query("[objectId='"+tabTarget.rowKey+"']", this.domNode)[0];
		var objectId = dojo.attr(currentRowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		if (this.selectedRows[objectId]) {
			this.removeFromSelection([objectId]);
		} else {
			this.addToSelection([objectId]);
		}
	},
	
	_onSelectAndMoveKey: function(goUp) {
		// summary:
		//		Handle keypress to select row that currently has focus and
		//		move focus up or down to the next row.
		if (this._onSelectRowKey()) {
			return true;
		}
		return this._onUpDownArrowKey(goUp);
	},
	
	_onCellActionKey: function() {
		// summary:
		//		Invokes the action for a cell.
		// returns:
		//		True if no action function is called on the cell.  This is
		//		necessary because there are cases where keys are overloaded,
		//		so this handler must not consume the key if it does not apply
		//		to the current cell, but should allow it to be passed on to
		//		a later handler.
		var rowInfo = this._getCurrentRowInfo(),
			cellInfo = this._getCurrentCellInfo(rowInfo);
		if (rowInfo.actionMode) {
			// don't perform action when row in action mode - let the
			// key bubble up.
			return true;
		}
		if (cellInfo) {
			if (cellInfo.columnId === com.ibm.rdm.web.grid.GridModel.COL_SELECTOR) {
				// special case for the selector column
				// toggle the selector - this is exactly the same as we do for
				// other select row keys.
				this._onSelectRowKey();
				return;
			} else if (cellInfo.columnId === com.ibm.rdm.web.grid.GridModel.COL_MENU) {
				// special case for invoking the row menu
				var menuLauncherNode = dojo.query("img.menuLauncher", cellInfo.node)[0];
				this._openRowActionMenu(null, rowInfo.node, rowInfo.type, menuLauncherNode);
				return;
			} else if (dojo.hasClass(cellInfo.node, "expandoContainer")) {
				// Special case for expandos.  Expando can be in three states - "open", "closed"
				// and "none".  For the "none" case the expando is inactive, so we fall through to
				// the general action handling for the cell, allowing any action specified there
				// to occur.
				var expandoNode = dojo.query(".expandoWrapper img", cellInfo.node)[0];
				if (dojo.hasClass(expandoNode, "open") || dojo.hasClass(expandoNode, "closed")) {
					this.applyExpando(rowInfo.type, rowInfo.key);
					return;
				}
			}
			// The usual case - look for a function on the grid model and execute it if it is there.
			var actionFunction = this.gridmodel.getNavModeActionFunction(rowInfo.type, cellInfo.columnId);
			if (actionFunction) {
				actionFunction(rowInfo.key, cellInfo.columnId, cellInfo.node);
				// key has been handled so is consumed
				return;
			}
		}
		// key is not consumed
		return true;
	},
	
	_onStartActionModeKey: function() {
		// summary:
		//		Switch the row to "Actionable Mode" and start editing on the current cell.
		// description:
		//		In addition to opening an editor on the current cell (and setting
		//		focus to it), this function also creates editors for all of other
		//		cells within the row that are editable by the user.  We do this so
		//		that the user can more readily know what navgiation would be
		//		performed by Tab and Shift-Tab.
		
		var rowInfo = this._getCurrentRowInfo(),
			focusCellInfo = this._getCurrentCellInfo(rowInfo),
			i;
		// Find out if the current cell is editable.
		if (dojo.hasClass(focusCellInfo.node, com.ibm.rdm.web.grid.Module.ClassNames.CELL_EDITABLE)) {
			// Record that the row is now in action mode.
			this._navigationTabTarget.actionMode = true;
			// Start editing on the current cell (and open editors for
			// other cells within the row that are editable).
			var cols = this.gridmodel.getUserCols(RESOURCE_ROW),
				artifactURI = rowInfo.key;
			for (i = 0; i < cols.length; i++) {
				var col = cols[i],
					cellInfo = this._getCellInfo(rowInfo, col.colId);
				if (cellInfo.node
					&& dojo.hasClass(cellInfo.node, com.ibm.rdm.web.grid.Module.ClassNames.CELL_EDITABLE)
					) {
					var editStarted =
						this.gridmodel.invokeEdit(
							rowInfo.type, col.colId, artifactURI, cellInfo.node, null, false
						);

					if (editStarted && focusCellInfo.columnId === col.colId) {
						this.notifyEditStartedOnCell(artifactURI, col.colId);
					}
				}
			}
			var refresh = [{key: artifactURI, rowtype: RESOURCE_ROW}];
			this.refreshRows(refresh);
			//console.log("_onStartActionModeKey just invoked edit.  Focus is ",  document.activeElement);
		}
	},
	
	_onActionModeTabKey: function(goBackward, event) {
		// summary:
		//		Handler for the Tab key.  Moves between cells in action mode.
		// goForward:
		//		false for a normal tab.  true for a backward tab (shift-tab).
		var rowInfo = this._getCurrentRowInfo();
		//console.log("_onActionModeTabKey: ", goBackward, " Synthetic: ", event.synthetic, " Target: ", event.target);
		if (!rowInfo || !rowInfo.actionMode) {
			// Not in action mode - propagate the key to allow normal
			// tab handling to occur.
			return true;
		}
		// Find the next actionable column for the current row.
		// The tab may originate in any column that is being edited, and that may not
		// be the column that we last set fucus to - because focus may have been moved
		// by a mouse click.  So find the column that originated the tab key event.
		
		var tabCellNode = this._findCellNodeForEvent(event, rowInfo);
		
		//var cellInfo =  this._getCurrentCellInfo(rowInfo);
		if (tabCellNode) {
			var nextCellNode = tabCellNode;
			while (true) {
				nextCellNode = goBackward ? nextCellNode.previousSibling : nextCellNode.nextSibling;
				if (!nextCellNode) {
					// We tried all cells in that direction.  Need to leave action mode
					// and transfer focus to the row's first cell.
					this._moveFocusToFirstCell(rowInfo);
					this.onSaveKey();
					return;
				}
				var columnId = dojo.attr(nextCellNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
				var isUnderEdit = this.gridmodel.isUnderEdit(rowInfo.type, columnId, rowInfo.key);
				var isEditable = dojo.hasClass(nextCellNode, com.ibm.rdm.web.grid.Module.ClassNames.CELL_EDITABLE);
				if (isUnderEdit || isEditable) {
					// Found editable cell, focus on it and start edit.
					var nextColId = dojo.attr(nextCellNode, com.ibm.rdm.web.grid.Module.AttributeNames.COLUMNID);
					this._moveGridFocus(rowInfo.key, rowInfo.type, nextColId, true /*isActionMode*/, event);
					if (!isUnderEdit) {
						this.gridmodel.invokeEdit(rowInfo.type, nextColId, rowInfo.key, nextCellNode);
					}
					return;
				}
			}
		}
		return true;
	},
	
	_onOpenMenuKey: function() {
		// summary:
		//		Handler for key to open the menu on the row.  Effectively brings the menu column into
		//		focus and activates it.
		var rowInfo = this._getCurrentRowInfo();
		if (rowInfo.actionMode) {
			// don't do anything when row in action mode - let the
			// key bubble up.
			return true;
		}
		this._moveGridFocus(rowInfo.key, rowInfo.type, com.ibm.rdm.web.grid.GridModel.COL_MENU, rowInfo.actionMode);
		return this._onCellActionKey();
	},
	
	_onStartIntraCellNavigationKey: function() {
		// summary:
		//		Handler for key to start navigation within a cell.
		var rowInfo = this._getCurrentRowInfo(),
			cellInfo = this._getCurrentCellInfo(rowInfo);
		if (rowInfo.actionMode) {
			// don't do anything when row in action mode - let the
			// key bubble up.
			return true;
		}
		var startNavigationFunction = this.gridmodel.getStartIntraCellNavigationFunction(rowInfo.type, cellInfo.columnId);
		if (!startNavigationFunction) {
			// column doesn't do this sort of navigation.  Do nothing - let key bubble up.
			return true;
		}
		startNavigationFunction(dojo.hitch(this, this._endIntraCellNavigation), cellInfo.node);
	},
	
	_endIntraCellNavigation: function(/*?String*/direction) {
		// summary:
		//		Callback handler passed to the startIntraCellNavigationFunction.  Is invoked when a cell that is
		//		doing intra-cell navigation wants to return focus to the grid.
		// direction:
		//		Optional string indicating whether focus should be moved or left at the current cell.
		//		"left" and "right" are recognised and any other value puts focus on the current cell (i.e. the
		//		cell that was doing the intra-cell navigation.
		
		this.retakeFocus();
		if (direction === "left") {
			this._onLeftRightArrowKey(true);
		} else if (direction === "right") {
			this._onLeftRightArrowKey(false);
		}
	},
	
	_onSelectAllKey: function() {
		// summary:
		//		Handle keypress that selects all items in the grid.

		// Delegate handling to a function that may be overridden, to that
		// sub-class can do appropriate handling for virtual grid.

		// This event should only be fired if an artifact is not currently under edit.
		if(!this.isInEditMode()) {
			return this.handleSelectAllKey();
		} else {
			// True tells the handler NOT to consume (dojo.stopEvent) this event,
			// which allows the browser to handle the event natively.
			return true;
		}
	},
	
	handleSelectAllKey: function() {
		// summary:
		//		Handle keypress to select all items in the grid.  Should be overridden
		//		by sub-class to provide approprate handling for virtual grid.
		// returns:
		//		undefined (false-ish) always in this implementation, to indicate that
		//		the key has been handled.
		this.selectAll(true);
	},
	
	_saveAllRows: function() {
		// summary:
		//		Saves any rows that are under edit, and switch the current row to
		//		navigate mode.
		var rowInfo = this._getCurrentRowInfo();
		var cellInfo = this._getCurrentCellInfo(rowInfo);
		this._moveGridFocus(rowInfo.key, rowInfo.type, cellInfo.columnId, false/* isActionMode*/);
		this.onSaveKey();
	},
	
	_cancelAllEdits: function() {
		// summary:
		//		Cancels edits to any rows that are under edit.
		var rowInfo = this._getCurrentRowInfo();
		var cellInfo = this._getCurrentCellInfo(rowInfo);
		this._moveGridFocus(rowInfo.key, rowInfo.type, cellInfo.columnId, false/* isActionMode*/);
		this.onEscKey();
	},
	
	_focusOnNode: function (targetFocusNode) {
		// summary:
		//		Changes the cell/row that is in focus.
		//		CHANGES TO FOCUS MUST GO THROUGH THIS METHOD, SO THAT
		//		PROPER BLUR HANDLING IS APPLIED.  WITHOUT BLUR HANDLING
		//		TABBING INTO AND OUT OF THE GRID WILL BREAK.
		dojo.attr(targetFocusNode, "tabIndex", "0");
		// remove ability to take focus from _focusCatcherNode, so that shift-tab doesn't bring us back here.
		dojo.attr(this._focusCatcherNode, "tabIndex", "-1");

		targetFocusNode.focus();

		// Note: see Defect 86926 to find out why we're making sure to call the
		// 'onNodeFocused' function here.
		this.onNodeFocused(targetFocusNode);

		var handle = this.connect(targetFocusNode, "onblur", function(event){
			// Put focus back onto the _focusCatcherNode, so that it routes focus next time.
			// Only do this if the current focus has not been moved to another node within the
			// grid before the blur event is received and if the document still has focus.
			var currentFocusNode = this._findCurrentFocusNode();
			if (dojo.doc.hasFocus()) {
				if (targetFocusNode === currentFocusNode) {
					dojo.attr(this._focusCatcherNode, "tabIndex", "0");
				}
				dojo.removeAttr(targetFocusNode, "tabIndex");
			}
			this.disconnect(handle);
		});
	},
	
	retakeFocus: function() {
		// summary:
		//		Re-establishes focus on the current row or cell in the grid.
		
		if (!this.manageFocusNavigation) {
			// We are not managing focus, so don't try to retake it.
			return;
		}
		
		var rowInfo = this._getCurrentRowInfo(),
			cellInfo = this._getCurrentCellInfo(rowInfo);
			
		if (!rowInfo) {
			// There is no current row set for the grid.  Probably a new grid.  Don't grab focus, because it may
			// have been created as a side-effect of an operation somewhere else on the page.
			// Just make sure the focus catcher will catch tabs into the grid.  This should not be necessary on a
			// new grid, but let's do it just in case.
			dojo.attr(this._focusCatcherNode, "tabIndex", "0");
			return;
		}
		if (!rowInfo.node) {
			// The row no longer exists in the grid.
			dojo.attr(this._focusCatcherNode, "tabIndex", "0");
			return;
		}
		if (rowInfo.rowFocus) {
			this._moveGridFocus(rowInfo.key);
		} else {
			this._moveGridFocus(rowInfo.key, rowInfo.type, cellInfo.columnId, rowInfo.actionMode);
		}
	},
	
	focus: function() {
		// summary:
		//		Take focus.  Focus is moved to the row/cell that was last in
		//		focus, or set for focus.
		this.retakeFocus();
	},
	
	simulateTabIntoGrid: function() {
		// summary:
		//		Simulate the user tabbing into the grid.
		// description:
		//		We use a FirstArtifactInModuleHelper based view (not grid based
		//		view) when we show the contents of an empty module.  When the
		//		user uses "Add Existing Artifact" or "Paste" from that view
		//		we show a new row but don't ALSO give this row focus.  This
		//		method can be used to give focus to that first row by simulating
		//		what would occur if the user were to tab into the grid.

		this._onGridFocus();
	},

	moveFocusToCellOnRow: function(/*String*/rowKey) {
		// summary:
		//		Moves the focus cell of the grid to a cell in the row indentified by
		//		rowKey.  The column chosen is the column that was last focussed on that
		//		type of row - typically this will be the same column that was previously
		//		in focus on another row.  If a cell cannot be determinied, the whole
		//		row is given focus.
		//		Will take focus if possible, but if the row is not yet displayed, still
		//		sets the grid such that the next time it takes focus, the given cell will
		//		be focused.
		// rowKey:
		//		Identifier for the row on which to focus.
		// returns:
		//		true if focus was taken.
		var tabTarget = this._navigationTabTarget;
		var oldRowKey = tabTarget.rowKey;
		tabTarget.rowKey = rowKey;
		tabTarget.actionMode = false;
		// leave column and rowFocus information as-is, so that focus moves to
		// similar column on the new row.
		
		// Move focus if the row is on display.
		var rowNode = dojo.query("[objectId='" + rowKey + "']", this.domNode)[0];
		if (rowNode) {
			var rowType = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			var columnId = tabTarget.columnIdMap[rowType];
			this._moveGridFocus(rowKey, null, tabTarget.rowFocus ? null : columnId);
			return true;
		}
		// else don't focus, but assume the grid will retake focus at a suitable time later
		// when it has finished refreshing and the row is displayed.
	},
	
	_moveGridFocus: function(rowKey, rowType, columnId, /*Boolean?*/isActionMode, /*?event*/event) {
		// summary:
		//		Changes the grid focus to the given row and column.  If no column information is
		//		supplied, focus is on the whole row.
		// rowKey:
		//		Key (not necessarily object ID) for the row to focus on.
		// rowType:
		//		optional type for the row.  May be supplied for cell focus.  If available,
		//		it should be supplied to avoid a repeat lookup of the rowType.  If not
		//		available to the caller, may be set to null, in which case it is looked
		//		up from to row if required.
		// columnId:
		//		optional id for the column.  Must be supplied for cell focus.
		// isActionMode:
		//		optional flag to record that the new row is in action mode.  defaults to false-ish.
		// event:
		//		optional event that caused focus to move.  When focusing on an editor this is passed to
		//		the editor, and may influence which field within the editor is given focus.
		
		//console.log("moveGridFocus");
		// remove focus from the current focus node
		var tabTarget = this._navigationTabTarget;
		// Update the recorded target node, and set focus to it.
		if (columnId) {
			// moving focus to a cell.
			// Get the rowType if we've not been given it.
			if (!rowType) {
				var rowNode = dojo.query("[objectId='" + rowKey + "']", this.domNode)[0];
				if (!rowNode) {
					// We seem to have been asked to focus on a row that does not yet exist in the
					// grid.  Not much we can do but ignore this call, leaving the navigation
					// target unchanged.
					return;
				}
				rowType = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			}
			
			// special case for resourceRow and relatedRow Types.  Treat these as equivalent for
			// keyboard navigation.  They have the same columns, so by setting columnID for them
			// both to be identical we ensure that vertical scrolling through them will pick the same
			// columns.
			if (rowType === "resourceRow" || rowType === "relatedRow") {
				tabTarget.columnIdMap["resourceRow"] = columnId;
				tabTarget.columnIdMap["relatedRow"] = columnId;
			} else {
				tabTarget.columnIdMap[rowType] = columnId;
			}
			tabTarget.rowFocus = false;
		} else {
			// moving focus to a row
			tabTarget.rowFocus = true;
		}
		tabTarget.rowKey = rowKey;
		tabTarget.actionMode = isActionMode;
		
		var newFocusNode = this._findCurrentFocusNode();
		var focused;
		
		if (newFocusNode) {
			if (isActionMode) {
				// In action mode focus on editor if it has been started, but if it has not yet
				// been started focus on the cell.
				focused = this.gridmodel.focusOnEditor(rowType, columnId, rowKey, newFocusNode, event);
				//console.log("_moveGridFocus: ", focused, columnId);
			}
			if (!focused) {
				this._focusOnNode(newFocusNode);
			}
		}
	},
	
	_findCurrentFocusNode: function() {
		var cellNode,
			tabTarget = this._navigationTabTarget,
			rowNode = dojo.query("[objectId='"+tabTarget.rowKey+"']", this.domNode)[0];
		
		// Don't bother looking for a cell if current focus is the row.
		if (rowNode && !tabTarget.rowFocus) {
			var rowType = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			var columnId = tabTarget.columnIdMap[rowType];
			if (columnId) {
				cellNode = dojo.query("td[colid='" + columnId + "']", rowNode)[0];
			}
		}
		var focusNode = cellNode ? cellNode : rowNode;
		return focusNode;
	},
	
	_findCellNodeForEvent: function(event, rowInfo){
		// summary:
		//		Find the cell node within the given row corresponding to the cell in which an
		//		event (e.g. keypress) originated.
		// event:
		//		An event.
		// rowInfo:
		//		rowInfo for the row in which the event is expected to have originated.
		// returns:
		//		The cell node (td) in which the event originated.  Undefined if the cell cannot
		//		be determined.
		
		if (!rowInfo || !rowInfo.node) {
			return;
		}
		var curCellNode,
			eventSource = event.target || event.srcElement || event.originalTarget;
		var cellNodes = dojo.query("td[colid]", rowInfo.node);
		for (var i = 0 ; i < cellNodes.length ; i++) {
			curCellNode = cellNodes[i];
			if (dojo.isDescendant(eventSource, curCellNode)) {
				return curCellNode;
			}
		}
		// didn't find a match - implicitly return undefined.
	},
	
	makeLinkColDropTarget: function(colTD) {
		this._toDestroy.push(new com.ibm.rdm.web.common.dnd.DropZone({
				node: colTD,
				accepts: this.dragDropArgs.validDropTypes,
				onDrop: dojo.hitch(this, function(ev) {
					var linkColID = colTD.getAttribute("colid");
					var parent = dojo.query(colTD).parents("table")[0];
					this.dragDropArgs.onLinkDrop(ev, dojo.attr(parent, "objectid"), linkColID);
					this._buildLinkColTarget();
					this._linkDropEvent = true;
				}),
				onDragOver: function(ev){
					// Kill the IE event here...
					if(dojo.isIE) {
						ev.stopPropagation();
					}
					
					if (ev.dataTransfer) {
						ev.dataTransfer.dropEffect = "link";
					}
				},
				onDragEnter: dojo.hitch(this, function(ev){
					dojo.place(this._linkColTarget, colTD.firstChild, "first");
					var parent = dojo.query(colTD).parents("table")[0];
					this.dragOverCommon(ev, parent, true);
					dojo.addClass(colTD, "link-active");
					this._linkColEvent = true;
				})
		}));
	},
	
	setDragDropArgs: function(dragDropArgs) {
		this.dragDropArgs = dragDropArgs;
		
		// Destroy them if we have them. We would hate to leak these nodes
		dojo.destroy(this._dndWrapper);
		dojo.destroy(this._dragPointer);
		dojo.destroy(this._linkBox);

		
		this._dndWrapper = dojo.create("div", {
			"class": "dndWrapper"
		});
		
		this._linkBox = dojo.create("div", {
			"class": "link",
			innerHTML: "<img src=" + this._blankGif + " class='addLinksCommand'>"
		}, this._dndWrapper);
		
		this._buildLinkColTarget();
		
		// Hook up Link Target
		this._toDestroy.push(new com.ibm.rdm.web.common.dnd.DropZone({
				node: this._linkBox,
				accepts: this.dragDropArgs.validDropTypes,
				onDrop: dojo.hitch(this, function(ev) {
					
					var nl = new dojo.NodeList(ev.target),
					dropTarget = this._dndWrapper.previousSibling;
					
					if (!dropTarget || !dojo.attr(dropTarget, "objectid")) {
						dropTarget = this._dndWrapper.nextSibling;
					}
					dojo.removeClass(this.domNode, "link-box-active");
					this.dragDropArgs.onLinkDrop(ev, dojo.attr(dropTarget, "objectid"));
				}),
				onDragOver: function(ev) {
					if (ev.dataTransfer) {
						ev.dataTransfer.dropEffect = "link";
					}
				},
				onDragEnter: dojo.hitch(this, function(ev){
					// this should check if we can actually create a link before doing this...
					dojo.addClass(this.domNode, "link-box-active");
					dojo.addClass(this._linkBox, "active");
					dojo.query("td.link-active", this.domNode).removeClass("link-active");
					this._linkBoxEvent = true;
				}),
				onDragLeave: dojo.hitch(this, function(ev){
					if(ev.target === this._linkBox) {
						// We are still in linkBox...
						return;
					}
					dojo.removeClass(this.domNode, "link-box-active");
					dojo.removeClass(this._linkBox, "active");
					this._linkBoxEvent = null;
				})
		}));
		
		
		if(this.dragDropArgs.isDropEnabled()) {
			this._dragPointer = dojo.create("div", {
				"class": "moduleDropPointer"
			}, this._dndWrapper);
			
			var innerBox = dojo.create("div", {
				"class": "inner"
			}, this._dragPointer);
			
			// User should be able to drop on the pointer
			this._toDestroy.push(new com.ibm.rdm.web.common.dnd.DropZone({
					node: this._dragPointer,
					accepts: this.dragDropArgs.validDropTypes,
					onDragOver: this.dragDropArgs.onDragOver,
					onDrop: dojo.hitch(this, function(ev) {
						var nl = new dojo.NodeList(ev.target),
						dropTarget = this._dndWrapper.previousSibling,
						beforeFirst = false;
						
						if (!dropTarget || !dojo.attr(dropTarget, "objectid")) {
							dropTarget = this._dndWrapper.nextSibling;
							beforeFirst = true;
						}
						
						this.dragDropArgs.onDrop(ev, dojo.attr(dropTarget, "objectid"), beforeFirst);
					})
			}));
		}
	},
	
	updateLinkBoxPosition: function(){
		if(this.dragDropArgs && this.dragDropArgs.isModule && this._linkBox){
			if(this.domNode.scrollWidth > this.domNode.offsetWidth) {
				var rightOffset = ((this.domNode.scrollWidth - this.domNode.offsetWidth) + 5);
				this._linkBox.style.right = rightOffset + "px";
			} else {
				this._linkBox.style.right = "";
			}
		}
	},
	
	_buildLinkColTarget: function() {
		dojo.destroy(this._linkColTarget);
		this._linkColTarget = dojo.create("div", {
			"class": "lk linkColBox",
			innerHTML: "<img src=" + this._blankGif + " class='addLinksCommand'>"
		});
	},
	
	_isFocusCell: function(key, columnId) {
		// summary:
		//		indicates whether the given row and column is the current focus for the grid.
		
		var rowInfo = this._getCurrentRowInfo(),
			cellInfo = this._getCurrentCellInfo(rowInfo);
		return (rowInfo && cellInfo && key === rowInfo.key && columnId ===  cellInfo.columnId);
	},

	_getCurrentRowInfo: function() {
		// summary:
		//		Gets information about the current row.
		// returns:
		//		undefined if there is no current row.  If there is a current row,
		//		returns a hash containing node, type and key for the row.
		var rowType, objectId, rowNode,
			tabTarget = this._navigationTabTarget;
		if (!tabTarget.rowKey) {
			return;
		}
		rowNode = dojo.query("[objectId='" + tabTarget.rowKey + "']", this.domNode)[0];
		if (rowNode) {
			rowType = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.ROWTYPEID);
			objectId = dojo.attr(rowNode, com.ibm.rdm.web.grid.Module.AttributeNames.OBJECT_ID);
		}
		return {
			node: rowNode,
			type: rowType,
			key: objectId,
			actionMode: tabTarget.actionMode,
			rowFocus: tabTarget.rowFocus
		};
	},

	_getCellInfo: function(rowInfo, columnId) {
		var cellNode = dojo.query("td[colid='" + columnId + "']", rowInfo.node)[0];
		return {columnId: columnId, node: cellNode};
	},
	
	_getCurrentCellInfo: function(rowInfo) {
		// summary:
		//		Gets information about the current cell.
		// rowInfo:
		//		Row information as returned from _getCurrentRowInfo().
		// returns:
		//		A hash containing columnId and node.  Returns undefined
		//		if there is no current column.
		if (!rowInfo) {
			return;
		}
		var tabTarget = this._navigationTabTarget,
			columnId = tabTarget.columnIdMap[rowInfo.type];
		if (!columnId) {
			return;
		}
		return this._getCellInfo(rowInfo, columnId);
	},
	
	/////////////////////////////
	// KEYBOARD NAVIGATION END //
	/////////////////////////////
	
	_getPositionRelativeToScrollParent: function(node) {
		// summary:
		//		For the given node, return the number of pixels
		//		from the top of the grid.
		return node.offsetTop;
	}

});

// set up the constants as attributes of the constructor function
// itself rather than the prototype.

com.ibm.rdm.web.grid.Module.ClassNames = {
	ROW_SELECTED: "row-selected",
	ROW_CHECKED: "row-checked",
	ROW_STARTLINK: "row-startlink",
	UNDER_EDIT: "underEdit",
	SAVING: "saving",
	REFRESHING: "refreshing",
	COLUMN_PREFIX: "column",
	ROW_HOVER: "rowHover",
	ROW_FOCUS: "rowFocus",
	SELECTED_EDIT: "selectedEdit",
	EDITOR_LAUNCHER_HOVER: "hoverOverMenuLauncher",
	MAIN_SELECT: "mainSelect",
	ROW_CUT: "row-cut",
	ROW_COPY: "row-copy",
	CELL_EDITABLE: "cellEditable"
};

/* Constants for HTML Attributes used to convey information
 * from the server.
 */
com.ibm.rdm.web.grid.Module.AttributeNames = {
	OBJECT_ID		: "objectId",
	MENU_TYPE		: "menuType",
	ROWTYPEID       : "rowtype",
	COLUMNID        : "colid",
	FOCUSABLE		: "focusable",
	// The row label, used to build an ARIA row label.
	ROW_LABEL		: "rowLabel",
	ROW_DESCRIPTION : "rowDescription"
};


com.ibm.rdm.web.grid.Module.HAND_CURSOR = "pointer";

com.ibm.rdm.web.grid.Module.UNIQUE_ID = 0;

})();
